// Copyright 2002-2012, University of Colorado
/**
 * Shared constants used in multiple locations within the sim.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Function = require( 'edu.colorado.phet.common.phetcommon.math.Function' );
  var Vector2 = require( 'DOT/Vector2' );
  var Function1 = require( 'edu.colorado.phet.common.phetcommon.util.function.Function1' );
  var Brick = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/Brick' );

// In Kelvin.
  var ROOM_TEMPERATURE = 296;
// In Kelvin.
  var FREEZING_POINT_TEMPERATURE = 273.15;
// In Kelvin.
  var BOILING_POINT_TEMPERATURE = 373.15;
// Time values for normal and fast-forward motion.
  var FRAMES_PER_SECOND = 30.0;
  var SIM_TIME_PER_TICK_NORMAL = 1 / FRAMES_PER_SECOND;
  var SIM_TIME_PER_TICK_FAST_FORWARD = SIM_TIME_PER_TICK_NORMAL * 4;
  var MAX_HEAT_EXCHANGE_TIME_STEP = SIM_TIME_PER_TICK_NORMAL;
// Constants used for creating projections that have a 3D-ish look.
  var Z_TO_X_OFFSET_MULTIPLIER = -0.25;
  var Z_TO_Y_OFFSET_MULTIPLIER = -0.25;
  var MAP_Z_TO_XY_OFFSET = new Function1().withAnonymousClassBody( {
    apply: function( zValue ) {
      return new Vector2( zValue * Z_TO_X_OFFSET_MULTIPLIER, zValue * Z_TO_Y_OFFSET_MULTIPLIER );
    }
  } );
// For comparing temperatures.
// In degrees K.
  var SIGNIFICANT_TEMPERATURE_DIFFERENCE = 1E-3;
// Constant function for energy chunk mapping. The basis for this function
// is that the brick has 2 energy chunks at room temp, one at the freezing
// point of water.

  //private
  var LOW_ENERGY_FOR_MAP_FUNCTION = Brick.ENERGY_AT_WATER_FREEZING_TEMPERATURE;

  //private
  var HIGH_ENERGY_FOR_MAP_FUNCTION = Brick.ENERGY_AT_ROOM_TEMPERATURE;

  //private
  var NUM_ENERGY_CHUNKS_IN_BRICK_AT_FREEZING = 1.25;
// Close to rounding to 3 so that little energy needed to transfer a chunk.

  //private
  var NUM_ENERGY_CHUNKS_IN_BRICK_AT_ROOM_TEMP = 2.4;

  //private
  var MAP_ENERGY_TO_NUM_CHUNKS_DOUBLE = new Function.LinearFunction( LOW_ENERGY_FOR_MAP_FUNCTION, HIGH_ENERGY_FOR_MAP_FUNCTION, NUM_ENERGY_CHUNKS_IN_BRICK_AT_FREEZING, NUM_ENERGY_CHUNKS_IN_BRICK_AT_ROOM_TEMP );

  //private
  var MAP_NUM_CHUNKS_TO_ENERGY_DOUBLE = MAP_ENERGY_TO_NUM_CHUNKS_DOUBLE.createInverse();
  var ENERGY_TO_NUM_CHUNKS_MAPPER = new Function1().withAnonymousClassBody( {
    apply: function( energy ) {
      return Math.max( Math.round( MAP_ENERGY_TO_NUM_CHUNKS_DOUBLE.evaluate( energy ) ), 0 );
    }
  } );
  var ENERGY_PER_CHUNK = MAP_NUM_CHUNKS_TO_ENERGY_DOUBLE.evaluate( 2 ) - MAP_NUM_CHUNKS_TO_ENERGY_DOUBLE.evaluate( 1 );
// Threshold for deciding when two temperatures can be considered equal.
// In Kelvin.
  var TEMPERATURES_EQUAL_THRESHOLD = 1E-6;
// Constant used by all of the "energy systems" in order to keep the amount
// of energy generated, converted, and consumed consistent.
// In joules/sec.
  var MAX_ENERGY_PRODUCTION_RATE = 10000;
// Colors that are used in multiple places.
  var NOMINAL_WATER_OPACITY = 0.75
  f;
  var WATER_COLOR_OPAQUE = new Color( 175, 238, 238 );
  var WATER_COLOR_IN_BEAKER = new Color( 175, 238, 238, (Math.round( NOMINAL_WATER_OPACITY * 255 )) );
  var FIRST_TAB_BACKGROUND_COLOR = new Color( 245, 235, 175 );
  var SECOND_TAB_BACKGROUND_COLOR = FIRST_TAB_BACKGROUND_COLOR;
// Pale gray green.  JB, NP, and AP voted on this as a fave.  Maybe too close to water though.
  var CONTROL_PANEL_BACKGROUND_COLOR = new Color( 199, 229, 199 );
  var CONTROL_PANEL_OUTLINE_STROKE = new BasicStroke( 1.5
  f
  )
  ;
  var CONTROL_PANEL_OUTLINE_COLOR = Color.BLACK;
  var CLOCK_CONTROL_BACKGROUND_COLOR = new Color( 120, 120, 120 );
// Model-view transform scale factor for Energy Systems tab.
  var ENERGY_SYSTEMS_MVT_SCALE_FACTOR = 2200;
// Constants that control the speed of the energy chunks
// In meters/sec.
  var ENERGY_CHUNK_VELOCITY = 0.04;

  return inherit( Object, EFACConstants, {
    },
//statics
    {
      ROOM_TEMPERATURE: ROOM_TEMPERATURE,
      FREEZING_POINT_TEMPERATURE: FREEZING_POINT_TEMPERATURE,
      BOILING_POINT_TEMPERATURE: BOILING_POINT_TEMPERATURE,
      FRAMES_PER_SECOND: FRAMES_PER_SECOND,
      SIM_TIME_PER_TICK_NORMAL: SIM_TIME_PER_TICK_NORMAL,
      SIM_TIME_PER_TICK_FAST_FORWARD: SIM_TIME_PER_TICK_FAST_FORWARD,
      MAX_HEAT_EXCHANGE_TIME_STEP: MAX_HEAT_EXCHANGE_TIME_STEP,
      Z_TO_X_OFFSET_MULTIPLIER: Z_TO_X_OFFSET_MULTIPLIER,
      Z_TO_Y_OFFSET_MULTIPLIER: Z_TO_Y_OFFSET_MULTIPLIER,
      MAP_Z_TO_XY_OFFSET: MAP_Z_TO_XY_OFFSET,
      SIGNIFICANT_TEMPERATURE_DIFFERENCE: SIGNIFICANT_TEMPERATURE_DIFFERENCE,
      ENERGY_TO_NUM_CHUNKS_MAPPER: ENERGY_TO_NUM_CHUNKS_MAPPER,
      ENERGY_PER_CHUNK: ENERGY_PER_CHUNK,
      TEMPERATURES_EQUAL_THRESHOLD: TEMPERATURES_EQUAL_THRESHOLD,
      MAX_ENERGY_PRODUCTION_RATE: MAX_ENERGY_PRODUCTION_RATE,
      NOMINAL_WATER_OPACITY: NOMINAL_WATER_OPACITY,
      WATER_COLOR_OPAQUE: WATER_COLOR_OPAQUE,
      WATER_COLOR_IN_BEAKER: WATER_COLOR_IN_BEAKER,
      FIRST_TAB_BACKGROUND_COLOR: FIRST_TAB_BACKGROUND_COLOR,
      SECOND_TAB_BACKGROUND_COLOR: SECOND_TAB_BACKGROUND_COLOR,
      CONTROL_PANEL_BACKGROUND_COLOR: CONTROL_PANEL_BACKGROUND_COLOR,
      CONTROL_PANEL_OUTLINE_STROKE: CONTROL_PANEL_OUTLINE_STROKE,
      CONTROL_PANEL_OUTLINE_COLOR: CONTROL_PANEL_OUTLINE_COLOR,
      CLOCK_CONTROL_BACKGROUND_COLOR: CLOCK_CONTROL_BACKGROUND_COLOR,
      ENERGY_SYSTEMS_MVT_SCALE_FACTOR: ENERGY_SYSTEMS_MVT_SCALE_FACTOR,
      ENERGY_CHUNK_VELOCITY: ENERGY_CHUNK_VELOCITY
    } );
} );

