// Copyright 2002-2012, University of Colorado
/**
 * Model element that represents a beaker that contains some amount of water,
 * and the water contains energy, which includes energy chunks, and has
 * temperature.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Random = require( 'java.util.Random' );
  var Function = require( 'edu.colorado.phet.common.phetcommon.math.Function' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunkContainerSlice = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EnergyChunkContainerSlice' );
  var EnergyChunkDistributor = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EnergyChunkDistributor' );
  var EnergyContainerCategory = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EnergyContainerCategory' );
  var HorizontalSurface = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/HorizontalSurface' );
  var RectangularThermalMovableModelElement = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/RectangularThermalMovableModelElement' );
  var ThermalContactArea = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/ThermalContactArea' );
  var BOILING_POINT_TEMPERATURE = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants/BOILING_POINT_TEMPERATURE' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// In meters.

  //private
  var MATERIAL_THICKNESS = 0.001;

  //private
  var NUM_SLICES = 6;
// This is seeded for consistent initial energy chunk distribution.

  //private
  var RAND = new Random( 1 );
// Number of degrees Kelvin over which steam is emitted.

  //private
  var STEAMING_RANGE = 10;
// Constants that control the nature of the fluid in the beaker.
// In J/kg-K.  The real value for water is 4186, but this was adjusted so that there

  //private
  var WATER_SPECIFIC_HEAT = 3000;
// aren't too many chunks and so that a chunk is needed as soon as heating starts.
// In kg/m^3, source = design document (and common knowledge).

  //private
  var WATER_DENSITY = 1000.0;
  var INITIAL_FLUID_LEVEL = 0.5;
  /*
   * Constructor.  Initial position is the center bottom of the beaker
   * rectangle.
   */
  function Beaker( clock, initialPosition, width, height, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Property that is used to control and track the amount of fluid in the beaker.
    this.fluidLevel = new Property( INITIAL_FLUID_LEVEL );
    // Top surface, which is the surface upon which other model elements can
    // sit.  For the beaker, this is only slightly above the bottom surface.

    //private
    this.topSurface = new Property( null );
    // Bottom surface, which is the surface that touches the ground, or sits
    // on a burner, etc.

    //private
    this.bottomSurface = new Property( null );
    // Property that allows temperature changes to be monitored.
    this.temperature = new Property( EFACConstants.ROOM_TEMPERATURE );
    // Indicator of how much steam is being emitted.  Ranges from 0 to 1, where
    // 0 is no steam, 1 is the max amount (full boil).
    this.steamingProportion = 0;
    // Max height above water where steam still affects the temperature.

    //private
    this.maxSteamHeight;
    RectangularThermalMovableModelElement.call( this, clock, initialPosition, width, height, calculateWaterMass( width, height * INITIAL_FLUID_LEVEL ), WATER_SPECIFIC_HEAT, energyChunksVisible );
    this.maxSteamHeight = 2 * height;
    // Update the top and bottom surfaces whenever the position changes.
    position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( immutableVector2D ) {
        updateSurfaces();
      }
    } ) );
  }

  return inherit( RectangularThermalMovableModelElement, Beaker, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------

      //private
      updateSurfaces: function() {
        topSurface.set( new HorizontalSurface( new DoubleRange( getRect().getMinX(), getRect().getMaxX() ), getRect().getMinY() + MATERIAL_THICKNESS, this ) );
        bottomSurface.set( new HorizontalSurface( new DoubleRange( getRect().getMinX(), getRect().getMaxX() ), getRect().getMinY(), this ) );
      },
      /*
       * Get the untranslated rectangle that defines the shape of the beaker.
       */
      getRawOutlineRect: function() {
        return new Rectangle.Number( -width / 2, 0, width, height );
      },
      getRect: function() {
        return new Rectangle.Number( position.get().getX() - width / 2, position.get().getY(), width, height );
      },
      stepInTime: function( dt ) {
        super.stepInTime( dt );
        temperature.set( getTemperature() );
        steamingProportion = 0;
        if ( EFACConstants.BOILING_POINT_TEMPERATURE - temperature.get() < STEAMING_RANGE ) {
          // Water is emitting some amount of steam.  Set the proportionate amount.
          steamingProportion = MathUtil.clamp( 0, 1 - ((EFACConstants.BOILING_POINT_TEMPERATURE - temperature.get()) / STEAMING_RANGE), 1 );
        }
      },
      getTopSurfaceProperty: function() {
        return topSurface;
      },
      getBottomSurfaceProperty: function() {
        return bottomSurface;
      },
      addInitialEnergyChunks: function() {
        for ( var slice in slices ) {
          slice.energyChunkList.clear();
        }
        var targetNumChunks = EFACConstants.ENERGY_TO_NUM_CHUNKS_MAPPER.apply( energy );
        var initialChunkBounds = getSliceBounds();
        while ( getNumEnergyChunks() < targetNumChunks ) {
          // Add a chunk at a random location in the beaker.
          addEnergyChunkToNextSlice( new EnergyChunk( EnergyType.THERMAL, EnergyChunkDistributor.generateRandomLocation( initialChunkBounds ), energyChunksVisible ) );
        }
        // Distribute the energy chunks within the beaker.
        for ( var i = 0; i < 1000; i++ ) {
          if ( !EnergyChunkDistributor.updatePositions( slices, EFACConstants.SIM_TIME_PER_TICK_NORMAL ) ) {
            break;
          }
        }
      },
      addEnergyChunkToNextSlice: function( ec ) {
        var totalSliceArea = 0;
        for ( var slice in slices ) {
          totalSliceArea += slice.getShape().getBounds2D().getWidth() * slice.getShape().getBounds2D().getHeight();
        }
        var sliceSelectionValue = RAND.nextDouble();
        var chosenSlice = slices.get( 0 );
        var accumulatedArea = 0;
        for ( var slice in slices ) {
          accumulatedArea += slice.getShape().getBounds2D().getWidth() * slice.getShape().getBounds2D().getHeight();
          if ( accumulatedArea / totalSliceArea >= sliceSelectionValue ) {
            chosenSlice = slice;
            break;
          }
        }
        chosenSlice.addEnergyChunk( ec );
      },

      //private
      calculateWaterMass: function( width, height ) {
        return Math.PI * Math.pow( width / 2, 2 ) * height * WATER_DENSITY;
      },
      getThermalContactArea: function() {
        return new ThermalContactArea( new Rectangle.Number( position.get().getX() - width / 2, position.get().getY(), width, height * fluidLevel.get() ), true );
      },
      /*
       * Get the area where the temperature of the steam can be sensed.
       */
      getSteamArea: function() {
        // Height of steam rectangle is based on beaker height and steamingProportion.
        var liquidWaterHeight = height * fluidLevel.get();
        return new Rectangle.Number( position.get().getX() - width / 2, position.get().getY() + liquidWaterHeight, width, maxSteamHeight );
      },
      getSteamTemperature: function( heightAboveWater ) {
        var mappingFunction = new Function.LinearFunction( 0, maxSteamHeight * steamingProportion, temperature.get(), EFACConstants.ROOM_TEMPERATURE );
        return Math.max( mappingFunction.evaluate( heightAboveWater ), EFACConstants.ROOM_TEMPERATURE );
      },
      addEnergyChunkSlices: function() {
        // Check that his has not been already called.
        assert && assert( slices.size() == 0 );
        var fluidRect = new Rectangle.Number( position.get().getX() - width / 2, position.get().getY(), width, height * INITIAL_FLUID_LEVEL );
        var widthYProjection = Math.abs( width * EFACConstants.Z_TO_Y_OFFSET_MULTIPLIER );
        for ( var i = 0; i < NUM_SLICES; i++ ) {
          var proportion = (i + 1) * (1 / (NUM_SLICES + 1));
          var slicePath = new DoubleGeneralPath();
          {
            // yield width value proportional to position in Z-space.
            var sliceWidth = (-Math.pow( (2 * proportion - 1), 2 ) + 1) * fluidRect.getWidth();
            var bottomY = fluidRect.getMinY() - (widthYProjection / 2) + (proportion * widthYProjection);
            var topY = bottomY + fluidRect.getHeight();
            var centerX = fluidRect.getCenterX();
            var controlPointYOffset = (bottomY - fluidRect.getMinY()) * 0.5;
            slicePath.moveTo( centerX - sliceWidth / 2, bottomY );
            slicePath.curveTo( centerX - sliceWidth * 0.33, bottomY + controlPointYOffset, centerX + sliceWidth * 0.33, bottomY + controlPointYOffset, centerX + sliceWidth / 2, bottomY );
            slicePath.lineTo( centerX + sliceWidth / 2, topY );
            slicePath.curveTo( centerX + sliceWidth * 0.33, topY + controlPointYOffset, centerX - sliceWidth * 0.33, topY + controlPointYOffset, centerX - sliceWidth / 2, topY );
            slicePath.lineTo( centerX - sliceWidth / 2, bottomY );
          }
          slices.add( new EnergyChunkContainerSlice( slicePath.getGeneralPath(), -proportion * width, position ) );
        }
      },
      getEnergyContainerCategory: function() {
        return EnergyContainerCategory.WATER;
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.beaker;
      },
      getEnergyBeyondMaxTemperature: function() {
        return Math.max( energy - (BOILING_POINT_TEMPERATURE * mass * specificHeat), 0 );
      },
// Limit max temp to the boiling point.
      getTemperature: function() {
        return Math.min( super.getTemperature(), BOILING_POINT_TEMPERATURE );
      },
      /*
       * This override handles the case where the point is above the beaker.
       * In this case, we want to pull from all slices evenly, and not favor
       * the slices the bump up at the top in order to match the 3D look of the
       * water surface.
       */
      extractClosestEnergyChunk: function( point ) {
        var pointIsAboveWaterSurface = true;
        for ( var slice in slices ) {
          if ( point.getY() < slice.getShape().getBounds2D().getMaxY() ) {
            pointIsAboveWaterSurface = false;
            break;
          }
        }
        if ( !pointIsAboveWaterSurface ) {
          return super.extractClosestEnergyChunk( point );
        }
        // density, since this is where we will get the energy chunk.
        var maxSliceDensity = 0;
        var densestSlice = null;
        for ( var slice in slices ) {
          var sliceDensity = slice.energyChunkList.size() / (slice.getShape().getBounds2D().getWidth() * slice.getShape().getBounds2D().getHeight());
          if ( sliceDensity > maxSliceDensity ) {
            maxSliceDensity = sliceDensity;
            densestSlice = slice;
          }
        }
        if ( densestSlice == null || densestSlice.energyChunkList.size() == 0 ) {
          console.log( getClass().getName() + " - Warning: No energy chunks in the beaker, can't extract any." );
          return null;
        }
        var highestEnergyChunk = densestSlice.energyChunkList.get( 0 );
        for ( var ec in densestSlice.energyChunkList ) {
          if ( ec.position.get().getY() > highestEnergyChunk.position.get().getY() ) {
            highestEnergyChunk = ec;
          }
        }
        removeEnergyChunk( highestEnergyChunk );
        return highestEnergyChunk;
      }
    },
//statics
    {
      INITIAL_FLUID_LEVEL: INITIAL_FLUID_LEVEL
    } );
} );

