// Copyright 2002-2012, University of Colorado
/**
 * Object that represents a beaker in the view.  This representation is split
 * between a front node and a back node, which must be separately added to the
 * canvas.  This is done to allow a layering effect.  Hence, this cannot be
 * added directly to the canvas, and the client must add each layer separately.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Area = require( 'java.awt.geom.Area' );
  var Ellipse2D = require( 'java.awt.geom.Ellipse2D' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var IClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.IClock' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var ColorUtils = require( 'edu.colorado.phet.common.phetcommon.view.util.ColorUtils' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var Beaker = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/Beaker' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyChunkContainerSliceNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EnergyChunkContainerSliceNode' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PText = require( 'edu.umd.cs.piccolo.nodes.PText' );
  var PClip = require( 'edu.umd.cs.piccolox.nodes.PClip' );


  //private
  var OUTLINE_STROKE = new BasicStroke( 3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL );

  //private
  var OUTLINE_COLOR = Color.LIGHT_GRAY;

  //private
  var PERSPECTIVE_PROPORTION = -EFACConstants.Z_TO_Y_OFFSET_MULTIPLIER;

  //private
  var LABEL_FONT = new PhetFont( 32, false );

  //private
  var SHOW_MODEL_RECT = false;

  //private
  var BEAKER_COLOR = new Color( 250, 250, 250, 100 );

  //private
  var RAND = new Random();

  // static class: PerspectiveWaterNode
  var PerspectiveWaterNode =

    //private
    define( function( require ) {

      //private
      var LIQUID_WATER_OUTLINE_COLOR = ColorUtils.darkerColor( EFACConstants.WATER_COLOR_IN_BEAKER, 0.2 );

      //private
      var WATER_OUTLINE_STROKE = new BasicStroke( 2 );
// Number of degrees Kelvin over which steam is visible.

      //private
      var STEAMING_RANGE = 10;
// In screen coords (basically pixels) per second.

      //private
      var STEAM_BUBBLE_SPEED_RANGE = new DoubleRange( 100, 125 );
// In screen coords (basically pixels).

      //private
      var STEAM_BUBBLE_DIAMETER_RANGE = new DoubleRange( 20, 50 );

      //private
      var MAX_STEAM_BUBBLE_HEIGHT = 300;
// Bubbles per second.

      //private
      var STEAM_BUBBLE_PRODUCTION_RATE_RANGE = new DoubleRange( 20, 40 );
// Proportion per second.

      //private
      var STEAM_BUBBLE_GROWTH_RATE = 0.2;
// Proportion, 1 is max.

      //private
      var MAX_STEAM_BUBBLE_OPACITY = 0.7;

      // static class: SteamBubble
      var SteamBubble =

        //private
        define( function( require ) {
          function SteamBubble( initialDiameter, initialOpacity ) {
            PhetPPath.call( this, new Ellipse2D.Number( -initialDiameter / 2, -initialDiameter / 2, initialDiameter, initialDiameter ), new Color( 255, 255, 255, (initialOpacity * 255) ) );
          }

          return inherit( PhetPPath, SteamBubble, {
            setOpacity: function( opacity ) {
              assert && assert( opacity <= 1.0 );
              setPaint( new Color( 255, 255, 255, (opacity * 255) ) );
            },
            getDiameter: function() {
              return getFullBoundsReference().getWidth();
            },
            setDiameter: function( newDiameter ) {
              setPathTo( new Ellipse2D.Number( -newDiameter / 2, -newDiameter / 2, newDiameter, newDiameter ) );
            }
          } );
        } );
      ;
      /*
       * Constructor.
       */

      //private
      function PerspectiveWaterNode( clock, beakerOutlineRect, waterLevel, temperature ) {
        // Nodes that comprise this node.

        //private
        this.liquidWaterTopNode = new PhetPPath( EFACConstants.WATER_COLOR_IN_BEAKER, WATER_OUTLINE_STROKE, LIQUID_WATER_OUTLINE_COLOR );

        //private
        this.liquidWaterBodyNode = new PhetPPath( EFACConstants.WATER_COLOR_IN_BEAKER, WATER_OUTLINE_STROKE, LIQUID_WATER_OUTLINE_COLOR );

        //private
        this.steamBubbles = [];

        //private
        this.steamNode;
        // Miscellaneous other variables.

        //private
        this.bubbleProductionRemainder;
        addChild( liquidWaterBodyNode );
        addChild( liquidWaterTopNode );
        steamNode = new Node();
        addChild( steamNode );
        clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
          clockTicked: function( clockEvent ) {
            updateAppearance( waterLevel.get(), beakerOutlineRect, temperature.get(), clockEvent.getSimulationTimeChange() );
          },
          simulationTimeReset: function( clockEvent ) {
            // Get rid of steam when a reset occurs.
            steamBubbles.clear();
            steamNode.removeAllChildren();
          }
        } ) );
        updateAppearance( waterLevel.get(), beakerOutlineRect, temperature.get(), 1 / EFACConstants.FRAMES_PER_SECOND );
      }

      return inherit( Node, PerspectiveWaterNode, {

        //private
        updateAppearance: function( fluidLevel, beakerOutlineRect, temperature, dt ) {
          var waterHeight = beakerOutlineRect.getHeight() * fluidLevel;
          var liquidWaterRect = new Rectangle.Number( beakerOutlineRect.getX(), beakerOutlineRect.getMaxY() - waterHeight, beakerOutlineRect.getWidth(), waterHeight );
          var ellipseWidth = beakerOutlineRect.getWidth();
          var ellipseHeight = PERSPECTIVE_PROPORTION * ellipseWidth;
          var liquidWaterTopEllipse = new Ellipse2D.Number( liquidWaterRect.getMinX(), liquidWaterRect.getMinY() - ellipseHeight / 2, liquidWaterRect.getWidth(), ellipseHeight );
          var bottomEllipse = new Ellipse2D.Number( liquidWaterRect.getMinX(), liquidWaterRect.getMaxY() - ellipseHeight / 2, liquidWaterRect.getWidth(), ellipseHeight );
          // Update shape of the the liquid water.
          var liquidWaterBodyArea = new Area( liquidWaterRect );
          liquidWaterBodyArea.add( new Area( bottomEllipse ) );
          liquidWaterBodyArea.subtract( new Area( liquidWaterTopEllipse ) );
          liquidWaterBodyNode.setPathTo( liquidWaterBodyArea );
          liquidWaterTopNode.setPathTo( liquidWaterTopEllipse );
          var steamingProportion = 0;
          if ( EFACConstants.BOILING_POINT_TEMPERATURE - temperature < STEAMING_RANGE ) {
            // Water is emitting some amount of steam.  Set the proportionate amount.
            steamingProportion = MathUtil.clamp( 0, 1 - ((EFACConstants.BOILING_POINT_TEMPERATURE - temperature) / STEAMING_RANGE), 1 );
          }
          if ( steamingProportion > 0 ) {
            // Add any new steam bubbles.
            var bubblesToProduceCalc = (STEAM_BUBBLE_PRODUCTION_RATE_RANGE.getMin() + STEAM_BUBBLE_PRODUCTION_RATE_RANGE.getLength() * steamingProportion) * dt;
            var bubblesToProduce = Math.floor( bubblesToProduceCalc );
            bubbleProductionRemainder += bubblesToProduceCalc - bubblesToProduce;
            if ( bubbleProductionRemainder >= 1 ) {
              bubblesToProduce += Math.floor( bubbleProductionRemainder );
              bubbleProductionRemainder -= Math.floor( bubbleProductionRemainder );
            }
            for ( var i = 0; i < bubblesToProduce; i++ ) {
              var steamBubbleDiameter = STEAM_BUBBLE_DIAMETER_RANGE.getMin() + RAND.nextDouble() * STEAM_BUBBLE_DIAMETER_RANGE.getLength();
              var steamBubbleCenterXPos = beakerOutlineRect.getCenterX() + (RAND.nextDouble() - 0.5) * (beakerOutlineRect.getWidth() - steamBubbleDiameter);
              var steamBubble = new SteamBubble( steamBubbleDiameter, steamingProportion );
              // Invisible to start, will fade in.
              steamBubble.setOffset( steamBubbleCenterXPos, liquidWaterRect.getMinY() );
              steamBubble.setOpacity( 0 );
              steamBubbles.add( steamBubble );
              steamNode.addChild( steamBubble );
            }
          }
          // Update the position and appearance of the existing steam bubbles.
          var steamBubbleSpeed = STEAM_BUBBLE_SPEED_RANGE.getMin() + steamingProportion * STEAM_BUBBLE_SPEED_RANGE.getLength();
          var unfilledBeakerHeight = beakerOutlineRect.getHeight() - waterHeight;
          for ( var steamBubble in new ArrayList( steamBubbles ) ) {
            steamBubble.setOffset( steamBubble.getXOffset(), steamBubble.getYOffset() + dt * (-steamBubbleSpeed) );
            if ( beakerOutlineRect.getMinY() - steamBubble.getYOffset() > MAX_STEAM_BUBBLE_HEIGHT ) {
              steamBubbles.remove( steamBubble );
              steamNode.removeChild( steamBubble );
            }
            else if ( steamBubble.getYOffset() < beakerOutlineRect.getMinY() ) {
              steamBubble.setDiameter( steamBubble.getDiameter() * (1 + (STEAM_BUBBLE_GROWTH_RATE * dt)) );
              var distanceFromCenterX = steamBubble.getXOffset() - beakerOutlineRect.getCenterX();
              steamBubble.setOffset( steamBubble.getXOffset() + (distanceFromCenterX * 0.2 * dt), steamBubble.getYOffset() );
              // Fade the bubble as it reaches the end of its range.
              steamBubble.setOpacity( (1 - (beakerOutlineRect.getMinY() - steamBubble.getYOffset()) / MAX_STEAM_BUBBLE_HEIGHT) * MAX_STEAM_BUBBLE_OPACITY );
            }
            else {
              // Fade the bubble in.
              var distanceFromWater = liquidWaterRect.getMinY() - steamBubble.getYOffset();
              steamBubble.setOpacity( MathUtil.clamp( 0, distanceFromWater / (unfilledBeakerHeight / 4), 1 ) * MAX_STEAM_BUBBLE_OPACITY );
            }
          }
        },
      } );
    } );
  ;
  function BeakerView( clock, beaker, energyChunksVisible, mvt ) {
    this.mvt;
    this.energyChunkClipNode;
    this.frontNode = new Node();

    //private
    this.backNode = new Node();
    this.grabNode = new Node();
    this.mvt = mvt;
    // shape from the position.
    var scaleTransform = AffineTransform.getScaleInstance( mvt.getTransform().getScaleX(), mvt.getTransform().getScaleY() );
    // location in the view.
    var beakerViewRect = scaleTransform.createTransformedShape( beaker.getRawOutlineRect() ).getBounds2D();
    // ellipses in order to create a 3D-ish look.
    var ellipseHeight = beakerViewRect.getWidth() * PERSPECTIVE_PROPORTION;
    var topEllipse = new Ellipse2D.Number( beakerViewRect.getMinX(), beakerViewRect.getMinY() - ellipseHeight / 2, beakerViewRect.getWidth(), ellipseHeight );
    var bottomEllipse = new Ellipse2D.Number( beakerViewRect.getMinX(), beakerViewRect.getMaxY() - ellipseHeight / 2, beakerViewRect.getWidth(), ellipseHeight );
    // Add the bottom ellipse.
    backNode.addChild( new PhetPPath( bottomEllipse, BEAKER_COLOR, OUTLINE_STROKE, OUTLINE_COLOR ) );
    // Add the water.  It will adjust its size based on the fluid level.
    var water = new PerspectiveWaterNode( clock, beakerViewRect, beaker.fluidLevel, beaker.temperature );
    frontNode.addChild( water );
    // Create and add the shape for the body of the beaker.
    var beakerBody = new Area( beakerViewRect );
    beakerBody.add( new Area( bottomEllipse ) );
    beakerBody.subtract( new Area( topEllipse ) );
    frontNode.addChild( new PhetPPath( beakerBody, BEAKER_COLOR, OUTLINE_STROKE, OUTLINE_COLOR ) );
    // Add the top ellipse.  It is behind the water for proper Z-order behavior.
    backNode.addChild( new PhetPPath( topEllipse, BEAKER_COLOR, OUTLINE_STROKE, OUTLINE_COLOR ) );
    // grab the beaker.
    backNode.addChild( new PhetPPath( beakerViewRect, new Color( 0, 0, 0, 0 ) ) );
    // remove things from the beaker.
    frontNode.setPickable( false );
    frontNode.setChildrenPickable( false );
    backNode.setPickable( false );
    backNode.setChildrenPickable( false );
    // Add the label.  Position it just below the front, top water line.
    var label = new PText( EnergyFormsAndChangesResources.Strings.WATER );
    label.setFont( LABEL_FONT );
    label.setOffset( beakerViewRect.getCenterX() - label.getFullBoundsReference().width / 2, beakerViewRect.getMaxY() - beakerViewRect.getHeight() * beaker.fluidLevel.get() + topEllipse.getHeight() / 2 );
    label.setPickable( false );
    label.setChildrenPickable( false );
    frontNode.addChild( label );
    // other model elements.
    var energyChunkRootNode = new Node();
    backNode.addChild( energyChunkRootNode );
    energyChunkClipNode = new PClip();
    // Not sure that this is what is needed here. Bigger for chunks that are leaving? Needs thought.
    energyChunkClipNode.setPathTo( beakerViewRect );
    energyChunkRootNode.addChild( energyChunkClipNode );
    energyChunkClipNode.setStroke( null );
    for ( var i = beaker.getSlices().size() - 1; i >= 0; i-- ) {
      energyChunkClipNode.addChild( new EnergyChunkContainerSliceNode( beaker.getSlices().get( i ), mvt ) );
    }
    // this model element and add/remove them as needed.
    beaker.approachingEnergyChunks.addElementAddedObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( addedEnergyChunk ) {
        var energyChunkNode = new EnergyChunkNode( addedEnergyChunk, mvt );
        energyChunkRootNode.addChild( energyChunkNode );
        beaker.approachingEnergyChunks.addElementRemovedObserver( new VoidFunction1().withAnonymousClassBody( {
          apply: function( removedEnergyChunk ) {
            if ( removedEnergyChunk == addedEnergyChunk ) {
              energyChunkRootNode.removeChild( energyChunkNode );
              beaker.approachingEnergyChunks.removeElementRemovedObserver( this );
            }
          }
        } ) );
      }
    } ) );
    // Add the node that can be used to grab and move the beaker.
    var grabNodeShape = new Area( beakerBody );
    grabNodeShape.add( new Area( topEllipse ) );
    // Invisible, yet pickable.
    grabNode.addChild( new PhetPPath( grabNodeShape, new Color( 0, 0, 0, 0 ) ) );
    // beaker's position in the model.
    if ( SHOW_MODEL_RECT ) {
      frontNode.addChild( new PhetPPath( beakerViewRect, new BasicStroke( 1 ), Color.RED ) );
    }
    // Update the offset if and when the model position changes.
    beaker.position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( position ) {
        frontNode.setOffset( mvt.modelToView( position ).toPoint2D() );
        backNode.setOffset( mvt.modelToView( position ).toPoint2D() );
        grabNode.setOffset( mvt.modelToView( position ).toPoint2D() );
        // nodes can handle their own positioning.
        energyChunkRootNode.setOffset( mvt.modelToView( position ).getRotatedInstance( Math.PI ).toPoint2D() );
      }
    } ) );
    // chunk visibility.
    energyChunksVisible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyChunksVisible ) {
        label.setTransparency( energyChunksVisible ? 0.5f
        :
        1
        f
        )
        ;
        water.setTransparency( energyChunksVisible ? EFACConstants.NOMINAL_WATER_OPACITY / 2 : EFACConstants.NOMINAL_WATER_OPACITY );
      }
    } ) );
  }

  return inherit( Object, BeakerView, {
    getFrontNode: function() {
      return frontNode;
    },
    getBackNode: function() {
      return backNode;
    },
    getGrabNode: function() {
      return grabNode;
    }
  } );
} );

