// Copyright 2002-2012, University of Colorado
/**
 * Node that represents a burner stand in the view.  A burner stand is a faux-
 * 3D representation of an object upon which other objects may be placed in
 * order to be heated or cooled.  The burner is not included.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var BasicStroke = require( 'java.awt.BasicStroke' );
  var Color = require( 'SCENERY/util/Color' );
  var Stroke = require( 'java.awt.Stroke' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Vector2 = require( 'DOT/Vector2' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var Node = require( 'SCENERY/nodes/Node' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Constants that control some aspect of appearance.  These can be made
// into constructor params if it is ever desirable to do so.

  //private
  var BURNER_STAND_STROKE = new BasicStroke( 4, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL );

  //private
  var BURNER_STAND_STROKE_COLOR = Color.BLACK;
// Positive is counterclockwise, a value of 0 produces a non-skewed rectangle.
  var PERSPECTIVE_ANGLE = Math.PI / 4;

  // static class: BurnerStandSide
  var BurnerStandSide =

    //private
    define( function( require ) {

      //private
      function BurnerStandSide( topCenter, height, edgeLength ) {
        // Draw the side as a parallelogram.
        var upperLeftCorner = topCenter.plus( new Vector2( -edgeLength / 2, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE ) );
        var lowerLeftCorner = upperLeftCorner.plus( new Vector2( 0, height ) );
        var lowerRightCorner = lowerLeftCorner.plus( new Vector2( edgeLength, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE ) );
        var upperRightCorner = lowerRightCorner.plus( new Vector2( 0, -height ) );
        var path = new DoubleGeneralPath( topCenter );
        path.lineTo( upperLeftCorner );
        path.lineTo( lowerLeftCorner );
        path.lineTo( lowerRightCorner );
        path.lineTo( upperRightCorner );
        path.closePath();
        addChild( new PhetPPath( path.getGeneralPath(), BURNER_STAND_STROKE, BURNER_STAND_STROKE_COLOR ) );
      }

      return inherit( Node, BurnerStandSide, {
      } );
    } );
  ;
  // static class: BurnerStandTop
  var BurnerStandTop =

    //private
    define( function( require ) {

      //private
      function BurnerStandTop( leftCenter, width, edgeLength ) {
        // Create the points for the outline of the perspective rectangle.
        var upperLeftCorner = leftCenter.plus( new Vector2( edgeLength / 2, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE ) );
        var upperRightCorner = upperLeftCorner.plus( new Vector2( width, 0 ) );
        var lowerRightCorner = upperRightCorner.plus( new Vector2( -edgeLength, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE ) );
        var lowerLeftCorner = lowerRightCorner.plus( new Vector2( -width, 0 ) );
        // Create the points for the circular opening in the top.
        var upperLeftCircularOpeningCorner = upperLeftCorner.plus( new Vector2( width * 0.25, 0 ) );
        var upperRightCircularOpeningCorner = upperLeftCorner.plus( new Vector2( width * 0.75, 0 ) );
        var lowerLeftCircularOpeningCorner = lowerLeftCorner.plus( new Vector2( width * 0.25, 0 ) );
        var lowerRightCircularOpeningCorner = lowerLeftCorner.plus( new Vector2( width * 0.75, 0 ) );
        // Create the control points for the circular opening in the top.
        var circularOpeningPerspectiveVector = new Vector2( edgeLength * 0.5, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE );
        var path = new DoubleGeneralPath();
        path.moveTo( upperLeftCorner );
        path.lineTo( upperLeftCircularOpeningCorner );
        path.curveTo( upperLeftCircularOpeningCorner.plus( circularOpeningPerspectiveVector ), upperRightCircularOpeningCorner.plus( circularOpeningPerspectiveVector ), upperRightCircularOpeningCorner );
        path.lineTo( upperRightCorner );
        path.lineTo( lowerRightCorner );
        path.lineTo( lowerRightCircularOpeningCorner );
        path.curveTo( lowerRightCircularOpeningCorner.minus( circularOpeningPerspectiveVector ), lowerLeftCircularOpeningCorner.minus( circularOpeningPerspectiveVector ), lowerLeftCircularOpeningCorner );
        path.lineTo( lowerLeftCorner );
        path.closePath();
        addChild( new PhetPPath( path.getGeneralPath(), BURNER_STAND_STROKE, BURNER_STAND_STROKE_COLOR ) );
      }

      return inherit( Node, BurnerStandTop, {
      } );
    } );
  ;
  function BurnerStandNode( rect, projectedEdgeLength ) {
    // Add the left and right sides of the stand.
    addChild( new BurnerStandSide( new Vector2( rect.getX(), rect.getY() ), rect.getHeight(), projectedEdgeLength ) );
    addChild( new BurnerStandSide( new Vector2( rect.getMaxX(), rect.getY() ), rect.getHeight(), projectedEdgeLength ) );
    // Add the top of the burner stand.
    addChild( new BurnerStandTop( new Vector2( rect.getX(), rect.getY() ), rect.getWidth(), projectedEdgeLength ) );
  }

  return inherit( Node, BurnerStandNode, {
    },
//statics
    {
      PERSPECTIVE_ANGLE: PERSPECTIVE_ANGLE
    } );
} );

