// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a chunk of energy in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var HashMap = require( 'java.util.HashMap' );
  var Map = require( 'java.util.Map' );
  var Vector2 = require( 'DOT/Vector2' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var PhetPText = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPText' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PImage = require( 'edu.umd.cs.piccolo.nodes.PImage' );

// In meters.
  var Z_DISTANCE_WHERE_FULLY_FADED = 0.1;
// In screen coords, which is close to pixels.

  //private
  var WIDTH = 24;

  //private
  var mapEnergyTypeToImage = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyType.THERMAL, EnergyFormsAndChangesResources.Images.E_THERM_BLANK_ORANGE );
      put( EnergyType.ELECTRICAL, EnergyFormsAndChangesResources.Images.E_ELECTRIC_BLANK );
      put( EnergyType.MECHANICAL, EnergyFormsAndChangesResources.Images.E_MECH_BLANK );
      put( EnergyType.LIGHT, EnergyFormsAndChangesResources.Images.E_LIGHT_BLANK );
      put( EnergyType.CHEMICAL, EnergyFormsAndChangesResources.Images.E_CHEM_BLANK_LIGHT );
      put( EnergyType.HIDDEN, EnergyFormsAndChangesResources.Images.E_DASHED_BLANK );
    }
  } );

  function EnergyChunkNode( energyChunk, mvt ) {
    // Control the overall visibility of this node.
    energyChunk.visible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( visible ) {
        setVisible( visible );
      }
    } ) );
    // Set up updating of transparency based on Z position.
    energyChunk.zPosition.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( zPosition ) {
        updateTransparency( zPosition );
      }
    } ) );
    // Monitor the energy type and make the image match it.
    energyChunk.energyType.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyType ) {
        removeAllChildren();
        addChild( createEnergyChunkNode( energyType ) );
      }
    } ) );
    // Set this node's position when the corresponding model element moves.
    energyChunk.position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( immutableVector2D ) {
        setOffset( mvt.modelToView( immutableVector2D ).toPoint2D() );
      }
    } ) );
  }

  return inherit( Node, EnergyChunkNode, {
// Update the transparency, which is a function of several factors.

      //private
      updateTransparency: function( zPosition ) {
        var zFadeValue = 1;
        if ( zPosition < 0 ) {
          zFadeValue = Math.max( (Z_DISTANCE_WHERE_FULLY_FADED + zPosition) / Z_DISTANCE_WHERE_FULLY_FADED, 0 );
        }
        setTransparency( zFadeValue );
      },
      createEnergyChunkNode: function( energyType ) {
        var background = new PImage( mapEnergyTypeToImage.get( energyType ) );
        background.addChild( new PhetPText( EnergyFormsAndChangesResources.Strings.ENERGY_CHUNK_LABEL, new PhetFont( 16, true ) ).withAnonymousClassBody( {
          initializer: function() {
            setScale( Math.min( background.getFullBoundsReference().width / getFullBoundsReference().width, background.getFullBoundsReference().height / getFullBoundsReference().height ) * 0.95 );
            setOffset( background.getFullBoundsReference().width / 2 - getFullBoundsReference().width / 2, background.getFullBoundsReference().height / 2 - getFullBoundsReference().height / 2 );
          }
        } ) );
        background.setScale( WIDTH / background.getFullBoundsReference().width );
        background.setOffset( -background.getFullBoundsReference().width / 2, -background.getFullBoundsReference().height / 2 );
        return background;
      }
    },
//statics
    {
      Z_DISTANCE_WHERE_FULLY_FADED: Z_DISTANCE_WHERE_FULLY_FADED
    } );
} );

