// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a chunk of energy in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var BasicStroke = require( 'java.awt.BasicStroke' );
  var Color = require( 'SCENERY/util/Color' );
  var Point = require( 'java.awt.Point' );
  var Stroke = require( 'java.awt.Stroke' );
  var Dimension2D = require( 'java.awt.geom.Dimension2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var RoundRectangle2D = require( 'java.awt.geom.RoundRectangle2D' );
  var HashMap = require( 'java.util.HashMap' );
  var Map = require( 'java.util.Map' );
  var JFrame = require( 'javax.swing.JFrame' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var PhetColorScheme = require( 'edu.colorado.phet.common.phetcommon.view.PhetColorScheme' );
  var RoundGradientPaint = require( 'edu.colorado.phet.common.phetcommon.view.graphics.RoundGradientPaint' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var ColorUtils = require( 'edu.colorado.phet.common.phetcommon.view.util.ColorUtils' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var PhetPCanvas = require( 'edu.colorado.phet.common.piccolophet.PhetPCanvas' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var PhetPText = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPText' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PDimension = require( 'edu.umd.cs.piccolo.util.PDimension' );

// In screen coords, which is close to pixels.

  //private
  var WIDTH = 24;
// In meters.

  //private
  var Z_DISTANCE_WHERE_FULLY_FADED = 0.1;

  //private
  var OUTLINE_STROKE = new BasicStroke( 2 );

  //private
  var OUTLINE_STROKE_COLOR = Color.BLACK;

  //private
  var mapEnergyTypeToColor = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      // TODO: If we end up using these, these color will need adjustment.
      put( EnergyType.THERMAL, PhetColorScheme.RED_COLORBLIND );
      put( EnergyType.ELECTRICAL, Color.YELLOW );
      put( EnergyType.MECHANICAL, Color.GRAY );
      put( EnergyType.LIGHT, Color.YELLOW );
      put( EnergyType.CHEMICAL, Color.GREEN );
    }
  } );

  function EnergyChunkNode2( energyChunk, mvt ) {
    // Control the overall visibility of this node.
    energyChunk.visible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( visible ) {
        setVisible( visible );
      }
    } ) );
    // Set up updating of transparency based on Z position.
    energyChunk.zPosition.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( zPosition ) {
        updateTransparency( zPosition );
      }
    } ) );
    // Draw the energy chunks.
    var body = new PhetPPath( new RoundRectangle2D.Number( -WIDTH / 2, -WIDTH / 2, WIDTH, WIDTH, WIDTH / 4, WIDTH / 4 ), OUTLINE_STROKE, OUTLINE_STROKE_COLOR );
    addChild( body );
    // Monitor the energy type and set the color appropriately.
    energyChunk.energyType.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyType ) {
        var baseColor = mapEnergyTypeToColor.get( energyType );
        var paint = new RoundGradientPaint( WIDTH / 4, -WIDTH / 4, ColorUtils.brighterColor( baseColor, 0.5 ), new Vector2( -WIDTH / 2, WIDTH / 2 ), baseColor );
        body.setPaint( paint );
      }
    } ) );
    // Add the label.
    addChild( new PhetPText( "E", new PhetFont( 20, true ) ).withAnonymousClassBody( {
      initializer: function() {
        setOffset( -getFullBoundsReference().width / 2, -getFullBoundsReference().height / 2 );
        setTextPaint( Color.BLACK );
      }
    } ) );
    // Set this node's position when the corresponding model element moves.
    energyChunk.position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( immutableVector2D ) {
        setOffset( mvt.modelToView( immutableVector2D ).toPoint2D() );
      }
    } ) );
  }

  return inherit( Node, EnergyChunkNode2, {
// Update the transparency, which is a function of several factors.

    //private
    updateTransparency: function( zPosition ) {
      var zFadeValue = 1;
      if ( zPosition < 0 ) {
        zFadeValue = Math.max( (Z_DISTANCE_WHERE_FULLY_FADED + zPosition) / Z_DISTANCE_WHERE_FULLY_FADED, 0 );
      }
      setTransparency( zFadeValue );
    },
// Test harness.
    main: function( args ) {
      var stageSize = new PDimension( 200, 100 );
      var canvas = new PhetPCanvas();
      // Set up the canvas-screen transform.
      canvas.setWorldTransformStrategy( new PhetPCanvas.CenteredStage( canvas, stageSize ) );
      var mvt = ModelViewTransform.createSinglePointScaleInvertedYMapping( new Vector2( 0, 0 ), new Point( Math.round( stageSize.getWidth() * 0.5 ), Math.round( stageSize.getHeight() * 0.50 ) ), // "Zoom factor" - smaller zooms out, larger zooms in.
        1 );
      canvas.getLayer().addChild( new EnergyChunkNode2( new EnergyChunk( EnergyType.ELECTRICAL, 0, 0, new BooleanProperty( true ) ), mvt ) );
      // Boiler plate app stuff.
      var frame = new JFrame();
      frame.setContentPane( canvas );
      frame.setSize( stageSize.getWidth(), stageSize.getHeight() );
      frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
      // Center.
      frame.setLocationRelativeTo( null );
      frame.setVisible( true );
    }
  } );
} );

