// Copyright 2002-2012, University of Colorado
/**
 * This class represents a bicycle being peddled by a rider in order to
 * generate energy.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var ChangeObserver = require( 'edu.colorado.phet.common.phetcommon.model.property.ChangeObserver' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var Images = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images' );//static ///*

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// In radians/sec.
  var MAX_ANGULAR_VELOCITY_OF_CRANK = 3 * Math.PI;
// In radians/(sec^2).

  //private
  var ANGULAR_ACCELERATION = Math.PI / 2;
// In joules / sec

  //private
  var MAX_ENERGY_OUTPUT_WHEN_CONNECTED_TO_GENERATOR = EFACConstants.MAX_ENERGY_PRODUCTION_RATE;
// In joules / sec

  //private
  var MAX_ENERGY_OUTPUT_WHEN_RUNNING_FREE = MAX_ENERGY_OUTPUT_WHEN_CONNECTED_TO_GENERATOR / 5;

  //private
  var CRANK_TO_REAR_WHEEL_RATIO = 1;

  //private
  var INITIAL_NUM_ENERGY_CHUNKS = 15;

  //private
  var RAND = new Random();

  //private
  var MECHANICAL_TO_THERMAL_CHUNK_RATIO = 5;
// In meters, must be worked out with the image.
  var REAR_WHEEL_RADIUS = 0.02;
// Offset of the bike frame center.  Most other image offsets are relative
// to this one.

  //private
  var FRAME_CENTER_OFFSET = new Vector2( 0.0, 0.01 );
// Images used when representing this model element in the view.  The
// offsets, which are in meters, were empirically determined.  The values
// aren't really to scale, since there are so many things in this model
// with very different scales.
  var FRAME_IMAGE = new ModelElementImage( BICYCLE_FRAME_3, FRAME_CENTER_OFFSET );
  var REAR_WHEEL_SPOKES_IMAGE = new ModelElementImage( BICYCLE_SPOKES, FRAME_CENTER_OFFSET.plus( new Vector2( 0.035, -0.020 ) ) );
  var RIDER_NORMAL_UPPER_BODY_IMAGE = new ModelElementImage( BICYCLE_RIDER, FRAME_CENTER_OFFSET.plus( new Vector2( -0.0025, 0.062 ) ) );
  var RIDER_TIRED_UPPER_BODY_IMAGE = new ModelElementImage( BICYCLE_RIDER_TIRED, FRAME_CENTER_OFFSET.plus( new Vector2( -0.0032, 0.056 ) ) );

  //private
  var LEG_IMAGE_OFFSET = new Vector2( 0.009, 0.002 );
// Images used for animation.
  var FRONT_LEG_IMAGES = [].withAnonymousClassBody( {
    initializer: function() {
      add( new ModelElementImage( FRONT_LEG_01, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_02, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_03, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_04, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_05, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_06, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_07, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_08, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_09, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_10, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_11, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_12, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_13, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_14, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_15, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_16, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_17, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_18, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_19, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_20, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_21, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_22, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_23, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( FRONT_LEG_24, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
    }
  } );
  var BACK_LEG_IMAGES = [].withAnonymousClassBody( {
    initializer: function() {
      add( new ModelElementImage( BACK_LEG_01, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_02, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_03, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_04, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_05, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_06, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_07, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_08, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_09, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_10, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_11, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_12, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_13, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_14, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_15, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_16, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_17, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_18, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_19, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_20, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_21, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_22, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_23, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
      add( new ModelElementImage( BACK_LEG_24, FRAME_CENTER_OFFSET.plus( LEG_IMAGE_OFFSET ) ) );
    }
  } );
  var NUM_LEG_IMAGES = FRONT_LEG_IMAGES.size();
// Offsets used for creating energy chunk paths.  These need to be
// coordinated with the images.

  //private
  var BIKER_BUTTOCKS_OFFSET = new Vector2( 0.02, 0.04 );

  //private
  var TOP_TUBE_ABOVE_CRANK_OFFSET = new Vector2( 0.007, 0.015 );

  //private
  var BIKE_CRANK_OFFSET = new Vector2( 0.0052, -0.006 );
  var CENTER_OF_BACK_WHEEL_OFFSET = new Vector2( 0.03, -0.01 );

  //private
  var BOTTOM_OF_BACK_WHEEL_OFFSET = new Vector2( 0.03, -0.03 );

  //private
  var NEXT_ENERGY_SYSTEM_OFFSET = new Vector2( 0.13, -0.01 );

  function Biker( energyChunksVisible, mechanicalPoweredSystemIsNext ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // In radians.

    //private
    this.crankAngle = new Property( 0.0 );
    // In radians.

    //private
    this.rearWheelAngle = new Property( 0.0 );
    // In radians/s.

    //private
    this.crankAngularVelocity = 0;

    //private
    this.energyChunksVisible;

    //private
    this.energyChunkMovers = [];

    //private
    this.energyProducedSinceLastChunkEmitted = EFACConstants.ENERGY_PER_CHUNK * 0.9;

    //private
    this.mechanicalChunksSinceLastThermal = 0;
    this.bikerHasEnergy = new BooleanProperty( true );
    // Property through which the target pedaling rate is set.
    // In radians/sec
    this.targetCrankAngularVelocity = new Property( 0.0 );

    //private
    this.mechanicalPoweredSystemIsNext;
    EnergySource.call( this, EnergyFormsAndChangesResources.Images.BICYCLE_ICON );
    this.energyChunksVisible = energyChunksVisible;
    this.mechanicalPoweredSystemIsNext = mechanicalPoweredSystemIsNext;
    // Monitor target rotation rate for validity.
    targetCrankAngularVelocity.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( targetAngularVelocity ) {
        assert && assert( targetAngularVelocity >= 0 && targetAngularVelocity <= MAX_ANGULAR_VELOCITY_OF_CRANK );
      }
    } ) );
    // Add initial set of energy chunks.
    replenishEnergyChunks();
    // Get the crank into a position where animation will start right away.
    setCrankToPoisedPosition();
    // to the next energy system and that system is swapped out.
    mechanicalPoweredSystemIsNext.addObserver( new ChangeObserver().withAnonymousClassBody( {
      update: function( isNext, wasNext ) {
        var hubPosition = getPosition().plus( CENTER_OF_BACK_WHEEL_OFFSET );
        for ( var energyChunkMover in new ArrayList( energyChunkMovers ) ) {
          var ec = energyChunkMover.energyChunk;
          if ( ec.energyType.get() == EnergyType.MECHANICAL ) {
            if ( ec.position.get().getX() > hubPosition.getX() ) {
              // Just remove this energy chunk.
              energyChunkMovers.remove( energyChunkMover );
              energyChunkList.remove( ec );
            }
            else {
              // Make sure that this energy chunk turns to thermal energy.
              energyChunkMovers.remove( energyChunkMover );
              energyChunkMovers.add( new EnergyChunkPathMover( ec, createMechanicalToThermalEnergyChunkPath( getPosition(), ec.position.get() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
            }
          }
        }
      }
    } ) );
  }

  return inherit( EnergySource, Biker, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt ) {
        if ( isActive() ) {
          // Update energy state.
          bikerHasEnergy.set( bikerHasEnergy() );
          // the current set point.
          var target = bikerHasEnergy.get() ? targetCrankAngularVelocity.get() : 0;
          // Speed up or slow down the angular velocity of the crank.
          var previousAngularVelocity = crankAngularVelocity;
          var angularVelocityDiffFromTarget = target - crankAngularVelocity;
          if ( angularVelocityDiffFromTarget != 0 ) {
            var change = ANGULAR_ACCELERATION * dt;
            if ( angularVelocityDiffFromTarget > 0 ) {
              // Accelerate
              crankAngularVelocity = Math.min( crankAngularVelocity + change, targetCrankAngularVelocity.get() );
            }
            else {
              // Decelerate
              crankAngularVelocity = Math.max( crankAngularVelocity - change, 0 );
            }
          }
          crankAngle.set( (crankAngle.get() + crankAngularVelocity * dt) % (2 * Math.PI) );
          rearWheelAngle.set( (rearWheelAngle.get() + crankAngularVelocity * dt * CRANK_TO_REAR_WHEEL_RATIO) % (2 * Math.PI) );
          if ( crankAngularVelocity == 0 && previousAngularVelocity != 0 ) {
            // right away when motion is restarted.
            setCrankToPoisedPosition();
          }
          // Determine how much energy is produced in this time step.
          if ( targetCrankAngularVelocity.get() > 0 ) {
            // Less energy is produced if not hooked up to generator.
            var maxEnergyProductionRate = mechanicalPoweredSystemIsNext.get() ? MAX_ENERGY_OUTPUT_WHEN_CONNECTED_TO_GENERATOR : MAX_ENERGY_OUTPUT_WHEN_RUNNING_FREE;
            energyProducedSinceLastChunkEmitted += maxEnergyProductionRate * (crankAngularVelocity / MAX_ANGULAR_VELOCITY_OF_CRANK) * dt;
          }
          // Decide if new chem energy chunk should start on its way.
          if ( energyProducedSinceLastChunkEmitted >= EFACConstants.ENERGY_PER_CHUNK && targetCrankAngularVelocity.get() > 0 ) {
            // Start a new chunk moving.
            if ( bikerHasEnergy() ) {
              var energyChunk = findNonMovingEnergyChunk();
              energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createChemicalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
              energyProducedSinceLastChunkEmitted = 0;
            }
          }
          // Move the energy chunks.
          for ( var energyChunkMover in new ArrayList( energyChunkMovers ) ) {
            energyChunkMover.moveAlongPath( dt );
            if ( energyChunkMover.isPathFullyTraversed() ) {
              var energyChunk = energyChunkMover.energyChunk;
              if ( energyChunk.energyType.get() == EnergyType.CHEMICAL ) {
                // Turn this into mechanical energy.
                energyChunk.energyType.set( EnergyType.MECHANICAL );
                energyChunkMovers.remove( energyChunkMover );
                // Add new mover for the mechanical energy chunk.
                if ( mechanicalChunksSinceLastThermal >= MECHANICAL_TO_THERMAL_CHUNK_RATIO || !mechanicalPoweredSystemIsNext.get() ) {
                  // will become a chunk of thermal energy.
                  energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createMechanicalToThermalEnergyChunkPath( getPosition(), energyChunk.position.get() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
                  mechanicalChunksSinceLastThermal = 0;
                }
                else {
                  // Send this chunk to the next energy system.
                  energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createMechanicalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
                  mechanicalChunksSinceLastThermal++;
                }
              }
              else if ( energyChunk.energyType.get() == EnergyType.MECHANICAL && energyChunk.position.get().distance( getPosition().plus( CENTER_OF_BACK_WHEEL_OFFSET ) ) < 1E-6 ) {
                // to the hub and should now become thermal energy.
                energyChunkMovers.remove( energyChunkMover );
                energyChunk.energyType.set( EnergyType.THERMAL );
                energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createThermalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
              }
              else if ( energyChunk.energyType.get() == EnergyType.THERMAL ) {
                // reached the end of its route.  Delete it.
                energyChunkMovers.remove( energyChunkMover );
                energyChunkList.remove( energyChunk );
              }
              else {
                // the next energy system.
                outgoingEnergyChunks.add( energyChunk );
                energyChunkMovers.remove( energyChunkMover );
              }
            }
          }
        }
        return new Energy( EnergyType.MECHANICAL, Math.abs( crankAngularVelocity / MAX_ANGULAR_VELOCITY_OF_CRANK * MAX_ENERGY_OUTPUT_WHEN_CONNECTED_TO_GENERATOR * dt ), -Math.PI / 2 );
      },
      preLoadEnergyChunks: function() {
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.MECHANICAL, Math.abs( crankAngularVelocity / MAX_ANGULAR_VELOCITY_OF_CRANK * MAX_ENERGY_OUTPUT_WHEN_CONNECTED_TO_GENERATOR ) );
      },
      /*
       * Set the crank to a position where a very small amount of motion will
       * cause a new image to be chosen.  This is generally done when the biker
       * stops so that the animation starts right away the next time the motion
       * starts.
       */

      //private
      setCrankToPoisedPosition: function() {
        var currentImage = mapAngleToImageIndex( crankAngle.get() );
        var radiansPerImage = 2 * Math.PI / NUM_LEG_IMAGES;
        crankAngle.set( (currentImage % NUM_LEG_IMAGES * radiansPerImage + (radiansPerImage - 1E-7)) );
      },
      deactivate: function() {
        super.deactivate();
        targetCrankAngularVelocity.set( 0.0 );
        crankAngularVelocity = targetCrankAngularVelocity.get();
      },
      activate: function() {
        super.activate();
        // was a fairly arbitrary decision, and can be changed if desired.
        replenishEnergyChunks();
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        energyChunkMovers.clear();
      },
      getRearWheelAngle: function() {
        return rearWheelAngle;
      },
      getCrankAngle: function() {
        return crankAngle;
      },
      replenishEnergyChunks: function() {
        var nominalInitialOffset = new Vector2( 0.019, 0.05 );
        for ( var i = 0; i < INITIAL_NUM_ENERGY_CHUNKS; i++ ) {
          var displacement = new Vector2( (RAND.nextDouble() - 0.5) * 0.02, 0 ).getRotatedInstance( Math.PI * 0.7 );
          var newEnergyChunk = new EnergyChunk( EnergyType.CHEMICAL, getPosition().plus( nominalInitialOffset ).plus( displacement ), energyChunksVisible );
          energyChunkList.add( newEnergyChunk );
        }
      },
      mapAngleToImageIndex: function( angle ) {
        var temp = (Math.floor( (angle % (2 * Math.PI)) / (Math.PI * 2 / NUM_LEG_IMAGES) ));
        if ( temp >= NUM_LEG_IMAGES || temp < 0 ) {
          assert && assert( false );
        }
        return (Math.floor( (angle % (2 * Math.PI)) / (Math.PI * 2 / NUM_LEG_IMAGES) ));
      },

      //private
      createChemicalEnergyChunkPath: function( centerPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( centerPosition.plus( BIKER_BUTTOCKS_OFFSET ) );
            add( centerPosition.plus( TOP_TUBE_ABOVE_CRANK_OFFSET ) );
          }
        } );
      },

      //private
      createMechanicalEnergyChunkPath: function( centerPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( centerPosition.plus( BIKE_CRANK_OFFSET ) );
            add( centerPosition.plus( BOTTOM_OF_BACK_WHEEL_OFFSET ) );
            add( centerPosition.plus( NEXT_ENERGY_SYSTEM_OFFSET ) );
          }
        } );
      },
// Create a path for an energy chunk that will travel to the hub and then become thermal.

      //private
      createMechanicalToThermalEnergyChunkPath: function( centerPosition, currentPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            var crankPosition = centerPosition.plus( BIKE_CRANK_OFFSET );
            if ( currentPosition.getY() > crankPosition.getY() ) {
              // indicates that the chunk hasn't reached the crank yet.
              add( centerPosition.plus( BIKE_CRANK_OFFSET ) );
            }
            add( centerPosition.plus( CENTER_OF_BACK_WHEEL_OFFSET ) );
          }
        } );
      },

      //private
      createThermalEnergyChunkPath: function( centerPosition ) {
        var segmentLength = 0.05;
        var maxAngle = Math.PI / 8;
        var numSegments = 3;
        return [].withAnonymousClassBody( {
          initializer: function() {
            var offset = centerPosition.plus( CENTER_OF_BACK_WHEEL_OFFSET );
            add( new Vector2( offset ) );
            // The chuck needs to move up and to the right to avoid overlapping with the biker.
            offset = offset.plus( new Vector2( segmentLength, 0 ).getRotatedInstance( Math.PI * 0.4 ) );
            // Add a set of path segments that make the chunk move up in a somewhat random path.
            add( new Vector2( offset ) );
            for ( var i = 0; i < numSegments; i++ ) {
              offset = offset.plus( new Vector2( 0, segmentLength ).getRotatedInstance( (RAND.nextDouble() - 0.5) * maxAngle ) );
              add( new Vector2( offset ) );
            }
          }
        } );
      },
// Choose a non-moving energy chunk, returns null if all chunks are moving.

      //private
      findNonMovingEnergyChunk: function() {
        var movingEnergyChunks = [];
        for ( var energyChunkMover in energyChunkMovers ) {
          movingEnergyChunks.add( energyChunkMover.energyChunk );
        }
        var nonMovingEnergyChunk = null;
        for ( var energyChunk in energyChunkList ) {
          if ( !movingEnergyChunks.contains( energyChunk ) ) {
            nonMovingEnergyChunk = energyChunk;
            break;
          }
        }
        return nonMovingEnergyChunk;
      },

      //private
      bikerHasEnergy: function() {
        return energyChunkList.size() > 0 && energyChunkList.size() > energyChunkMovers.size();
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectWaterPoweredGeneratorButton;
      }
    },
//statics
    {
      MAX_ANGULAR_VELOCITY_OF_CRANK: MAX_ANGULAR_VELOCITY_OF_CRANK,
      REAR_WHEEL_RADIUS: REAR_WHEEL_RADIUS,
      FRAME_IMAGE: FRAME_IMAGE,
      REAR_WHEEL_SPOKES_IMAGE: REAR_WHEEL_SPOKES_IMAGE,
      RIDER_NORMAL_UPPER_BODY_IMAGE: RIDER_NORMAL_UPPER_BODY_IMAGE,
      RIDER_TIRED_UPPER_BODY_IMAGE: RIDER_TIRED_UPPER_BODY_IMAGE,
      FRONT_LEG_IMAGES: FRONT_LEG_IMAGES,
      BACK_LEG_IMAGES: BACK_LEG_IMAGES,
      NUM_LEG_IMAGES: NUM_LEG_IMAGES,
      CENTER_OF_BACK_WHEEL_OFFSET: CENTER_OF_BACK_WHEEL_OFFSET
    } );
} );

