// Copyright 2002-2012, University of Colorado
/**
 * This class implements a container of sorts for positionable model elements.
 * The model elements are positioned by this class, and an API is provided
 * that allows clients to move elements to the "selected" position.  Changes
 * to the selected element are animated.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------

  //private
  var TRANSITION_DURATION = 0.5;

  function Carousel( selectedElementPosition, offsetBetweenElements ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // The position in model space where the currently selected element should be.

    //private
    this.selectedElementPosition;
    // Offset between elements in the carousel.

    //private
    this.offsetBetweenElements;
    // List of the elements whose position is managed by this carousel.

    //private
    this.managedElements = [];
    // Target selected element.  Will be the same as the current selection
    // except when animating to a new selection.  This property is the API for
    // selecting elements in the carousel.
    this.targetIndex = new Property( 0 );
    // Indicator for when animations are in progress, meant to be monitored by
    // clients that need to be aware of this.

    //private
    this.animationInProgress = new BooleanProperty( false );

    //private
    this.elapsedTransitionTime = 0;

    //private
    this.currentCarouselOffset = new Vector2( 0, 0 );

    //private
    this.carouselOffsetWhenTransitionStarted = new Vector2( 0, 0 );
    this.selectedElementPosition = selectedElementPosition;
    this.offsetBetweenElements = offsetBetweenElements;
    // animation each time the target changes.
    targetIndex.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( targetIndexValue ) {
        // Bounds checking.
        assert && assert( targetIndexValue == 0 || targetIndexValue < managedElements.size() );
        elapsedTransitionTime = 0;
        carouselOffsetWhenTransitionStarted = currentCarouselOffset;
        animationInProgress.set( true );
      }
    } ) );
  }

  return inherit( Object, Carousel, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    add: function( element ) {
      // Set the element's position to be at the end of the carousel.
      if ( managedElements.size() == 0 ) {
        element.setPosition( selectedElementPosition );
      }
      else {
        element.setPosition( managedElements.get( managedElements.size() - 1 ).getPosition().plus( offsetBetweenElements ) );
      }
      // Add element to the list of managed elements.
      managedElements.add( element );
      // Update opacities.
      updateManagedElementOpacities();
    },
    getNumElements: function() {
      return managedElements.size();
    },
    getElement: function( index ) {
      if ( index <= managedElements.size() ) {
        return managedElements.get( index );
      }
      else {
        console.log( getClass().getName() + " - Warning: Request of out of range element from carousel, index = " + index );
        return null;
      }
    },
    getElementList: function() {
      return new ArrayList( managedElements );
    },
    getSelectedElementPosition: function() {
      return selectedElementPosition;
    },
    getSelectedElement: function() {
      if ( targetIndex.get() < managedElements.size() ) {
        return managedElements.get( targetIndex.get() );
      }
      return null;
    },
    /*
     * Perform any animation changes needed.
     */
    stepInTime: function( dt ) {
      if ( !atTargetPosition() ) {
        elapsedTransitionTime += dt;
        var targetCarouselOffset = offsetBetweenElements.times( -targetIndex.get() );
        var totalTravelVector = targetCarouselOffset.minus( carouselOffsetWhenTransitionStarted );
        var proportionOfTimeElapsed = MathUtil.clamp( 0, elapsedTransitionTime / TRANSITION_DURATION, 1 );
        currentCarouselOffset = carouselOffsetWhenTransitionStarted.plus( totalTravelVector.times( computeSlowInSlowOut( proportionOfTimeElapsed ) ) );
        updateManagedElementPositions();
        if ( proportionOfTimeElapsed == 1 ) {
          currentCarouselOffset = targetCarouselOffset;
        }
        if ( currentCarouselOffset == targetCarouselOffset ) {
          animationInProgress.set( false );
        }
        updateManagedElementOpacities();
      }
    },

    //private
    updateManagedElementPositions: function() {
      for ( var i = 0; i < managedElements.size(); i++ ) {
        managedElements.get( i ).setPosition( selectedElementPosition.plus( offsetBetweenElements.times( i ) ).plus( currentCarouselOffset ) );
      }
    },

    //private
    updateManagedElementOpacities: function() {
      for ( var managedElement in managedElements ) {
        var distanceFromSelectionPosition = managedElement.getPosition().distance( selectedElementPosition );
        var opacity = MathUtil.clamp( 0, 1 - (distanceFromSelectionPosition / offsetBetweenElements.magnitude()), 1 );
        managedElement.opacity.set( opacity );
      }
    },

    //private
    atTargetPosition: function() {
      var targetCarouselOffset = new Vector2( offsetBetweenElements.times( -targetIndex.get() ) );
      return currentCarouselOffset.equals( targetCarouselOffset );
    },
    getAnimationInProgressProperty: function() {
      return animationInProgress;
    },
    /*
     * Maps a linear value to a value that can be used for animated motion that
     * accelerates, then slows as destination is reached.  This was leveraged
     * from Piccolo's PInterpolatingActivity class.
     */

    //private
    computeSlowInSlowOut: function( zeroToOne ) {
      if ( zeroToOne < 0.5{
        f
      }
      )
      {
        return 2.0
        f * zeroToOne * zeroToOne;
      }
      else
      {
        var complement = 1.0
        f - zeroToOne;
        return 1.0
        f - 2.0
        f * complement * complement;
      }
    }
  } );
} );

