// Copyright 2002-2012, University of Colorado
/**
 * This class represents an electrical generator that is powered by flowing
 * water, steam, or directly driven using a drive belt.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var EnergyChunkNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/EnergyChunkNode' );
  var CONNECTOR = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/CONNECTOR' );//static
  var GENERATOR = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/GENERATOR' );//static
  var GENERATOR_WHEEL_HUB_2 = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/GENERATOR_WHEEL_HUB_2' );//static
  var GENERATOR_WHEEL_PADDLES_SHORT = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/GENERATOR_WHEEL_PADDLES_SHORT' );//static
  var GENERATOR_WHEEL_SPOKES = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/GENERATOR_WHEEL_SPOKES' );//static
  var WIRE_BLACK_LEFT = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/WIRE_BLACK_LEFT' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Attributes of the wheel and generator.
// In kg.

  //private
  var WHEEL_MOMENT_OF_INERTIA = 5;
// Controls max speed and rate of slow down, empirically determined.

  //private
  var RESISTANCE_CONSTANT = 3;
// In radians/sec, empirically determined.

  //private
  var MAX_ROTATIONAL_VELOCITY = Math.PI / 2;
// Images used to represent this model element in the view.
  var HOUSING_IMAGE = new ModelElementImage( GENERATOR, new Vector2( 0, 0 ) );
  var WHEEL_CENTER_OFFSET = new Vector2( 0, 0.03 );

  //private
  var LEFT_SIDE_OF_WHEEL_OFFSET = new Vector2( -0.030, 0.03 );
  var WHEEL_PADDLES_IMAGE = new ModelElementImage( GENERATOR_WHEEL_PADDLES_SHORT, WHEEL_CENTER_OFFSET );
  var WHEEL_HUB_IMAGE = new ModelElementImage( GENERATOR_WHEEL_HUB_2, WHEEL_CENTER_OFFSET );
  var SHORT_SPOKES_IMAGE = new ModelElementImage( GENERATOR_WHEEL_SPOKES, WHEEL_CENTER_OFFSET );

  //private
  var CONNECTOR_OFFSET = new Vector2( 0.057, -0.04 );
// Offset empirically determined for optimal look.
  var CONNECTOR_IMAGE = new ModelElementImage( CONNECTOR, CONNECTOR_OFFSET );
// Offset empirically determined for optimal look.
  var WIRE_CURVED_IMAGE = new ModelElementImage( WIRE_BLACK_LEFT, new Vector2( 0.0185, -0.015 ) );
  var WHEEL_RADIUS = WHEEL_HUB_IMAGE.getWidth() / 2;
// Offsets used to create the paths followed by the energy chunks.

  //private
  var START_OF_WIRE_CURVE_OFFSET = WHEEL_CENTER_OFFSET.plus( 0.01, -0.05 );

  //private
  var WIRE_CURVE_POINT_1_OFFSET = WHEEL_CENTER_OFFSET.plus( 0.015, -0.06 );

  //private
  var WIRE_CURVE_POINT_2_OFFSET = WHEEL_CENTER_OFFSET.plus( 0.03, -0.07 );

  //private
  var CENTER_OF_CONNECTOR_OFFSET = CONNECTOR_OFFSET;

  function ElectricalGenerator( energyChunkVisibilityControl ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // In radians.

    //private
    this.wheelRotationalAngle = new Property( 0.0 );
    // In radians/s.

    //private
    this.wheelRotationalVelocity = 0;

    //private
    this.energyChunkMovers = [];

    //private
    this.energyChunkVisibilityControl;
    // Flag that controls "direct coupling mode", which means that the
    // generator wheel turns at a rate that is directly proportionate to the
    // incoming energy, with no rotational inertia.
    this.directCouplingMode = new BooleanProperty( false );
    // The electrical energy chunks are kept on a separate list to support
    // placing them on a different layer in the view.
    this.electricalEnergyChunks = new ObservableArray();
    // The "hidden" energy chunks are kept on a separate list mainly for
    // code clarity.
    this.hiddenEnergyChunks = new ObservableArray();
    EnergyConverter.call( this, EnergyFormsAndChangesResources.Images.GENERATOR_ICON );
    this.energyChunkVisibilityControl = energyChunkVisibilityControl;
  }

  return inherit( EnergyConverter, ElectricalGenerator, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt, incomingEnergy ) {
        if ( isActive() ) {
          // Handle different wheel rotation modes.
          if ( directCouplingMode.get() ) {
            // energy source, e.g. through a belt or drive shaft.
            if ( incomingEnergy.type == EnergyType.MECHANICAL ) {
              // Convention is positive is counter clockwise.
              wheelRotationalVelocity = (incomingEnergy.amount / dt) / EFACConstants.MAX_ENERGY_PRODUCTION_RATE * MAX_ROTATIONAL_VELOCITY * (Math.sin( incomingEnergy.direction ) > 0 ? -1 : 1);
              wheelRotationalAngle.set( wheelRotationalAngle.get() + wheelRotationalVelocity * dt );
            }
          }
          else {
            // energy, such as water, and has inertia.
            var torqueFromIncomingEnergy = 0;
            if ( incomingEnergy.type == EnergyType.MECHANICAL ) {
              // Empirically determined to reach max energy after a second or two.
              var energyToTorqueConstant = 0.5;
              // Convention is positive is counter clockwise.
              torqueFromIncomingEnergy = incomingEnergy.amount * WHEEL_RADIUS * energyToTorqueConstant * (Math.sin( incomingEnergy.direction ) > 0 ? -1 : 1);
            }
            var torqueFromResistance = -wheelRotationalVelocity * RESISTANCE_CONSTANT;
            var angularAcceleration = (torqueFromIncomingEnergy + torqueFromResistance) / WHEEL_MOMENT_OF_INERTIA;
            wheelRotationalVelocity = MathUtil.clamp( -MAX_ROTATIONAL_VELOCITY, wheelRotationalVelocity + (angularAcceleration * dt), MAX_ROTATIONAL_VELOCITY );
            if ( Math.abs( wheelRotationalVelocity ) < 1E-3 ) {
              // Prevent the wheel from moving forever.
              wheelRotationalVelocity = 0;
            }
            wheelRotationalAngle.set( wheelRotationalAngle.get() + wheelRotationalVelocity * dt );
          }
          // Handle any incoming energy chunks.
          if ( !incomingEnergyChunks.isEmpty() ) {
            for ( var incomingEnergyChunk in new ArrayList( incomingEnergyChunks ) ) {
              if ( incomingEnergyChunk.energyType.get() == EnergyType.MECHANICAL ) {
                energyChunkList.add( incomingEnergyChunk );
                incomingEnergyChunks.remove( incomingEnergyChunk );
                // the center of the wheel.
                energyChunkMovers.add( new EnergyChunkPathMover( incomingEnergyChunk, createMechanicalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
              }
              else {
                // By design, this shouldn't happen, so warn if it does.
                console.log( getClass().getName() + " - Warning: Ignoring energy chunk with unexpected type, type = " + incomingEnergyChunk.energyType.get().toString() );
              }
            }
            incomingEnergyChunks.clear();
          }
          // Move the energy chunks and update their state.
          updateEnergyChunkPositions( dt );
        }
        // Produce the appropriate amount of energy.
        return new Energy( EnergyType.ELECTRICAL, Math.abs( (wheelRotationalVelocity / MAX_ROTATIONAL_VELOCITY) * EFACConstants.MAX_ENERGY_PRODUCTION_RATE ) * dt );
      },

      //private
      updateEnergyChunkPositions: function( dt ) {
        for ( var energyChunkMover in new ArrayList( energyChunkMovers ) ) {
          energyChunkMover.moveAlongPath( dt );
          if ( energyChunkMover.isPathFullyTraversed() ) {
            var ec = energyChunkMover.energyChunk;
            switch( ec.energyType.get() ) {
              case MECHANICAL:
                // be seen by the user.
                energyChunkList.remove( ec );
                energyChunkMovers.remove( energyChunkMover );
                ec.energyType.set( EnergyType.ELECTRICAL );
                electricalEnergyChunks.add( ec );
                energyChunkMovers.add( new EnergyChunkPathMover( energyChunkMover.energyChunk, createElectricalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
                var hiddenEnergyChunk = new EnergyChunk( EnergyType.HIDDEN, ec.position.get(), energyChunkVisibilityControl );
                hiddenEnergyChunk.zPosition.set( -EnergyChunkNode.Z_DISTANCE_WHERE_FULLY_FADED / 2 );
                hiddenEnergyChunks.add( hiddenEnergyChunk );
                energyChunkMovers.add( new EnergyChunkPathMover( hiddenEnergyChunk, createHiddenEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
                break;
              case ELECTRICAL:
                // energy system.
                energyChunkMovers.remove( energyChunkMover );
                outgoingEnergyChunks.add( ec );
                break;
              case HIDDEN:
                // should now be visible to the user.
                hiddenEnergyChunks.remove( energyChunkMover.energyChunk );
                energyChunkMovers.remove( energyChunkMover );
                break;
            }
          }
        }
      },
      preLoadEnergyChunks: function( incomingEnergyRate ) {
        clearEnergyChunks();
        if ( incomingEnergyRate.amount == 0 || incomingEnergyRate.type != EnergyType.MECHANICAL ) {
          // No energy chunk pre-loading needed.
          return;
        }
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        var energySinceLastChunk = EFACConstants.ENERGY_PER_CHUNK * 0.99;
        // Simulate energy chunks moving through the system.
        var preLoadComplete = false;
        while ( !preLoadComplete ) {
          energySinceLastChunk += incomingEnergyRate.amount * dt;
          // Determine if time to add a new chunk.
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            var newEnergyChunk = new EnergyChunk( EnergyType.MECHANICAL, getPosition().plus( LEFT_SIDE_OF_WHEEL_OFFSET ), energyChunkVisibilityControl );
            energyChunkList.add( newEnergyChunk );
            // Add a 'mover' for this energy chunk.
            energyChunkMovers.add( new EnergyChunkPathMover( newEnergyChunk, createMechanicalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
            // Update energy since last chunk.
            energySinceLastChunk = energySinceLastChunk - EFACConstants.ENERGY_PER_CHUNK;
          }
          updateEnergyChunkPositions( dt );
          if ( outgoingEnergyChunks.size() > 0 ) {
            // An energy chunks has made it all the way through the system.
            preLoadComplete = true;
          }
        }
      },

      //private
      createMechanicalEnergyChunkPath: function( panelPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( panelPosition.plus( WHEEL_CENTER_OFFSET ) );
          }
        } );
      },

      //private
      createElectricalEnergyChunkPath: function( panelPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( panelPosition.plus( START_OF_WIRE_CURVE_OFFSET ) );
            add( panelPosition.plus( WIRE_CURVE_POINT_1_OFFSET ) );
            add( panelPosition.plus( WIRE_CURVE_POINT_2_OFFSET ) );
            add( panelPosition.plus( CENTER_OF_CONNECTOR_OFFSET ) );
          }
        } );
      },

      //private
      createHiddenEnergyChunkPath: function( panelPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            // Overlaps with the electrical chunks until it reaches the window, then is done.
            add( panelPosition.plus( START_OF_WIRE_CURVE_OFFSET ) );
            add( panelPosition.plus( WIRE_CURVE_POINT_1_OFFSET ) );
          }
        } );
      },
      deactivate: function() {
        super.deactivate();
        wheelRotationalVelocity = 0;
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        electricalEnergyChunks.clear();
        hiddenEnergyChunks.clear();
        energyChunkMovers.clear();
      },
      getWheelRotationalAngle: function() {
        return wheelRotationalAngle;
      },
// Have to override, since outgoing chunks are on a separate list.
      extractOutgoingEnergyChunks: function() {
        var retVal = new ArrayList( outgoingEnergyChunks );
        electricalEnergyChunks.removeAll( outgoingEnergyChunks );
        outgoingEnergyChunks.clear();
        return retVal;
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.ELECTRICAL, Math.abs( wheelRotationalVelocity / MAX_ROTATIONAL_VELOCITY * EFACConstants.MAX_ENERGY_PRODUCTION_RATE ) );
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectWaterPoweredGeneratorButton;
      }
    },
//statics
    {
      HOUSING_IMAGE: HOUSING_IMAGE,
      WHEEL_CENTER_OFFSET: WHEEL_CENTER_OFFSET,
      WHEEL_PADDLES_IMAGE: WHEEL_PADDLES_IMAGE,
      WHEEL_HUB_IMAGE: WHEEL_HUB_IMAGE,
      SHORT_SPOKES_IMAGE: SHORT_SPOKES_IMAGE,
      CONNECTOR_IMAGE: CONNECTOR_IMAGE,
      WIRE_CURVED_IMAGE: WIRE_CURVED_IMAGE,
      WHEEL_RADIUS: WHEEL_RADIUS
    } );
} );

