// Copyright 2002-2012, University of Colorado
/**
 * Primary model class for the "Energy Systems" tab of the Energy Forms and
 * Changes simulation.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Vector2 = require( 'DOT/Vector2' );
  var Resettable = require( 'edu.colorado.phet.common.phetcommon.model.Resettable' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------

  //private
  var OFFSET_BETWEEN_ELEMENTS_ON_CAROUSEL = new Vector2( 0, -0.4 );

  function EnergySystemsModel() {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Main model clock.

    //private
    this.clock = new ConstantDtClock( 30.0 );
    // Boolean property that controls whether the energy chunks are visible to the user.
    this.energyChunksVisible = new BooleanProperty( false );
    // Carousels that control the positions of the energy sources, converters,
    // and users.
    this.energySourcesCarousel = new EnergySystemElementCarousel( new Vector2( -0.15, 0 ), OFFSET_BETWEEN_ELEMENTS_ON_CAROUSEL );
    this.energyConvertersCarousel = new EnergySystemElementCarousel( new Vector2( -0.025, 0 ), OFFSET_BETWEEN_ELEMENTS_ON_CAROUSEL );
    this.energyUsersCarousel = new EnergySystemElementCarousel( new Vector2( 0.09, 0 ), OFFSET_BETWEEN_ELEMENTS_ON_CAROUSEL );

    //private
    this.carousels = [].withAnonymousClassBody( {
      initializer: function() {
        add( energySourcesCarousel );
        add( energyConvertersCarousel );
        add( energyUsersCarousel );
      }
    } );
    // Energy Converters.
    this.waterPoweredGenerator = new ElectricalGenerator( energyChunksVisible );
    this.solarPanel = new SolarPanel( getClock(), energyChunksVisible );
    // Energy sources.
    this.faucet = new FaucetAndWater( energyChunksVisible, waterPoweredGenerator.getObservableActiveState() );
    this.sun = new Sun( solarPanel, energyChunksVisible );
    this.teaPot = new TeaPot( energyChunksVisible, waterPoweredGenerator.getObservableActiveState() );
    this.biker = new Biker( energyChunksVisible, waterPoweredGenerator.getObservableActiveState() );
    // Energy users.
    this.incandescentLightBulb = new IncandescentLightBulb( energyChunksVisible );
    this.fluorescentLightBulb = new FluorescentLightBulb( energyChunksVisible );
    this.beakerHeater = new BeakerHeater( clock, energyChunksVisible );
    // Items that span between energy system elements.
    this.belt;
    clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        stepInTime( clockEvent.getSimulationTimeChange() );
      }
    } ) );
    energySourcesCarousel.add( faucet );
    energySourcesCarousel.add( sun );
    energySourcesCarousel.add( teaPot );
    energySourcesCarousel.add( biker );
    energyConvertersCarousel.add( waterPoweredGenerator );
    energyConvertersCarousel.add( solarPanel );
    energyUsersCarousel.add( beakerHeater );
    energyUsersCarousel.add( incandescentLightBulb );
    energyUsersCarousel.add( fluorescentLightBulb );
    // in the right place.  Not entirely sure why.
    belt = new Belt( Biker.REAR_WHEEL_RADIUS, energySourcesCarousel.getSelectedElementPosition().plus( Biker.CENTER_OF_BACK_WHEEL_OFFSET ).plus( 0.005, 0 ), ElectricalGenerator.WHEEL_RADIUS, energyConvertersCarousel.getSelectedElementPosition().plus( ElectricalGenerator.WHEEL_CENTER_OFFSET ) );
    // biker and the generator.
    var beltVisibilityUpdated = new VoidFunction1().withAnonymousClassBody( {
      apply: function( isAnimating ) {
        var bikerAndGeneratorSelected = !isAnimating && biker.isActive() && waterPoweredGenerator.isActive();
        belt.isVisible.set( bikerAndGeneratorSelected );
        waterPoweredGenerator.directCouplingMode.set( bikerAndGeneratorSelected );
      }
    } );
    energySourcesCarousel.getAnimationInProgressProperty().addObserver( beltVisibilityUpdated );
    energyConvertersCarousel.getAnimationInProgressProperty().addObserver( beltVisibilityUpdated );
    // chunks continue to move through the user.
    energyConvertersCarousel.getAnimationInProgressProperty().addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( animating ) {
        energyUsersCarousel.getSelectedElement().clearEnergyChunks();
      }
    } ) );
    // can still operate when the sim is paused.
    var carouselClock = new ConstantDtClock( EFACConstants.FRAMES_PER_SECOND );
    carouselClock.start();
    carouselClock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        // Step the animation for the carousels.
        for ( var carousel in carousels ) {
          carousel.stepInTime( clockEvent.getSimulationTimeChange() );
        }
      }
    } ) );
  }

  return inherit( Object, EnergySystemsModel, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    reset: function() {
      clock.resetSimulationTime();
      energyChunksVisible.reset();
      energySourcesCarousel.getSelectedElement().deactivate();
      energyConvertersCarousel.getSelectedElement().deactivate();
      energyUsersCarousel.getSelectedElement().deactivate();
      energySourcesCarousel.targetIndex.set( 0 );
      energyConvertersCarousel.targetIndex.set( 0 );
      energyUsersCarousel.targetIndex.set( 0 );
      energySourcesCarousel.getSelectedElement().activate();
      energyConvertersCarousel.getSelectedElement().activate();
      energyUsersCarousel.getSelectedElement().activate();
    },
    getClock: function() {
      return clock;
    },

    //private
    stepInTime: function( dt ) {
      // Step the active elements in time to produce, convert, and use energy.
      var energyFromSource = energySourcesCarousel.getSelectedElement().stepInTime( dt );
      var energyFromConverter = energyConvertersCarousel.getSelectedElement().stepInTime( dt, energyFromSource );
      energyUsersCarousel.getSelectedElement().stepInTime( dt, energyFromConverter );
      // Transfer energy chunks between the elements.
      energyConvertersCarousel.getSelectedElement().injectEnergyChunks( energySourcesCarousel.getSelectedElement().extractOutgoingEnergyChunks() );
      energyUsersCarousel.getSelectedElement().injectEnergyChunks( energyConvertersCarousel.getSelectedElement().extractOutgoingEnergyChunks() );
    },
    /**
     * Pre-load the currently active energy system elements with energy
     * chunks so that the energy chunks are fully propagated into the
     * elements.
     */
    preLoadEnergyChunks: function() {
      energySourcesCarousel.getSelectedElement().preLoadEnergyChunks();
      energyConvertersCarousel.getSelectedElement().preLoadEnergyChunks( energySourcesCarousel.getSelectedElement().getEnergyOutputRate() );
      energyUsersCarousel.getSelectedElement().preLoadEnergyChunks( energyConvertersCarousel.getSelectedElement().getEnergyOutputRate() );
    }
  } );
} );

