// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a faucet that can be turned on to provide mechanical
 * energy to other energy system elements.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var PDimension = require( 'edu.umd.cs.piccolo.util.PDimension' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
  var OFFSET_FROM_CENTER_TO_WATER_ORIGIN = new Vector2( 0.065, 0.08 );
// In meters/second.

  //private
  var FALLING_ENERGY_CHUNK_VELOCITY = 0.09;
// In meters.

  //private
  var MAX_WATER_WIDTH = 0.015;
// In meters.

  //private
  var MAX_DISTANCE_FROM_FAUCET_TO_BOTTOM_OF_WATER = 0.5;

  //private
  var RAND = new Random();

  //private
  var ENERGY_CHUNK_TRANSFER_DISTANCE_RANGE = new DoubleRange( 0.05, 0.06 );
// The following acceleration constant defines the rate at which the water
// flows from the faucet.  The value used is not the actual value in
// Earth's gravitational field - it has been tweaked for optimal visual
// effect.

  //private
  var ACCELERATION_DUE_TO_GRAVITY = new Vector2( 0, -0.15 );

  function FaucetAndWater( energyChunksVisible, waterPowerableElementInPlace ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Proportion of full available flow that is occurring.
    this.flowProportion = new Property( 0.0 );
    // Water drops that comprise the stream of water.
    this.waterDrops = new ObservableArray();
    // List of chunks that are not being transferred to the next energy system
    // element.

    //private
    this.exemptFromTransferEnergyChunks = [];

    //private
    this.energySinceLastChunk = 0;

    //private
    this.energyChunksVisible;
    // Flag that is used to decide whether to pass energy chunks to the next
    // energy system element.

    //private
    this.waterPowerableElementInPlace;
    // Flag for whether next chunk should be transferred or kept, used to
    // alternate transfer with non-transfer.

    //private
    this.transferNextAvailableChunk = true;
    EnergySource.call( this, EnergyFormsAndChangesResources.Images.FAUCET_ICON );
    this.energyChunksVisible = energyChunksVisible;
    this.waterPowerableElementInPlace = waterPowerableElementInPlace;
  }

  return inherit( EnergySource, FaucetAndWater, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt ) {
        if ( isActive() ) {
          // Add water droplets as needed based on flow rate.
          if ( flowProportion.get() > 0 ) {
            var initialWidth = flowProportion.get() * MAX_WATER_WIDTH * (1 + (RAND.nextDouble() - 0.5) * 0.2);
            waterDrops.add( new WaterDrop( OFFSET_FROM_CENTER_TO_WATER_ORIGIN.plus( 0, 0.01 ), new Vector2( 0, 0 ), new PDimension( initialWidth, initialWidth ) ) );
          }
          // Make the water droplets fall.
          for ( var waterDrop in new ArrayList( waterDrops ) ) {
            waterDrop.velocity.set( waterDrop.velocity.get().plus( ACCELERATION_DUE_TO_GRAVITY.times( dt ) ) );
            waterDrop.offsetFromParent.set( waterDrop.offsetFromParent.get().plus( waterDrop.velocity.get().times( dt ) ) );
            if ( waterDrop.offsetFromParent.get().distance( getPosition() ) > MAX_DISTANCE_FROM_FAUCET_TO_BOTTOM_OF_WATER ) {
              waterDrops.remove( waterDrop );
            }
          }
          // Check if time to emit an energy chunk and, if so, do it.
          energySinceLastChunk += EFACConstants.MAX_ENERGY_PRODUCTION_RATE * flowProportion.get() * dt;
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            energyChunkList.add( createNewChunk() );
            energySinceLastChunk = energySinceLastChunk - EFACConstants.ENERGY_PER_CHUNK;
          }
          // Update energy chunk positions.
          for ( var energyChunk in new ArrayList( energyChunkList ) ) {
            // Make the chunk fall.
            energyChunk.translateBasedOnVelocity( dt );
            // to the next energy system.
            if ( waterPowerableElementInPlace.get() && ENERGY_CHUNK_TRANSFER_DISTANCE_RANGE.contains( getPosition().plus( OFFSET_FROM_CENTER_TO_WATER_ORIGIN ).getY() - energyChunk.position.get().y ) && !exemptFromTransferEnergyChunks.contains( energyChunk ) ) {
              if ( transferNextAvailableChunk ) {
                // Send this chunk to the next energy system.
                outgoingEnergyChunks.add( energyChunk );
                // Alternate sending or keeping chunks.
                transferNextAvailableChunk = false;
              }
              else {
                // Don't transfer this chunk.
                exemptFromTransferEnergyChunks.add( energyChunk );
                // Set up to transfer the next one.
                transferNextAvailableChunk = true;
              }
            }
            // Remove it if it is out of visible range.
            if ( getPosition().plus( OFFSET_FROM_CENTER_TO_WATER_ORIGIN ).distance( energyChunk.position.get() ) > MAX_DISTANCE_FROM_FAUCET_TO_BOTTOM_OF_WATER ) {
              energyChunkList.remove( energyChunk );
              exemptFromTransferEnergyChunks.remove( energyChunk );
            }
          }
        }
        // Generate the appropriate amount of energy.
        return new Energy( EnergyType.MECHANICAL, EFACConstants.MAX_ENERGY_PRODUCTION_RATE * flowProportion.get() * dt, -Math.PI / 2 );
      },
      preLoadEnergyChunks: function() {
        clearEnergyChunks();
        // In seconds, empirically determined.
        var preLoadTime = 3;
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        var tempEnergyChunkList = [];
        // Simulate energy chunks moving through the system.
        while ( preLoadTime > 0 ) {
          energySinceLastChunk += getEnergyOutputRate().amount * dt;
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            tempEnergyChunkList.add( createNewChunk() );
            energySinceLastChunk = energySinceLastChunk - EFACConstants.ENERGY_PER_CHUNK;
          }
          for ( var energyChunk in tempEnergyChunkList ) {
            // Make the chunk fall.
            energyChunk.translateBasedOnVelocity( dt );
          }
          preLoadTime -= dt;
        }
        // Now that they are positioned, add these to the 'real' list of energy chunks.
        energyChunkList.addAll( tempEnergyChunkList );
      },

      //private
      createNewChunk: function() {
        var initialPosition = getPosition().plus( OFFSET_FROM_CENTER_TO_WATER_ORIGIN ).plus( (RAND.nextDouble() - 0.5) * flowProportion.get() * MAX_WATER_WIDTH / 2, 0 );
        return new EnergyChunk( EnergyType.MECHANICAL, initialPosition, new Vector2( 0, -FALLING_ENERGY_CHUNK_VELOCITY ), energyChunksVisible );
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.MECHANICAL, EFACConstants.MAX_ENERGY_PRODUCTION_RATE * flowProportion.get() );
      },
      deactivate: function() {
        super.deactivate();
        waterDrops.clear();
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        exemptFromTransferEnergyChunks.clear();
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectFaucetButton;
      }
    },
//statics
    {
      OFFSET_FROM_CENTER_TO_WATER_ORIGIN: OFFSET_FROM_CENTER_TO_WATER_ORIGIN
    } );
} );

