// Copyright 2002-2012, University of Colorado
/**
 * Base class for light bulbs in the model.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var ELEMENT_BASE_BACK = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/ELEMENT_BASE_BACK' );//static
  var ELEMENT_BASE_FRONT = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/ELEMENT_BASE_FRONT' );//static
  var WIRE_BLACK_62 = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/WIRE_BLACK_62' );//static
  var WIRE_BLACK_RIGHT = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/WIRE_BLACK_RIGHT' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Images used to represent this element in the view.
  var WIRE_FLAT_IMAGE = new ModelElementImage( WIRE_BLACK_62, new Vector2( -0.036, -0.04 ) );
  var WIRE_CURVE_IMAGE = new ModelElementImage( WIRE_BLACK_RIGHT, new Vector2( -0.009, -0.016 ) );
  var ELEMENT_BASE_FRONT_IMAGE = new ModelElementImage( ELEMENT_BASE_FRONT, new Vector2( 0, 0.0 ) );
  var ELEMENT_BASE_BACK_IMAGE = new ModelElementImage( ELEMENT_BASE_BACK, new Vector2( 0, 0.0 ) );
// Offsets need for creating the path followed by the energy chunks.  These
// were empirically determined based on images, will need to change if the
// images are changed.

  //private
  var OFFSET_TO_LEFT_SIDE_OF_WIRE = new Vector2( -0.04, -0.04 );

  //private
  var OFFSET_TO_LEFT_SIDE_OF_WIRE_BEND = new Vector2( -0.02, -0.04 );

  //private
  var OFFSET_TO_FIRST_WIRE_CURVE_POINT = new Vector2( -0.01, -0.0375 );

  //private
  var OFFSET_TO_SECOND_WIRE_CURVE_POINT = new Vector2( -0.001, -0.025 );

  //private
  var OFFSET_TO_THIRD_WIRE_CURVE_POINT = new Vector2( -0.0005, -0.0175 );

  //private
  var OFFSET_TO_BOTTOM_OF_CONNECTOR = new Vector2( 0, -0.01 );

  //private
  var OFFSET_TO_RADIATE_POINT = new Vector2( 0, 0.066 );
// Miscellaneous other constants.

  //private
  var RADIATED_ENERGY_CHUNK_MAX_DISTANCE = 0.5;

  //private
  var RAND = new Random();

  //private
  var THERMAL_ENERGY_CHUNK_TIME_ON_FILAMENT = new DoubleRange( 2, 2.5 );

  //private
  var ENERGY_TO_FULLY_LIGHT = EFACConstants.MAX_ENERGY_PRODUCTION_RATE;
// In meters.

  //private
  var LIGHT_CHUNK_LIT_BULB_RADIUS = 0.1;
// In proportion per second.

  //private
  var LIGHT_CHANGE_RATE = 0.5;

  function LightBulb( userComponent, icon, hasFilament, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.userComponent;
    this.litProportion = new Property( 0.0 );

    //private
    this.electricalEnergyChunkMovers = [];

    //private
    this.filamentEnergyChunkMovers = [];

    //private
    this.radiatedEnergyChunkMovers = [];

    //private
    this.hasFilament;

    //private
    this.energyChunksVisible;

    //private
    this.proportionOfThermalChunksRadiated;

    //private
    this.goRightNextTime = true;
    EnergyUser.call( this, icon );
    this.userComponent = userComponent;
    this.hasFilament = hasFilament;
    this.energyChunksVisible = energyChunksVisible;
    if ( hasFilament ) {
      proportionOfThermalChunksRadiated = 0.35;
    }
    else {
      // Fewer thermal energy chunks are radiated for bulbs without a filament.
      proportionOfThermalChunksRadiated = 0.2;
    }
  }

  return inherit( EnergyUser, LightBulb, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt, incomingEnergy ) {
        if ( isActive() ) {
          // Handle any incoming energy chunks.
          if ( !incomingEnergyChunks.isEmpty() ) {
            for ( var incomingEnergyChunk in incomingEnergyChunks ) {
              if ( incomingEnergyChunk.energyType.get() == EnergyType.ELECTRICAL ) {
                // Add the energy chunk to the list of those under management.
                energyChunkList.add( incomingEnergyChunk );
                // the wire to the bulb.
                electricalEnergyChunkMovers.add( new EnergyChunkPathMover( incomingEnergyChunk, createElectricalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
              }
              else {
                // By design, this shouldn't happen, so warn if it does.
                console.log( getClass().getName() + " - Warning: Ignoring energy chunk with unexpected type, type = " + incomingEnergyChunk.energyType.get().toString() );
              }
            }
            incomingEnergyChunks.clear();
          }
          // Move all of the energy chunks.
          moveElectricalEnergyChunks( dt );
          moveFilamentEnergyChunks( dt );
          moveRadiatedEnergyChunks( dt );
          // Set how lit the bulb is.
          if ( energyChunksVisible.get() ) {
            // dependent upon whether light energy chunks are present.
            var lightChunksInLitRadius = 0;
            for ( var lightEnergyChunkMover in radiatedEnergyChunkMovers ) {
              if ( lightEnergyChunkMover.energyChunk.position.get().distance( getPosition().plus( OFFSET_TO_RADIATE_POINT ) ) < LIGHT_CHUNK_LIT_BULB_RADIUS ) {
                lightChunksInLitRadius++;
              }
            }
            if ( lightChunksInLitRadius > 0 ) {
              // Light is on.
              litProportion.set( Math.min( 1, litProportion.get() + LIGHT_CHANGE_RATE * dt ) );
            }
            else {
              // Light is off.
              litProportion.set( Math.max( 0, litProportion.get() - LIGHT_CHANGE_RATE * dt ) );
            }
          }
          else {
            if ( isActive() && incomingEnergy.type == EnergyType.ELECTRICAL ) {
              litProportion.set( MathUtil.clamp( 0, incomingEnergy.amount / (ENERGY_TO_FULLY_LIGHT * dt), 1 ) );
            }
            else {
              litProportion.set( 0.0 );
            }
          }
        }
      },

      //private
      moveRadiatedEnergyChunks: function( dt ) {
        for ( var radiatedEnergyChunkMover in new ArrayList( radiatedEnergyChunkMovers ) ) {
          radiatedEnergyChunkMover.moveAlongPath( dt );
          if ( radiatedEnergyChunkMover.isPathFullyTraversed() ) {
            // Remove the chunk and its mover.
            energyChunkList.remove( radiatedEnergyChunkMover.energyChunk );
            radiatedEnergyChunkMovers.remove( radiatedEnergyChunkMover );
          }
        }
      },

      //private
      moveFilamentEnergyChunks: function( dt ) {
        for ( var filamentEnergyChunkMover in new ArrayList( filamentEnergyChunkMovers ) ) {
          filamentEnergyChunkMover.moveAlongPath( dt );
          if ( filamentEnergyChunkMover.isPathFullyTraversed() ) {
            // Cause this energy chunk to be radiated from the bulb.
            filamentEnergyChunkMovers.remove( filamentEnergyChunkMover );
            radiateEnergyChunk( filamentEnergyChunkMover.energyChunk );
          }
        }
      },

      //private
      moveElectricalEnergyChunks: function( dt ) {
        for ( var energyChunkMover in new ArrayList( electricalEnergyChunkMovers ) ) {
          energyChunkMover.moveAlongPath( dt );
          if ( energyChunkMover.isPathFullyTraversed() ) {
            electricalEnergyChunkMovers.remove( energyChunkMover );
            if ( hasFilament ) {
              // Turn this energy chunk into thermal energy on the filament.
              energyChunkMover.energyChunk.energyType.set( EnergyType.THERMAL );
              var energyChunkPath = createThermalEnergyChunkPath( energyChunkMover.energyChunk.position.get() );
              filamentEnergyChunkMovers.add( new EnergyChunkPathMover( energyChunkMover.energyChunk, energyChunkPath, getTotalPathLength( energyChunkMover.energyChunk.position.get(), energyChunkPath ) / generateThermalChunkTimeOnFilament() ) );
            }
            else {
              // There is no filament, so just radiate the chunk.
              radiateEnergyChunk( energyChunkMover.energyChunk );
            }
          }
        }
      },
      preLoadEnergyChunks: function( incomingEnergyRate ) {
        clearEnergyChunks();
        if ( incomingEnergyRate.amount < EFACConstants.MAX_ENERGY_PRODUCTION_RATE / 10 || incomingEnergyRate.type != EnergyType.ELECTRICAL ) {
          // No energy chunk pre-loading needed.
          return;
        }
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        var energySinceLastChunk = EFACConstants.ENERGY_PER_CHUNK * 0.99;
        // Simulate energy chunks moving through the system.
        var preLoadComplete = false;
        while ( !preLoadComplete ) {
          energySinceLastChunk += incomingEnergyRate.amount * dt;
          // Determine if time to add a new chunk.
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            var newEnergyChunk = new EnergyChunk( EnergyType.ELECTRICAL, getPosition().plus( OFFSET_TO_LEFT_SIDE_OF_WIRE ), energyChunksVisible );
            energyChunkList.add( newEnergyChunk );
            // the wire to the heating element.
            electricalEnergyChunkMovers.add( new EnergyChunkPathMover( newEnergyChunk, createElectricalEnergyChunkPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
            // Update energy since last chunk.
            energySinceLastChunk = energySinceLastChunk - EFACConstants.ENERGY_PER_CHUNK;
          }
          moveElectricalEnergyChunks( dt );
          moveFilamentEnergyChunks( dt );
          if ( radiatedEnergyChunkMovers.size() > 1 ) {
            // A couple of chunks are radiating, which completes the pre-load.
            preLoadComplete = true;
          }
        }
      },

      //private
      radiateEnergyChunk: function( energyChunk ) {
        if ( RAND.nextDouble() > proportionOfThermalChunksRadiated ) {
          energyChunk.energyType.set( EnergyType.LIGHT );
        }
        else {
          energyChunk.energyType.set( EnergyType.THERMAL );
        }
        var lightPath = [].withAnonymousClassBody( {
          initializer: function() {
            add( getPosition().plus( OFFSET_TO_RADIATE_POINT ).plus( new Vector2( 0, RADIATED_ENERGY_CHUNK_MAX_DISTANCE ).getRotatedInstance( (RAND.nextDouble() - 0.5) * (Math.PI / 2) ) ) );
          }
        } );
        radiatedEnergyChunkMovers.add( new EnergyChunkPathMover( energyChunk, lightPath, EFACConstants.ENERGY_CHUNK_VELOCITY ) );
      },

      //private
      createThermalEnergyChunkPath: function( startingPoint ) {
        var filamentWidth = 0.03;
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( startingPoint.plus( new Vector2( (0.5 + RAND.nextDouble() / 2) * filamentWidth / 2 * (goRightNextTime ? 1 : -1), 0 ) ) );
            goRightNextTime = !goRightNextTime;
          }
        } );
      },

      //private
      createElectricalEnergyChunkPath: function( centerPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( centerPosition.plus( OFFSET_TO_LEFT_SIDE_OF_WIRE_BEND ) );
            add( centerPosition.plus( OFFSET_TO_FIRST_WIRE_CURVE_POINT ) );
            add( centerPosition.plus( OFFSET_TO_SECOND_WIRE_CURVE_POINT ) );
            add( centerPosition.plus( OFFSET_TO_THIRD_WIRE_CURVE_POINT ) );
            add( centerPosition.plus( OFFSET_TO_BOTTOM_OF_CONNECTOR ) );
            add( centerPosition.plus( OFFSET_TO_RADIATE_POINT ) );
          }
        } );
      },

      //private
      generateThermalChunkTimeOnFilament: function() {
        return THERMAL_ENERGY_CHUNK_TIME_ON_FILAMENT.getMin() + RAND.nextDouble() * THERMAL_ENERGY_CHUNK_TIME_ON_FILAMENT.getLength();
      },

      //private
      getTotalPathLength: function( startingLocation, pathPoints ) {
        if ( pathPoints.size() == 0 ) {
          return 0;
        }
        var pathLength = startingLocation.distance( pathPoints.get( 0 ) );
        for ( var i = 0; i < pathPoints.size() - 1; i++ ) {
          pathLength += pathPoints.get( i ).distance( pathPoints.get( i + 1 ) );
        }
        return pathLength;
      },
      deactivate: function() {
        super.deactivate();
        litProportion.set( 0.0 );
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        electricalEnergyChunkMovers.clear();
        filamentEnergyChunkMovers.clear();
        radiatedEnergyChunkMovers.clear();
      },
      getUserComponent: function() {
        return userComponent;
      }
    },
//statics
    {
      WIRE_FLAT_IMAGE: WIRE_FLAT_IMAGE,
      WIRE_CURVE_IMAGE: WIRE_CURVE_IMAGE,
      ELEMENT_BASE_FRONT_IMAGE: ELEMENT_BASE_FRONT_IMAGE,
      ELEMENT_BASE_BACK_IMAGE: ELEMENT_BASE_BACK_IMAGE
    } );
} );

