// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a solar panel in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var IClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.IClock' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var CONNECTOR = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/CONNECTOR' );//static
  var SOLAR_PANEL = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/SOLAR_PANEL' );//static
  var SOLAR_PANEL_GEN = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/SOLAR_PANEL_GEN' );//static
  var SOLAR_PANEL_POST_2 = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/SOLAR_PANEL_POST_2' );//static
  var WIRE_BLACK_LEFT = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/WIRE_BLACK_LEFT' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------

  //private
  var RAND = new Random();

  //private
  var SOLAR_PANEL_OFFSET = new Vector2( 0, 0.044 );
  var SOLAR_PANEL_IMAGE = new ModelElementImage( SOLAR_PANEL, 0.15, SOLAR_PANEL_OFFSET );

  //private
  var CONVERTER_IMAGE_OFFSET = new Vector2( 0.015, -0.040 );
  var CONVERTER_IMAGE = new ModelElementImage( SOLAR_PANEL_GEN, CONVERTER_IMAGE_OFFSET );
  var CURVED_WIRE_IMAGE = new ModelElementImage( WIRE_BLACK_LEFT, CONVERTER_IMAGE_OFFSET.plus( 0.009, 0.024 ) );
  var POST_IMAGE = new ModelElementImage( SOLAR_PANEL_POST_2, CONVERTER_IMAGE_OFFSET.plus( new Vector2( 0, 0.04 ) ) );

  //private
  var CONNECTOR_IMAGE_OFFSET = new Vector2( 0.057, -0.04 );
  var CONNECTOR_IMAGE = new ModelElementImage( CONNECTOR, CONNECTOR_IMAGE_OFFSET );

  //private
  var PANEL_IMAGE_BOUNDS = new Rectangle.Number( -SOLAR_PANEL_IMAGE.getWidth() / 2, -SOLAR_PANEL_IMAGE.getHeight() / 2, SOLAR_PANEL_IMAGE.getWidth(), SOLAR_PANEL_IMAGE.getHeight() );

  //private
  var ABSORPTION_SHAPE = new DoubleGeneralPath().withAnonymousClassBody( {
    initializer: function() {
      var absorptionZoneWidth = PANEL_IMAGE_BOUNDS.getWidth() * 0.2;
      moveTo( PANEL_IMAGE_BOUNDS.getMinX(), PANEL_IMAGE_BOUNDS.getMinY() );
      moveTo( PANEL_IMAGE_BOUNDS.getMaxX() - absorptionZoneWidth, PANEL_IMAGE_BOUNDS.getMaxY() );
      lineTo( PANEL_IMAGE_BOUNDS.getMaxX(), PANEL_IMAGE_BOUNDS.getMaxY() );
      lineTo( PANEL_IMAGE_BOUNDS.getMinX() + absorptionZoneWidth, PANEL_IMAGE_BOUNDS.getMinY() );
      lineTo( PANEL_IMAGE_BOUNDS.getMinX(), PANEL_IMAGE_BOUNDS.getMinY() );
      closePath();
    }
  } );
// Constants used for creating the path followed by the energy chunks.
// Many of these numbers were empirically determined based on the images,
// and will need updating if the images change.

  //private
  var OFFSET_TO_CONVERGENCE_POINT = new Vector2( CONVERTER_IMAGE_OFFSET.getX(), 0.01 );

  //private
  var OFFSET_TO_FIRST_CURVE_POINT = new Vector2( CONVERTER_IMAGE_OFFSET.getX(), -0.025 );

  //private
  var OFFSET_TO_SECOND_CURVE_POINT = new Vector2( CONVERTER_IMAGE_OFFSET.getX() + 0.005, -0.033 );

  //private
  var OFFSET_TO_THIRD_CURVE_POINT = new Vector2( CONVERTER_IMAGE_OFFSET.getX() + 0.015, CONNECTOR_IMAGE_OFFSET.getY() );

  //private
  var OFFSET_TO_CONNECTOR_CENTER = CONNECTOR_IMAGE_OFFSET;
// Inter chunk spacing time for when the chunks reach the 'convergence
// point' at the bottom of the solar panel.  It is intended to
// approximately match the rate at which the sun emits energy chunks.
// In seconds.

  //private
  var MIN_INTER_CHUNK_TIME = 1 / (Sun.ENERGY_CHUNK_EMISSION_PERIOD * Sun.NUM_EMISSION_SECTORS);

  function SolarPanel( simulationClock, energyChunkVisibility ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.energyChunkMovers = [];

    //private
    this.simulationClock;
    // Used to prevent clumping of chunks.

    //private
    this.latestChunkArrivalTime = 0;

    //private
    this.energyOutputRate = 0;

    //private
    this.energyChunkVisibility;
    EnergyConverter.call( this, EnergyFormsAndChangesResources.Images.SOLAR_PANEL_ICON );
    this.simulationClock = simulationClock;
    this.energyChunkVisibility = energyChunkVisibility;
  }

  return inherit( EnergyConverter, SolarPanel, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt, incomingEnergy ) {
        if ( isActive() ) {
          // Handle any incoming energy chunks.
          if ( !incomingEnergyChunks.isEmpty() ) {
            for ( var incomingEnergyChunk in incomingEnergyChunks ) {
              if ( incomingEnergyChunk.energyType.get() == EnergyType.LIGHT ) {
                // the list of energy chunks being managed.
                incomingEnergyChunk.energyType.set( EnergyType.ELECTRICAL );
                energyChunkList.add( incomingEnergyChunk );
                // to the bottom of the solar panel.
                energyChunkMovers.add( new EnergyChunkPathMover( incomingEnergyChunk, createPathToPanelBottom( getPosition() ), chooseChunkVelocityOnPanel( incomingEnergyChunk ) ) );
              }
              else {
                // By design, this shouldn't happen, so warn if it does.
                console.log( getClass().getName() + " - Warning: Ignoring energy chunk with unexpected type, type = " + incomingEnergyChunk.energyType.get().toString() );
              }
            }
            incomingEnergyChunks.clear();
          }
          // Move the energy chunks that are currently under management.
          moveEnergyChunks( dt );
        }
        // Produce the appropriate amount of energy.
        var energyProduced = 0;
        if ( isActive() && incomingEnergy.type == EnergyType.LIGHT ) {
          // Perfectly efficient conversion. We should patent this.
          energyProduced = incomingEnergy.amount;
        }
        energyOutputRate = energyProduced / dt;
        return new Energy( EnergyType.ELECTRICAL, energyProduced, 0 );
      },

      //private
      moveEnergyChunks: function( dt ) {
        for ( var energyChunkMover in new ArrayList( energyChunkMovers ) ) {
          energyChunkMover.moveAlongPath( dt );
          if ( energyChunkMover.isPathFullyTraversed() ) {
            energyChunkMovers.remove( energyChunkMover );
            if ( energyChunkMover.energyChunk.position.get().equals( getPosition().plus( OFFSET_TO_CONVERGENCE_POINT ) ) ) {
              // Energy chunk has reached the bottom of the panel and now needs to move through the converter.
              energyChunkMovers.add( new EnergyChunkPathMover( energyChunkMover.energyChunk, createPathThroughConverter( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
            }
            else {
              // the converter, so pass it off to the next element in the system.
              outgoingEnergyChunks.add( energyChunkMover.energyChunk );
              energyChunkList.remove( energyChunkMover.energyChunk );
            }
          }
        }
      },
      preLoadEnergyChunks: function( incomingEnergyRate ) {
        clearEnergyChunks();
        if ( incomingEnergyRate.amount == 0 || incomingEnergyRate.type != EnergyType.LIGHT ) {
          // No energy chunk pre-loading needed.
          return;
        }
        var lowerLeftOfPanel = new Vector2( getAbsorptionShape().getBounds2D().getMinX(), getAbsorptionShape().getBounds2D().getMinY() );
        var upperRightOfPanel = new Vector2( getAbsorptionShape().getBounds2D().getMaxX(), getAbsorptionShape().getBounds2D().getMaxY() );
        var crossLineAngle = upperRightOfPanel.minus( lowerLeftOfPanel ).getAngle();
        var crossLineLength = lowerLeftOfPanel.distance( upperRightOfPanel );
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        var energySinceLastChunk = EFACConstants.ENERGY_PER_CHUNK * 0.99;
        // Simulate energy chunks moving through the system.
        var preLoadComplete = false;
        while ( !preLoadComplete ) {
          // Full energy rate generates too many chunks, so an adjustment factor is used.
          energySinceLastChunk += incomingEnergyRate.amount * dt * 0.4;
          // Determine if time to add a new chunk.
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            var initialPosition;
            if ( energyChunkList.size() == 0 ) {
              // For predictability of the algorithm, add the first chunk to the center of the panel.
              initialPosition = lowerLeftOfPanel.plus( new Vector2( crossLineLength * 0.5, 0 ).getRotatedInstance( crossLineAngle ) );
            }
            else {
              // Choose a random location along the center portion of the cross line.
              initialPosition = lowerLeftOfPanel.plus( new Vector2( crossLineLength * (0.5 * RAND.nextDouble() + 0.25), 0 ).getRotatedInstance( crossLineAngle ) );
            }
            var newEnergyChunk = new EnergyChunk( EnergyType.ELECTRICAL, initialPosition, energyChunkVisibility );
            energyChunkList.add( newEnergyChunk );
            // to the bottom of the solar panel.
            energyChunkMovers.add( new EnergyChunkPathMover( newEnergyChunk, createPathToPanelBottom( getPosition() ), chooseChunkVelocityOnPanel( newEnergyChunk ) ) );
            // Update energy since last chunk.
            energySinceLastChunk -= EFACConstants.ENERGY_PER_CHUNK;
          }
          moveEnergyChunks( dt );
          if ( outgoingEnergyChunks.size() > 0 ) {
            // An energy chunk has made it all the way through the system, which completes the pre-load.
            preLoadComplete = true;
          }
        }
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.ELECTRICAL, energyOutputRate );
      },
// Choose velocity of chunk on panel such that it won't clump up with
// other chunks.

      //private
      chooseChunkVelocityOnPanel: function( incomingEnergyChunk ) {
        // Start with default velocity.
        var chunkVelocity = EFACConstants.ENERGY_CHUNK_VELOCITY;
        // Count the number of chunks currently on the panel.
        var numChunksOnPanel = 0;
        for ( var energyChunkMover in energyChunkMovers ) {
          if ( energyChunkMover.getFinalDestination().equals( getPosition().plus( OFFSET_TO_CONVERGENCE_POINT ) ) ) {
            numChunksOnPanel++;
          }
        }
        // Compute the projected time of arrival at the convergence point.
        var distanceToConvergencePoint = incomingEnergyChunk.position.get().distance( getPosition().plus( OFFSET_TO_CONVERGENCE_POINT ) );
        var travelTime = distanceToConvergencePoint / chunkVelocity;
        var projectedArrivalTime = simulationClock.getSimulationTime() + travelTime;
        // panel.
        var minArrivalTimeSpacing = numChunksOnPanel <= 3 ? MIN_INTER_CHUNK_TIME : MIN_INTER_CHUNK_TIME / (numChunksOnPanel - 2);
        // chunk, slow down so that the minimum spacing is maintained.
        if ( latestChunkArrivalTime + minArrivalTimeSpacing > projectedArrivalTime ) {
          projectedArrivalTime = latestChunkArrivalTime + minArrivalTimeSpacing;
        }
        latestChunkArrivalTime = projectedArrivalTime;
        return distanceToConvergencePoint / (projectedArrivalTime - simulationClock.getSimulationTime());
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        energyChunkMovers.clear();
        latestChunkArrivalTime = 0;
      },

      //private
      createPathToPanelBottom: function( panelPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( panelPosition.plus( OFFSET_TO_CONVERGENCE_POINT ) );
          }
        } );
      },

      //private
      createPathThroughConverter: function( panelPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( panelPosition.plus( OFFSET_TO_FIRST_CURVE_POINT ) );
            add( panelPosition.plus( OFFSET_TO_SECOND_CURVE_POINT ) );
            add( panelPosition.plus( OFFSET_TO_THIRD_CURVE_POINT ) );
            add( panelPosition.plus( OFFSET_TO_CONNECTOR_CENTER ) );
          }
        } );
      },
      /**
       * Get the shape of the region that absorbs sunlight.
       *
       * @return A shape, in model space, of the region of the solar panel that
       *         can absorb sunlight.
       */
      getAbsorptionShape: function() {
        var transform = AffineTransform.getTranslateInstance( getPosition().getX() + SOLAR_PANEL_OFFSET.getX(), getPosition().getY() + SOLAR_PANEL_OFFSET.getY() );
        return transform.createTransformedShape( ABSORPTION_SHAPE.getGeneralPath() );
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectSolarPanelButton;
      }
    },
//statics
    {
      SOLAR_PANEL_IMAGE: SOLAR_PANEL_IMAGE,
      CONVERTER_IMAGE: CONVERTER_IMAGE,
      CURVED_WIRE_IMAGE: CURVED_WIRE_IMAGE,
      POST_IMAGE: POST_IMAGE,
      CONNECTOR_IMAGE: CONNECTOR_IMAGE
    } );
} );

