// Copyright 2002-2012, University of Colorado
/**
 * Class that represents the sun (as an energy source) in the model.  This
 * includes the clouds that can block the sun's rays.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var Collections = require( 'java.util.Collections' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// In meters, apparent size, not (obviously) actual size.
  var RADIUS = 0.02;
  var OFFSET_TO_CENTER_OF_SUN = new Vector2( -0.05, 0.12 );
// In seconds.
  var ENERGY_CHUNK_EMISSION_PERIOD = 0.11;

  //private
  var RAND = new Random();
// In meters.

  //private
  var MAX_DISTANCE_OF_E_CHUNKS_FROM_SUN = 0.5;
// Constants that control the nature of the emission sectors.  These are
// used to make emission look random yet still have a fairly steady rate
// within each sector.  One sector is intended to point at the solar panel.
  var NUM_EMISSION_SECTORS = 10;
  var EMISSION_SECTOR_SPAN = 2 * Math.PI / NUM_EMISSION_SECTORS;
// Used to tweak sector positions to make sure solar panel gets consistent flow of E's.
  var EMISSION_SECTOR_OFFSET = EMISSION_SECTOR_SPAN * 0.71;

  function Sun( solarPanel, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Clouds that can potentially block the sun's rays.  The positions are
    // set so that they appear between the sun and the solar panel, and must
    // not overlap with one another.
    this.clouds = [].withAnonymousClassBody( {
      initializer: function() {
        add( new Cloud( new Vector2( 0.02, 0.105 ), getObservablePosition() ) );
        add( new Cloud( new Vector2( 0.017, 0.0875 ), getObservablePosition() ) );
        add( new Cloud( new Vector2( -0.01, 0.08 ), getObservablePosition() ) );
      }
    } );
    this.cloudiness = new Property( 0.0 );
    // Solar panel that will, at times, be absorbing the sun's rays.
    this.solarPanel;

    //private
    this.energyChunkEmissionCountdownTimer = ENERGY_CHUNK_EMISSION_PERIOD;

    //private
    this.energyChunksVisible;

    //private
    this.sectorList = [].withAnonymousClassBody( {
      initializer: function() {
        for ( var i = 0; i < NUM_EMISSION_SECTORS; i++ ) {
          add( i );
        }
      }
    } );
    // List of energy chunks that should be allowed to pass through the clouds
    // without bouncing (i.e. being reflected).

    //private
    this.energyChunksPassingThroughClouds = [];

    //private
    this.currentSectorIndex = 0;

    //private
    this.sunPosition = OFFSET_TO_CENTER_OF_SUN;
    EnergySource.call( this, EnergyFormsAndChangesResources.Images.SUN_ICON );
    this.solarPanel = solarPanel;
    this.energyChunksVisible = energyChunksVisible;
    // Add/remove clouds based on the value of the cloudiness property.
    cloudiness.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( cloudiness ) {
        for ( var i = 0; i < clouds.size(); i++ ) {
          // Stagger the existence strength of the clouds.
          clouds.get( i ).existenceStrength.set( MathUtil.clamp( 0, cloudiness * clouds.size() - i, 1 ) );
        }
      }
    } ) );
    Collections.shuffle( sectorList );
    // Keep the sun position updated.
    getObservablePosition().addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( position ) {
        sunPosition = position.plus( OFFSET_TO_CENTER_OF_SUN );
      }
    } ) );
  }

  return inherit( EnergySource, Sun, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt ) {
        var energyProduced = 0;
        if ( isActive() ) {
          // See if it is time to emit a new energy chunk.
          energyChunkEmissionCountdownTimer -= dt;
          if ( energyChunkEmissionCountdownTimer <= 0 ) {
            // Create a new chunk and start it on its way.
            emitEnergyChunk();
            energyChunkEmissionCountdownTimer += ENERGY_CHUNK_EMISSION_PERIOD;
          }
          // Move the energy chunks.
          updateEnergyChunkPositions( dt );
          // Calculate the amount of energy produced.
          energyProduced = EFACConstants.MAX_ENERGY_PRODUCTION_RATE * (1 - cloudiness.get()) * dt;
        }
        // Produce the energy.
        return new Energy( EnergyType.LIGHT, energyProduced );
      },

      //private
      updateEnergyChunkPositions: function( dt ) {
        // Check for bouncing and absorption of the energy chunks.
        for ( var energyChunk in new ArrayList( energyChunkList ) ) {
          if ( solarPanel.isActive() && solarPanel.getAbsorptionShape().contains( energyChunk.position.get().toPoint2D() ) ) {
            // put it on the list of outgoing chunks.
            outgoingEnergyChunks.add( energyChunk );
          }
          else if ( energyChunk.position.get().distance( sunPosition.plus( OFFSET_TO_CENTER_OF_SUN ) ) > MAX_DISTANCE_OF_E_CHUNKS_FROM_SUN ) {
            // This energy chunk is out of visible range, so remove it.
            energyChunkList.remove( energyChunk );
            energyChunksPassingThroughClouds.remove( energyChunk );
          }
          else {
            for ( var cloud in clouds ) {
              if ( cloud.getCloudAbsorptionReflectionShape().contains( energyChunk.position.get().toPoint2D() ) && !energyChunksPassingThroughClouds.contains( energyChunk ) && Math.abs( energyChunk.getVelocity().getAngle() - energyChunk.position.get().minus( sunPosition ).getAngle() ) < Math.PI / 10 ) {
                // through the clouds or be reflected.
                if ( RAND.nextDouble() < cloud.existenceStrength.get() ) {
                  // code below tries to avoid that.
                  var angleTowardsSun = energyChunk.getVelocity().getAngle() + Math.PI;
                  var reflectionAngle = new Vector2( cloud.getCenterPosition(), energyChunk.position.get() ).getAngle();
                  if ( reflectionAngle < angleTowardsSun ) {
                    energyChunk.setVelocity( energyChunk.getVelocity().getRotatedInstance( 0.7 * Math.PI + RAND.nextDouble() * Math.PI / 8 ) );
                  }
                  else {
                    energyChunk.setVelocity( energyChunk.getVelocity().getRotatedInstance( -0.7 * Math.PI - RAND.nextDouble() * Math.PI / 8 ) );
                  }
                }
                else {
                  // Let is pass through the cloud.
                  energyChunksPassingThroughClouds.add( energyChunk );
                }
              }
            }
          }
        }
        // Move all the energy chunks away from the sun.
        for ( var energyChunk in energyChunkList ) {
          energyChunk.translateBasedOnVelocity( dt );
        }
      },

      //private
      emitEnergyChunk: function() {
        var directionAngle = chooseNextEmissionAngle();
        var initialPosition = sunPosition.plus( new Vector2( RADIUS / 2, 0 ).getRotatedInstance( directionAngle ) );
        var energyChunk = new EnergyChunk( EnergyType.LIGHT, initialPosition.x, initialPosition.y, energyChunksVisible );
        energyChunk.setVelocity( new Vector2( EFACConstants.ENERGY_CHUNK_VELOCITY, 0 ).getRotatedInstance( directionAngle ) );
        energyChunkList.add( energyChunk );
      },
      preLoadEnergyChunks: function() {
        clearEnergyChunks();
        // In simulated seconds, empirically determined.
        var preLoadTime = 6;
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        energyChunkEmissionCountdownTimer = 0;
        // Simulate energy chunks moving through the system.
        while ( preLoadTime > 0 ) {
          energyChunkEmissionCountdownTimer -= dt;
          if ( energyChunkEmissionCountdownTimer <= 0 ) {
            emitEnergyChunk();
            energyChunkEmissionCountdownTimer += ENERGY_CHUNK_EMISSION_PERIOD;
          }
          updateEnergyChunkPositions( dt );
          preLoadTime -= dt;
        }
        // Remove any chunks that actually made it to the solar panel.
        outgoingEnergyChunks.clear();
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.LIGHT, EFACConstants.MAX_ENERGY_PRODUCTION_RATE * (1 - cloudiness.get()) );
      },
// Choose the angle for the emission of an energy chunk from the sun.
// This uses history and probability to make the distribution somewhat
// even but still random looking.

      //private
      chooseNextEmissionAngle: function() {
        var sector = sectorList.get( currentSectorIndex );
        currentSectorIndex++;
        if ( currentSectorIndex >= NUM_EMISSION_SECTORS ) {
          currentSectorIndex = 0;
        }
        // within the sector.
        return sector * EMISSION_SECTOR_SPAN + (RAND.nextDouble() * EMISSION_SECTOR_SPAN) + EMISSION_SECTOR_OFFSET;
      },
      activate: function() {
        super.activate();
        // that are close, but not quite reaching, the solar panel.
        for ( var i = 0; i < 100; i++ ) {
          stepInTime( EFACConstants.SIM_TIME_PER_TICK_NORMAL );
        }
      },
      deactivate: function() {
        super.deactivate();
        cloudiness.reset();
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectSunButton;
      }
    },
//statics
    {
      RADIUS: RADIUS,
      OFFSET_TO_CENTER_OF_SUN: OFFSET_TO_CENTER_OF_SUN,
      ENERGY_CHUNK_EMISSION_PERIOD: ENERGY_CHUNK_EMISSION_PERIOD,
      NUM_EMISSION_SECTORS: NUM_EMISSION_SECTORS,
      EMISSION_SECTOR_SPAN: EMISSION_SECTOR_SPAN,
      EMISSION_SECTOR_OFFSET: EMISSION_SECTOR_OFFSET
    } );
} );

