// Copyright 2002-2012, University of Colorado
/**
 * Class that represents the steam-generating tea pot in the model.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var IUserComponent = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.IUserComponent' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var TEAPOT_LARGE = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources/Images/TEAPOT_LARGE' );//static
  var ENERGY_PER_CHUNK = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants/ENERGY_PER_CHUNK' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Images used in representing this energy system element in the view.

  //private
  var TEAPOT_OFFSET = new Vector2( 0.0, 0.015 );
  var TEAPOT_IMAGE = new ModelElementImage( TEAPOT_LARGE, TEAPOT_OFFSET );
// Offsets and other constants used for energy paths.  These are mostly
// empirically determined and coordinated with the image.

  //private
  var SPOUT_BOTTOM_OFFSET = new Vector2( 0.03, 0.02 );

  //private
  var SPOUT_TIP_OFFSET = new Vector2( 0.25, 0.3 );

  //private
  var DISTANT_TARGET_OFFSET = new Vector2( 1, 1 );
// From teapot position, in meters.

  //private
  var WATER_SURFACE_HEIGHT_OFFSET = 0;
// In meters, must be coordinated with heater position.

  //private
  var THERMAL_ENERGY_CHUNK_Y_ORIGIN = -0.05;
// In meters, must be coordinated with heater position.

  //private
  var THERMAL_ENERGY_CHUNK_X_ORIGIN_RANGE = new DoubleRange( -0.015, 0.015 );
// Miscellaneous other constants.
// In joules/second

  //private
  var MAX_ENERGY_CHANGE_RATE = EFACConstants.MAX_ENERGY_PRODUCTION_RATE / 5;
// Controls rate at which tea pot cools down, empirically determined.

  //private
  var COOLING_CONSTANT = 0.1;
// In joules/second

  //private
  var COOL_DOWN_COMPLETE_THRESHOLD = 30;

  //private
  var ENERGY_CHUNK_TRANSFER_DISTANCE_RANGE = new DoubleRange( 0.12, 0.15 );

  //private
  var RAND = new Random();
// Used to keep chunks evenly spaced.

  //private
  var ENERGY_CHUNK_WATER_TO_SPOUT_TIME = 0.7;

  function TeaPot( energyChunksVisible, steamPowerableElementInPlace ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    this.heatCoolAmount = new Property( 0.0 );

    //private
    this.energyProductionRate = new Property( 0.0 );

    //private
    this.heatEnergyProducedSinceLastChunk = ENERGY_PER_CHUNK / 2;

    //private
    this.energyChunksVisible;

    //private
    this.steamPowerableElementInPlace;

    //private
    this.energyChunkMovers = [];
    // List of chunks that are not being transferred to the next energy system
    // element.

    //private
    this.exemptFromTransferEnergyChunks = [];
    // Flag for whether next chunk should be transferred or kept, used to
    // alternate transfer with non-transfer.

    //private
    this.transferNextAvailableChunk = true;
    EnergySource.call( this, EnergyFormsAndChangesResources.Images.TEAPOT_ICON );
    this.energyChunksVisible = energyChunksVisible;
    this.steamPowerableElementInPlace = steamPowerableElementInPlace;
  }

  return inherit( EnergySource, TeaPot, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      stepInTime: function( dt ) {
        if ( isActive() ) {
          if ( heatCoolAmount.get() > 0 || energyProductionRate.get() > COOL_DOWN_COMPLETE_THRESHOLD ) {
            // Analogous to acceleration.
            var energyProductionIncreaseRate = heatCoolAmount.get() * MAX_ENERGY_CHANGE_RATE;
            // Analogous to friction.
            var energyProductionDecreaseRate = energyProductionRate.get() * COOLING_CONSTANT;
            energyProductionRate.set( Math.min( energyProductionRate.get() + energyProductionIncreaseRate * dt - energyProductionDecreaseRate * dt, // Analogous to velocity.
              EFACConstants.MAX_ENERGY_PRODUCTION_RATE ) );
          }
          else {
            // trickle on forever.
            energyProductionRate.set( 0.0 );
          }
          // See if it's time to emit a new energy chunk from the heater.
          heatEnergyProducedSinceLastChunk += Math.max( heatCoolAmount.get(), 0 ) * EFACConstants.MAX_ENERGY_PRODUCTION_RATE * dt;
          if ( heatEnergyProducedSinceLastChunk >= ENERGY_PER_CHUNK ) {
            // Emit a new thermal energy chunk.
            var initialPosition = new Vector2( getPosition().getX() + THERMAL_ENERGY_CHUNK_X_ORIGIN_RANGE.getMin() + RAND.nextDouble() * THERMAL_ENERGY_CHUNK_X_ORIGIN_RANGE.getLength(), getPosition().getY() + THERMAL_ENERGY_CHUNK_Y_ORIGIN );
            var energyChunk = new EnergyChunk( EnergyType.THERMAL, initialPosition, energyChunksVisible );
            energyChunkList.add( energyChunk );
            heatEnergyProducedSinceLastChunk -= ENERGY_PER_CHUNK;
            energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createThermalEnergyChunkPath( initialPosition, getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
          }
          // Move all energy chunks that are under this element's control.
          moveEnergyChunks( dt );
        }
        return new Energy( EnergyType.MECHANICAL, energyProductionRate.get() * dt, Math.PI / 2 );
      },

      //private
      moveEnergyChunks: function( dt ) {
        for ( var energyChunkMover in new ArrayList( energyChunkMovers ) ) {
          energyChunkMover.moveAlongPath( dt );
          var energyChunk = energyChunkMover.energyChunk;
          if ( energyChunkMover.isPathFullyTraversed() ) {
            energyChunkMovers.remove( energyChunkMover );
            if ( energyChunk.energyType.get() == EnergyType.THERMAL && energyChunk.position.get().getY() == getPosition().getY() + WATER_SURFACE_HEIGHT_OFFSET ) {
              // This is a thermal chunk that is coming out of the water.
              if ( RAND.nextDouble() > 0.2 ) {
                // Turn the chunk into mechanical energy.
                energyChunk.energyType.set( EnergyType.MECHANICAL );
              }
              // Set this chunk on a path to the base of the spout.
              var travelDistance = energyChunk.position.get().distance( getPosition().plus( SPOUT_BOTTOM_OFFSET ) );
              energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createPathToSpoutBottom( getPosition() ), travelDistance / ENERGY_CHUNK_WATER_TO_SPOUT_TIME ) );
            }
            else if ( energyChunk.position.get().equals( getPosition().plus( SPOUT_BOTTOM_OFFSET ) ) ) {
              // The chunk is moving out of the spout.
              energyChunkMovers.add( new EnergyChunkPathMover( energyChunk, createSpoutExitPath( getPosition() ), EFACConstants.ENERGY_CHUNK_VELOCITY ) );
            }
            else {
              // This chunk is out of view, and we are done with it.
              energyChunkList.remove( energyChunk );
            }
          }
          else {
            // next energy system.
            if ( energyChunk.energyType.get() == EnergyType.MECHANICAL && steamPowerableElementInPlace.get() && ENERGY_CHUNK_TRANSFER_DISTANCE_RANGE.contains( getPosition().distance( energyChunk.position.get() ) ) && !exemptFromTransferEnergyChunks.contains( energyChunk ) ) {
              if ( transferNextAvailableChunk ) {
                // Send this chunk to the next energy system.
                outgoingEnergyChunks.add( energyChunk );
                energyChunkMovers.remove( energyChunkMover );
                // Alternate sending or keeping chunks.
                transferNextAvailableChunk = false;
              }
              else {
                // Don't transfer this chunk.
                exemptFromTransferEnergyChunks.add( energyChunk );
                // Set up to transfer the next one.
                transferNextAvailableChunk = true;
              }
            }
          }
        }
      },
      preLoadEnergyChunks: function() {
        clearEnergyChunks();
        if ( energyProductionRate.get() == 0 ) {
          // No chunks to add.
          return;
        }
        var preLoadComplete = false;
        var dt = 1 / EFACConstants.FRAMES_PER_SECOND;
        var energySinceLastChunk = EFACConstants.ENERGY_PER_CHUNK * 0.99;
        // Simulate energy chunks moving through the system.
        while ( !preLoadComplete ) {
          energySinceLastChunk += energyProductionRate.get() * dt;
          if ( energySinceLastChunk >= EFACConstants.ENERGY_PER_CHUNK ) {
            // Create a chunk inside the teapot (at the water surface).
            var initialPosition = new Vector2( getPosition().getX(), getPosition().getY() + WATER_SURFACE_HEIGHT_OFFSET );
            var energyType = RAND.nextDouble() > 0.2 ? EnergyType.MECHANICAL : EnergyType.THERMAL;
            var newEnergyChunk = new EnergyChunk( energyType, initialPosition, energyChunksVisible );
            energyChunkList.add( newEnergyChunk );
            var travelDistance = newEnergyChunk.position.get().distance( getPosition().plus( SPOUT_BOTTOM_OFFSET ) );
            energyChunkMovers.add( new EnergyChunkPathMover( newEnergyChunk, createPathToSpoutBottom( getPosition() ), travelDistance / ENERGY_CHUNK_WATER_TO_SPOUT_TIME ) );
            energySinceLastChunk = energySinceLastChunk - EFACConstants.ENERGY_PER_CHUNK;
          }
          // Update energy chunk positions.
          moveEnergyChunks( dt );
          if ( outgoingEnergyChunks.size() > 0 ) {
            // An energy chunk has traversed to the output of this system, completing the preload.
            preLoadComplete = true;
          }
        }
      },
      getEnergyOutputRate: function() {
        return new Energy( EnergyType.MECHANICAL, energyProductionRate.get() );
      },

      //private
      createThermalEnergyChunkPath: function( startPosition, teapotPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( new Vector2( startPosition.getX(), teapotPosition.getY() + WATER_SURFACE_HEIGHT_OFFSET ) );
          }
        } );
      },

      //private
      createPathToSpoutBottom: function( parentElementPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( parentElementPosition.plus( SPOUT_BOTTOM_OFFSET ) );
          }
        } );
      },

      //private
      createSpoutExitPath: function( parentElementPosition ) {
        return [].withAnonymousClassBody( {
          initializer: function() {
            add( parentElementPosition.plus( SPOUT_TIP_OFFSET ) );
            add( parentElementPosition.plus( DISTANT_TARGET_OFFSET ) );
          }
        } );
      },
      deactivate: function() {
        super.deactivate();
        heatCoolAmount.reset();
        energyProductionRate.reset();
      },
      clearEnergyChunks: function() {
        super.clearEnergyChunks();
        exemptFromTransferEnergyChunks.clear();
        energyChunkMovers.clear();
      },
      getUserComponent: function() {
        return EnergyFormsAndChangesSimSharing.UserComponents.selectTeapotButton;
      },
      getEnergyProductionRate: function() {
        return energyProductionRate;
      }
    },
//statics
    {
      TEAPOT_IMAGE: TEAPOT_IMAGE
    } );
} );

