// Copyright 2002-2012, University of Colorado
/**
 * Piccolo canvas for the "Energy Systems" tab of the Energy Forms and Changes
 * simulation.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var BasicStroke = require( 'java.awt.BasicStroke' );
  var Color = require( 'SCENERY/util/Color' );
  var Point = require( 'java.awt.Point' );
  var Dimension2D = require( 'java.awt.geom.Dimension2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Resettable = require( 'edu.colorado.phet.common.phetcommon.model.Resettable' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var PropertyCheckBox = require( 'edu.colorado.phet.common.phetcommon.view.controls.PropertyCheckBox' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var PhetPCanvas = require( 'edu.colorado.phet.common.piccolophet.PhetPCanvas' );
  var ControlPanelNode = require( 'edu.colorado.phet.common.piccolophet.nodes.ControlPanelNode' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var ResetAllButtonNode = require( 'edu.colorado.phet.common.piccolophet.nodes.ResetAllButtonNode' );
  var PiccoloClockControlPanel = require( 'edu.colorado.phet.common.piccolophet.nodes.mediabuttons.PiccoloClockControlPanel' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergySystemsModel = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/energysystems/model/EnergySystemsModel' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PSwing = require( 'edu.umd.cs.piccolox.pswing.PSwing' );
  var DEFAULT_STAGE_SIZE = require( 'edu.colorado.phet.common.piccolophet.PhetPCanvas.CenteredStage.DEFAULT_STAGE_SIZE' );//static
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );//static ///*

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------

  //private
  var STAGE_SIZE = CenteredStage.DEFAULT_STAGE_SIZE;

  //private
  var CONTROL_INSET = 10;

  function EnergySystemsCanvas( model ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.energyChunkVizCheckBoxProperty = new BooleanProperty( false );

    //private
    this.model;
    this.model = model;
    // Set up the canvas-screen transform.
    setWorldTransformStrategy( new CenteredStage( this ) );
    // (smaller numbers zoom out, larger ones zoom in).
    var mvt = ModelViewTransform.createSinglePointScaleInvertedYMapping( new Vector2( 0, 0 ), new Point( Math.round( DEFAULT_STAGE_SIZE.getWidth() * 0.5 ), Math.round( DEFAULT_STAGE_SIZE.getHeight() * 0.475 ) ), // "Zoom factor" - smaller zooms out, larger zooms in.
      ENERGY_SYSTEMS_MVT_SCALE_FACTOR );
    setBackground( SECOND_TAB_BACKGROUND_COLOR );
    var clockControl = new PSwing( new PiccoloClockControlPanel( model.getClock() ).withAnonymousClassBody( {
      initializer: function() {
        var transparent = new Color( 0, 0, 0, 0 );
        setBackground( transparent );
        getButtonCanvas().setBackground( transparent );
        getBackgroundNode().setVisible( false );
        clearAllToolTips();
      }
    } ) );
    // Create the back drop for the clock controls.
    var clockControlBackground = new // Tall enough so that users are unlikely to see bottom.
      PhetPPath(// Tall enough so that users are unlikely to see bottom.
      new Rectangle.Number( 0, 0, STAGE_SIZE.getWidth() * 2, clockControl.getFullBoundsReference().getHeight() * 20 ), EFACConstants.CLOCK_CONTROL_BACKGROUND_COLOR, new BasicStroke( 1 ), Color.BLACK );
    // Create the control for showing/hiding energy chunks.
    var showEnergyCheckBox = new PropertyCheckBox( EnergyFormsAndChangesSimSharing.UserComponents.showEnergyCheckBox, EnergyFormsAndChangesResources.Strings.ENERGY_SYMBOLS, energyChunkVizCheckBoxProperty ).withAnonymousClassBody( {
      initializer: function() {
        setFont( new PhetFont( 20 ) );
      }
    } );
    var showEnergyControlPanel = new ControlPanelNode( new PSwing( showEnergyCheckBox ), CONTROL_PANEL_BACKGROUND_COLOR, CONTROL_PANEL_OUTLINE_STROKE, CONTROL_PANEL_OUTLINE_COLOR );
    // that energy chunk pre-loading can be done prior to making ECs visible.
    energyChunkVizCheckBoxProperty.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyVisible ) {
        if ( energyVisible ) {
          model.preLoadEnergyChunks();
        }
        model.energyChunksVisible.set( energyVisible );
      }
    } ) );
    // Create the legend for energy chunks.
    var energyChunkLegend = new EnergyChunkLegend();
    // Control the visibility of energy chunks and the energy chunk legend.
    model.energyChunksVisible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyChunksVisible ) {
        energyChunkLegend.setVisible( energyChunksVisible );
      }
    } ) );
    // Create the carousel control nodes.
    var energySourcesCarouselController = new EnergySystemElementSelector( model.energySourcesCarousel );
    var energyConvertersCarouselController = new EnergySystemElementSelector( model.energyConvertersCarousel );
    var energyUsersCarouselController = new EnergySystemElementSelector( model.energyUsersCarousel );
    // Create the various energy system elements.
    var faucetAndWaterNode = new FaucetAndWaterNode( model.faucet, mvt );
    var sunNode = new SunNode( model.sun, model.energyChunksVisible, mvt );
    var teaPotNode = new TeaPotNode( model.teaPot, model.getClock(), model.energyChunksVisible, mvt );
    var bikerNode = new BikerNode( model.biker, mvt );
    var electricalGeneratorNode = new ElectricalGeneratorNode( model.waterPoweredGenerator, mvt );
    var incandescentLightBulbNode = new IncandescentLightBulbNode( model.incandescentLightBulb, model.energyChunksVisible, mvt );
    var fluorescentLightBulbNode = new FluorescentLightBulbNode( model.fluorescentLightBulb, model.energyChunksVisible, mvt );
    var beakerHeaterNode = new BeakerHeaterNode( model.getClock(), model.beakerHeater, model.energyChunksVisible, mvt );
    var solarPanelNode = new SolarPanelNode( model.solarPanel, mvt );
    var beltNode = new BeltNode( model.belt, mvt );
    // Create the reset button.
    var resetButton = new ResetAllButtonNode( new Resettable[]
    { this, model }
  ,
    this, 20, Color.black, new Color( 255, 153, 0 )
  )
    ;
    resetButton.setConfirmationEnabled( false );
    var rootNode = new Node();
    addWorldChild( rootNode );
    // between the energy system elements.
    rootNode.addChild( faucetAndWaterNode );
    rootNode.addChild( beakerHeaterNode );
    rootNode.addChild( incandescentLightBulbNode );
    rootNode.addChild( fluorescentLightBulbNode );
    rootNode.addChild( solarPanelNode );
    rootNode.addChild( sunNode );
    rootNode.addChild( electricalGeneratorNode );
    rootNode.addChild( beltNode );
    rootNode.addChild( bikerNode );
    rootNode.addChild( teaPotNode );
    rootNode.addChild( clockControlBackground );
    rootNode.addChild( clockControl );
    //        rootNode.addChild( showEnergyControlPanel );
    rootNode.addChild( energyChunkLegend );
    rootNode.addChild( resetButton );
    rootNode.addChild( energySourcesCarouselController );
    rootNode.addChild( energyConvertersCarouselController );
    rootNode.addChild( energyUsersCarouselController );
    // Clock controls.  Positions tweaked to match first tab appearance.
    clockControl.setOffset( STAGE_SIZE.getWidth() / 2 - clockControl.getFullBoundsReference().getWidth() / 2, STAGE_SIZE.getHeight() - clockControl.getFullBoundsReference().height * 0.90 );
    clockControlBackground.setOffset( STAGE_SIZE.getWidth() / 2 - clockControlBackground.getFullBoundsReference().getWidth() / 2, clockControl.getFullBoundsReference().getMinY() + 5 );
    // Energy chunk control.
    {
      var energyControlsCenterX = STAGE_SIZE.getWidth() - Math.max( showEnergyControlPanel.getFullWidth() / 2, energyChunkLegend.getFullBoundsReference().getWidth() / 2 ) - CONTROL_INSET;
      showEnergyControlPanel.setOffset( energyControlsCenterX - showEnergyControlPanel.getFullBoundsReference().getWidth() / 2, CONTROL_INSET );
      energyChunkLegend.setOffset( STAGE_SIZE.getWidth() - energyChunkLegend.getFullBoundsReference().getWidth() - CONTROL_INSET, showEnergyControlPanel.getFullBoundsReference().getMaxY() + CONTROL_INSET );
    }
    // Carousel control.
    var carouselControllersCenterYPos = clockControlBackground.getFullBoundsReference().getMinY() - energyConvertersCarouselController.getFullBoundsReference().getHeight() / 2 - 5;
    energySourcesCarouselController.centerFullBoundsOnPoint( mvt.modelToViewX( model.energySourcesCarousel.getSelectedElementPosition().getX() ), carouselControllersCenterYPos );
    energyConvertersCarouselController.centerFullBoundsOnPoint( mvt.modelToViewX( model.energyConvertersCarousel.getSelectedElementPosition().getX() ), carouselControllersCenterYPos );
    energyUsersCarouselController.centerFullBoundsOnPoint( mvt.modelToViewX( model.energyUsersCarousel.getSelectedElementPosition().getX() ), carouselControllersCenterYPos );
    // Reset button.
    resetButton.setOffset( STAGE_SIZE.getWidth() - resetButton.getFullBounds().getWidth() - CONTROL_INSET, clockControlBackground.getFullBounds().getMinY() - resetButton.getFullBounds().getWidth() - CONTROL_INSET );
    resetButton.centerFullBoundsOnPoint( (energyUsersCarouselController.getFullBoundsReference().getMaxX() + STAGE_SIZE.getWidth()) / 2, energyUsersCarouselController.getFullBoundsReference().getCenterY() );
  }

  return inherit( PhetPCanvas, EnergySystemsCanvas, {
    reset: function() {
      energyChunkVizCheckBoxProperty.reset();
      model.getClock().setPaused( false );
    }
  } );
} );

