// Copyright 2002-2012, University of Colorado
/**
 * A view element that represents a device that can be used to heat or cool
 * things.  This is based on the HeaterCoolerNode that was in common code as
 * of early October 2012, but is different in that it doesn't extend PNode.
 * Instead, it provides an API that enables the user to get the front and back
 * representations separately so that they can be layered on a canvas and other
 * view elements can be placed between them.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var BasicStroke = require( 'java.awt.BasicStroke' );
  var Color = require( 'SCENERY/util/Color' );
  var Font = require( 'SCENERY/util/Font' );
  var GradientPaint = require( 'java.awt.GradientPaint' );
  var Paint = require( 'java.awt.Paint' );
  var Shape = require( 'java.awt.Shape' );
  var Stroke = require( 'java.awt.Stroke' );
  var KeyEvent = require( 'java.awt.event.KeyEvent' );
  var Ellipse2D = require( 'java.awt.geom.Ellipse2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var SettableProperty = require( 'edu.colorado.phet.common.phetcommon.model.property.SettableProperty' );
  var UserComponents = require( 'edu.colorado.phet.common.phetcommon.simsharing.messages.UserComponents' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var ColorUtils = require( 'edu.colorado.phet.common.phetcommon.view.util.ColorUtils' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var PiccoloPhetResources = require( 'edu.colorado.phet.common.piccolophet.PiccoloPhetResources' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var PhetPText = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPText' );
  var VSliderNode = require( 'edu.colorado.phet.common.piccolophet.nodes.slider.VSliderNode' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyChunkNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/EnergyChunkNode' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PBasicInputEventHandler = require( 'edu.umd.cs.piccolo.event.PBasicInputEventHandler' );
  var PInputEvent = require( 'edu.umd.cs.piccolo.event.PInputEvent' );
  var PImage = require( 'edu.umd.cs.piccolo.nodes.PImage' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Basic color used for the stove.

  //private
  var BASE_COLOR = new Color( 159, 182, 205 );
// Valid range of heat values.

  //private
  var heatRange = new DoubleRange( -1, 1 );

  // static class: HeaterCoolerSliderNode
  var HeaterCoolerSliderNode =
  /**
   * This class is the slider that is used to control the StoveNode, causing
   * it to add heat or cooling to the simulated system.
   *
   * @author Sam Reid
   * @author John Blanco
   */
    define( function( require ) {
// Meant to look warm.

      //private
      var TOP_SIDE_TRACK_COLOR = new Color( 255, 69, 0 );
// Meant to look cold.

      //private
      var BOTTOM_SIDE_TRACK_COLOR = new Color( 0, 0, 240 );

      //private
      var ALWAYS_SNAP_THRESHOLD = 1.0 / 8.0;

      //private
      var LABEL_FONT = new PhetFont( 20, true );

      // static class: TickMark
      var TickMark =
// Convenience class for creating a tick mark that works for this slider.

        //private
        define( function( require ) {

          //private
          var INDENT = 4;

          //private
          var LENGTH = 10;

          //private
          var STROKE_WIDTH = 2;

          //private
          var STROKE = new BasicStroke( STROKE_WIDTH );

          //private
          function TickMark() {
            var path = new DoubleGeneralPath( INDENT, STROKE_WIDTH / 2 ).withAnonymousClassBody( {
              initializer: function() {
                lineTo( INDENT + LENGTH, STROKE_WIDTH / 2 );
              }
            } );
            addChild( new PhetPPath( path.getGeneralPath(), STROKE, Color.BLACK ) );
          }

          return inherit( Node, TickMark, {
          } );
        } );
      ;
      function HeaterCoolerSliderNode( value, heatingEnabled, heatLabel, coolingEnabled, coolLabel, snapToZero ) {
        if ( !(heatingEnabled || coolingEnabled) ) {
          throw new IllegalArgumentException( "Either heating or cooling must be enabled." );
        }
        var slider = new VSliderNode( UserComponents.heaterCoolerSlider, coolingEnabled ? -1 : 0, heatingEnabled ? 1 : 0, 6, 75, value, new BooleanProperty( true ) );
        // Show labels for add, zero, and remove.
        slider.addLabel( 0.0, new TickMark() );
        slider.addLabel( +1, new PhetPText( heatLabel, LABEL_FONT ) );
        slider.addLabel( -1, new PhetPText( coolLabel, LABEL_FONT ) );
        // Return to 0 when the user releases the slider.
        addInputEventListener( new PBasicInputEventHandler().withAnonymousClassBody( {
          mouseReleased: function( event ) {
            if ( snapToZero || Math.abs( value.get() ) < ALWAYS_SNAP_THRESHOLD ) {
              value.set( 0.0 );
            }
          }
        } ) );
        // indicate the heat/coolness setting.
        slider.setTrackFillPaint( new GradientPaint( 0, 0, TOP_SIDE_TRACK_COLOR, 0, slider.trackLength, BOTTOM_SIDE_TRACK_COLOR, false ) );
        addChild( slider );
      }

      return inherit( Node, HeaterCoolerSliderNode, {
      } );
    } );
  ;
  /**
   * Constructor.
   *
   * @param heatCoolLevel Property that is set based on what the user does
   *                      when interacting with the slider.  Max value is 1
   *                      (max heat), min value is -1 (max cooling).
   * @param heatLabel     Textual label for the slider knob position that
   *                      corresponds to max heat.
   * @param coolLabel     Textual label for the slider knob position that
   * @param snapToZero    Controls whether the slider will snap to the off
   */
  function HeaterCoolerView( heatCoolLevel, heatEnabled, coolEnabled, heatLabel, coolLabel, width, height, openingHeight, snapToZero, energyChunkList, mvt ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.fireImage;

    //private
    this.iceImage;

    //private
    this.heatCoolLevel;

    //private
    this.burner;
    // This class maintains two layers and makes those layers available via its
    // API.  This is done so that its parts can be added to different layers,
    // thus making more easy to make things look like they are emerging from
    // the burner.

    //private
    this.holeLayer = new Node();

    //private
    this.frontLayer = new Node();

    //private
    this.energyChunkLayer = new Node();
    this.heatCoolLevel = heatCoolLevel;
    // Create the body of the stove.
    var stoveBodyShape = new DoubleGeneralPath().withAnonymousClassBody( {
      initializer: function() {
        // Start in upper left corner.
        moveTo( 0, 0 );
        // Curve to upper right corner.
        curveTo( 0, openingHeight / 2, width, openingHeight / 2, width, 0 );
        var bottomWidth = width * 0.8;
        // Line to lower right corner.
        lineTo( (width + bottomWidth) / 2, height );
        curveTo( (width + bottomWidth) / 2, height + openingHeight / 2, (width - bottomWidth) / 2, height + openingHeight / 2, (width - bottomWidth) / 2, // Curve to lower left corner.
          height );
        // Line back to the upper left corner.
        lineTo( 0, 0 );
        // Just to be sure.
        closePath();
      }
    } );
    var stoveBodyPaint = new GradientPaint( 0, 0, ColorUtils.brighterColor( BASE_COLOR, 0.5 ), width, 0, ColorUtils.darkerColor( BASE_COLOR, 0.5 ) );
    burner = new PhetPPath( stoveBodyShape.getGeneralPath(), stoveBodyPaint, new BasicStroke( 1 ), Color.BLACK );
    // Create the inside bowl of the burner, which is an ellipse.
    var burnerInteriorShape = new Ellipse2D.Number( 0, 0, width, openingHeight );
    var burnerInteriorPaint = new GradientPaint( 0, 0, ColorUtils.darkerColor( BASE_COLOR, 0.25 ), width, 0, ColorUtils.brighterColor( BASE_COLOR, 0.5 ) );
    var burnerInterior = new PhetPPath( burnerInteriorShape, burnerInteriorPaint, new BasicStroke( 1 ), Color.LIGHT_GRAY );
    // Create the slider.
    var stoveControlSlider = new HeaterCoolerSliderNode( HeaterCoolerView.this.heatCoolLevel, heatEnabled, heatLabel, coolEnabled, coolLabel, snapToZero ).withAnonymousClassBody( {
      initializer: function() {
        // may be scaled differently for different translations.
        var maxWidth = width * 0.8;
        var maxHeight = height * 0.8;
        var sliderScale = Math.min( maxWidth / getFullBoundsReference().width, maxHeight / getFullBoundsReference().height );
        setScale( sliderScale );
      }
    } );
    // Add the images for fire and ice that come out of the stove.
    fireImage = new PImage( PiccoloPhetResources.getImage( "flame.png" ) );
    fireImage.setScale( (width * 0.8) / fireImage.getFullBoundsReference().getWidth() );
    iceImage = new PImage( PiccoloPhetResources.getImage( "ice-cube-stack.png" ) );
    iceImage.setScale( (width * 0.8) / iceImage.getFullBoundsReference().getWidth() );
    // Monitor energy chunks and add/remove as needed.
    energyChunkList.addElementAddedObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( addedEnergyChunk ) {
        var energyChunkNode = new EnergyChunkNode( addedEnergyChunk, mvt );
        energyChunkLayer.addChild( energyChunkNode );
        energyChunkList.addElementRemovedObserver( new VoidFunction1().withAnonymousClassBody( {
          apply: function( removedEnergyChunk ) {
            if ( removedEnergyChunk == addedEnergyChunk ) {
              energyChunkLayer.removeChild( energyChunkNode );
              energyChunkList.removeElementRemovedObserver( this );
            }
          }
        } ) );
      }
    } ) );
    // desired layering.
    holeLayer.addChild( burnerInterior );
    holeLayer.addChild( fireImage );
    holeLayer.addChild( iceImage );
    holeLayer.addChild( energyChunkLayer );
    frontLayer.addChild( burner );
    frontLayer.addChild( stoveControlSlider );
    // Note - Goes a little negative in Y direction.
    burnerInterior.setOffset( 0, -burnerInterior.getFullBoundsReference().height / 2 );
    stoveControlSlider.setOffset( width / 2 - stoveControlSlider.getFullBoundsReference().width / 2, height / 2 - stoveControlSlider.getFullBoundsReference().height / 2 + burnerInterior.getFullBoundsReference().height / 2 );
    // changes.
    heatCoolLevel.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( heatCoolAmount ) {
        update();
      }
    } ) );
    // add and remove heat.
    frontLayer.addInputEventListener( new PBasicInputEventHandler().withAnonymousClassBody( {
      mousePressed: function( event ) {
        // Get the keyboard focus.
        event.getInputManager().setKeyboardFocus( event.getPath() );
      },
      keyPressed: function( event ) {
        var numIncrements = 50;
        switch( event.getKeyCode() ) {
          case KeyEvent.VK_UP:
          case KeyEvent.VK_KP_UP:
            HeaterCoolerView.this.heatCoolLevel.set( Math.min( HeaterCoolerView.this.heatCoolLevel.get() + heatRange.getLength() / numIncrements, 1 ) );
            break;
          case KeyEvent.VK_DOWN:
          case KeyEvent.VK_KP_DOWN:
            HeaterCoolerView.this.heatCoolLevel.set( Math.max( HeaterCoolerView.this.heatCoolLevel.get() - heatRange.getLength() / numIncrements, -1 ) );
            break;
          case KeyEvent.VK_ESCAPE:
          case KeyEvent.VK_0:
          case KeyEvent.VK_NUMPAD0:
          case KeyEvent.VK_ENTER:
            HeaterCoolerView.this.heatCoolLevel.set( 0.0 );
            break;
          default:
            // Ignore the key.
            break;
        }
      }
    } ) );
    update();
  }

  return inherit( Object, HeaterCoolerView, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------

    //private
    update: function() {
      if ( heatCoolLevel.get() > 0 ) {
        fireImage.setOffset( burner.getFullBoundsReference().width / 2 - fireImage.getFullBoundsReference().width / 2, -heatCoolLevel.get() * fireImage.getFullBoundsReference().height * 0.9 );
        iceImage.setOffset( burner.getFullBoundsReference().width / 2 - iceImage.getFullBoundsReference().width / 2, 0 );
      }
      else if ( heatCoolLevel.get() <= 0 ) {
        iceImage.setOffset( burner.getFullBoundsReference().width / 2 - iceImage.getFullBoundsReference().width / 2, heatCoolLevel.get() * iceImage.getFullBoundsReference().height * 0.9 );
        fireImage.setOffset( burner.getFullBoundsReference().width / 2 - fireImage.getFullBoundsReference().width / 2, 0 );
      }
      iceImage.setVisible( heatCoolLevel.get() < 0 );
      fireImage.setVisible( heatCoolLevel.get() > 0 );
    },
    getHoleNode: function() {
      return holeLayer;
    },
    getFrontNode: function() {
      return frontLayer;
    },
    setOffset: function( x, y ) {
      holeLayer.setOffset( x, y );
      // The energy chunk layer must remain in global space.
      energyChunkLayer.setOffset( -x, -y );
      frontLayer.setOffset( x, y );
    },
  } );
} );

