// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a ray of light in the view.  Rays of light can have
 * shapes that reduce or block the amount of light passing through.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Color = require( 'SCENERY/util/Color' );
  var Shape = require( 'java.awt.Shape' );
  var Line2D = require( 'java.awt.geom.Line2D' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var Collections = require( 'java.util.Collections' );
  var Comparator = require( 'java.util.Comparator' );
  var List = require( 'java.util.List' );
  var Vector2 = require( 'DOT/Vector2' );
  var SimpleObserver = require( 'edu.colorado.phet.common.phetcommon.util.SimpleObserver' );
  var Node = require( 'SCENERY/nodes/Node' );


  //private
  var STROKE_THICKNESS = 2;

  //private
  var SEARCH_ITERATIONS = 10;

  //private
  var FADE_COEFFICIENT_IN_AIR = 0.005;

  // static class: PointAndFadeCoefficient
  var PointAndFadeCoefficient =

    //private
    define( function( require ) {

      //private
      function PointAndFadeCoefficient( point, fadeCoefficient ) {
        this.point;
        this.fadeCoefficient;
        this.point = point;
        this.fadeCoefficient = fadeCoefficient;
      }

      return inherit( Object, PointAndFadeCoefficient, {
      } );
    } );
  ;
  function LightRayNode( origin, endpoint, color ) {

    //private
    this.lightAbsorbingShapes = [];

    //private
    this.pointAndFadeCoefficientList = [];

    //private
    this.origin;

    //private
    this.endpoint;

    //private
    this.color;
    this.origin = origin;
    this.endpoint = endpoint;
    this.color = color;
    updateLineSegments();
  }

  return inherit( Node, LightRayNode, {
    addLightAbsorbingShape: function( lightAbsorbingShape ) {
      lightAbsorbingShapes.add( lightAbsorbingShape );
      lightAbsorbingShape.lightAbsorptionCoefficient.addObserver( new SimpleObserver().withAnonymousClassBody( {
        update: function() {
          updateLineSegments();
        }
      } ) );
    },
    removeLightAbsorbingShape: function( lightAbsorbingShape ) {
      lightAbsorbingShape.lightAbsorptionCoefficient.removeAllObservers();
      lightAbsorbingShapes.remove( lightAbsorbingShape );
      updateLineSegments();
    },

    //private
    updateLineSegments: function() {
      removeAllChildren();
      pointAndFadeCoefficientList.clear();
      // Add the initial start and end points.
      pointAndFadeCoefficientList.add( new PointAndFadeCoefficient( origin, FADE_COEFFICIENT_IN_AIR ) );
      pointAndFadeCoefficientList.add( new PointAndFadeCoefficient( endpoint, 0 ) );
      // Add the entry and exit points for each shape.
      for ( var lightAbsorbingShape in lightAbsorbingShapes ) {
        if ( lineIntersectsShapeIntersects( origin, endpoint, lightAbsorbingShape.shape ) ) {
          var entryPoint = getShapeEntryPoint( origin, endpoint, lightAbsorbingShape.shape );
          // It's conceivable that we could handle a case where the line originates in a shape, but we don't for now.
          assert && assert( entryPoint != null );
          pointAndFadeCoefficientList.add( new PointAndFadeCoefficient( entryPoint, lightAbsorbingShape.lightAbsorptionCoefficient.get() ) );
          var exitPoint = getShapeExitPoint( origin, endpoint, lightAbsorbingShape.shape );
          if ( exitPoint != null ) {
            pointAndFadeCoefficientList.add( new PointAndFadeCoefficient( exitPoint, FADE_COEFFICIENT_IN_AIR ) );
          }
        }
      }
      // Sort the list by distance from the origin.
      Collections.sort( pointAndFadeCoefficientList, new Comparator().withAnonymousClassBody( {
        compare: function( p1, p2 ) {
          return Number.compare( p1.point.distance( origin ), p2.point.distance( origin ) );
        }
      } ) );
      // Add the segments that comprise the line.
      var opacity = 255;
      for ( var i = 0; i < pointAndFadeCoefficientList.size() - 1; i++ ) {
        var fadingLineNode = new FadingLineNode( pointAndFadeCoefficientList.get( i ).point, pointAndFadeCoefficientList.get( i + 1 ).point, new Color( color.getRed(), color.getGreen(), color.getBlue(), opacity ), pointAndFadeCoefficientList.get( i ).fadeCoefficient, STROKE_THICKNESS );
        addChild( fadingLineNode );
        opacity = fadingLineNode.getOpacityAtEndpoint();
      }
    },

    //private
    lineIntersectsShapeIntersects: function( startPoint, endPoint, shape ) {
      // This was adequate in this case, but take care if reusing.
      return shape.getBounds2D().intersectsLine( startPoint.x, startPoint.y, endPoint.x, endPoint.y );
    },

    //private
    getShapeEntryPoint: function( origin, endpoint, shape ) {
      var shapeRect = shape.getBounds2D();
      var entryPoint = null;
      if ( shapeRect.intersectsLine( new Line2D.Number( origin.toPoint2D(), endpoint.toPoint2D() ) ) ) {
        var boundsEntryPoint = getRectangleEntryPoint( origin, endpoint, shapeRect );
        if ( boundsEntryPoint == null ) {
          return null;
        }
        var boundsExitPoint = getRectangleExitPoint( origin, endpoint, shapeRect );
        var searchEndPoint = boundsExitPoint == null ? endpoint : boundsExitPoint;
        // need good general edge finding.
        var angle = endpoint.minus( origin ).getAngle();
        var incrementalDistance = boundsEntryPoint.distance( searchEndPoint ) / SEARCH_ITERATIONS;
        for ( var i = 0; i < SEARCH_ITERATIONS; i++ ) {
          var testPoint = boundsEntryPoint.plus( new Vector2( incrementalDistance * i, 0 ).getRotatedInstance( angle ) );
          if ( shape.contains( testPoint.toPoint2D() ) ) {
            entryPoint = testPoint;
            break;
          }
        }
      }
      return entryPoint;
    },

    //private
    getLineIntersection: function( line1, line2 ) {
      var denominator = ((line1.getP2().getX() - line1.getP1().getX()) * (line2.getP2().getY() - line2.getP1().getY())) - ((line1.getP2().getY() - line1.getP1().getY()) * (line2.getP2().getX() - line2.getP1().getX()));
      // Check if the lines are parallel, and thus don't intersect.
      if ( denominator == 0 ) {
        return null;
      }
      var numerator = ((line1.getP1().getY() - line2.getP1().getY()) * (line2.getP2().getX() - line2.getP1().getX())) - ((line1.getP1().getX() - line2.getP1().getX()) * (line2.getP2().getY() - line2.getP1().getY()));
      var r = numerator / denominator;
      var numerator2 = ((line1.getP1().getY() - line2.getP1().getY()) * (line1.getP2().getX() - line1.getP1().getX())) - ((line1.getP1().getX() - line2.getP1().getX()) * (line1.getP2().getY() - line1.getP1().getY()));
      var s = numerator2 / denominator;
      if ( (r < 0 || r > 1) || (s < 0 || s > 1) ) {
        return null;
      }
      // Find intersection point
      return new Vector2( line1.getP1().getX() + (r * (line1.getP2().getX() - line1.getP1().getX())), line1.getP1().getY() + (r * (line1.getP2().getY() - line1.getP1().getY())) );
    },

    //private
    getRectangleEntryPoint: function( origin, endpoint, rect ) {
      var intersectingPoints = getRectangleLineIntersectionPoints( rect, new Line2D.Number( origin.toPoint2D(), endpoint.toPoint2D() ) );
      // Determine which point is closest to the origin.
      var closestIntersectionPoint = null;
      for ( var intersectingPoint in intersectingPoints ) {
        if ( closestIntersectionPoint == null || closestIntersectionPoint.distance( origin ) > intersectingPoint.distance( origin ) ) {
          closestIntersectionPoint = intersectingPoint;
        }
      }
      return closestIntersectionPoint;
    },

    //private
    getRectangleExitPoint: function( origin, endpoint, rect ) {
      var intersectingPoints = getRectangleLineIntersectionPoints( rect, new Line2D.Number( origin.toPoint2D(), endpoint.toPoint2D() ) );
      if ( intersectingPoints.size() < 2 ) {
        // Line either doesn't intersect or ends inside the rectangle.
        return null;
      }
      // Determine which point is furthest from the origin.
      var furthestIntersectionPoint = null;
      for ( var intersectingPoint in intersectingPoints ) {
        if ( furthestIntersectionPoint == null || furthestIntersectionPoint.distance( origin ) < intersectingPoint.distance( origin ) ) {
          furthestIntersectionPoint = intersectingPoint;
        }
      }
      return furthestIntersectionPoint;
    },

    //private
    getRectangleLineIntersectionPoints: function( rect, line ) {
      var linesThatMakeUpRectangle = new ArrayList( 4 );
      linesThatMakeUpRectangle.add( new Line2D.Number( rect.getMinX(), rect.getMinY(), rect.getMinX(), rect.getMaxY() ) );
      linesThatMakeUpRectangle.add( new Line2D.Number( rect.getMinX(), rect.getMaxY(), rect.getMaxX(), rect.getMaxY() ) );
      linesThatMakeUpRectangle.add( new Line2D.Number( rect.getMaxX(), rect.getMaxY(), rect.getMaxX(), rect.getMinY() ) );
      linesThatMakeUpRectangle.add( new Line2D.Number( rect.getMaxX(), rect.getMinY(), rect.getMinX(), rect.getMinY() ) );
      var intersectingPoints = [];
      for ( var rectLine in linesThatMakeUpRectangle ) {
        var intersectingPoint = getLineIntersection( rectLine, line );
        if ( intersectingPoint != null ) {
          intersectingPoints.add( intersectingPoint );
        }
      }
      return intersectingPoints;
    },

    //private
    getShapeExitPoint: function( origin, endpoint, shape ) {
      var shapeRect = shape.getBounds2D();
      var exitPoint = null;
      if ( shape.contains( endpoint.toPoint2D() ) ) {
        // Line ends inside shape, return null.
        return null;
      }
      if ( !shape.contains( endpoint.toPoint2D() ) && shapeRect.intersectsLine( new Line2D.Number( origin.toPoint2D(), endpoint.toPoint2D() ) ) ) {
        // rectangle that encloses the shape.
        var angle = endpoint.minus( origin ).getAngle();
        var length = origin.distance( endpoint );
        var lengthChange = length / 2;
        for ( var i = 0; i < SEARCH_ITERATIONS; i++ ) {
          var testLine = new Line2D.Number( origin.plus( new Vector2( length, 0 ).getRotatedInstance( angle ) ).toPoint2D(), endpoint.toPoint2D() );
          length += lengthChange * (testLine.intersects( shapeRect ) ? 1 : -1);
          lengthChange = lengthChange / 2;
        }
        exitPoint = origin.plus( new Vector2( length, 0 ).getRotatedInstance( angle ) );
      }
      return exitPoint;
    },
    main: function( args ) {
      var testLine1 = new Line2D.Number( 0, 0, 100, 100 );
      var testLine2 = new Line2D.Number( 0, 100, 100, 0 );
      console.log( "Intersection = " + getLineIntersection( testLine1, testLine2 ) );
      var testRect = new Rectangle.Number( 1, -1, 2, 2 );
      var testLineOrigin = new Vector2( 0, 0 );
      var testLineEndpoint = new Vector2( 100, 5 );
      console.log( "Entry point with rect = " + getRectangleEntryPoint( testLineOrigin, testLineEndpoint, testRect ) );
      console.log( "Exit point with rect = " + getRectangleExitPoint( testLineOrigin, testLineEndpoint, testRect ) );
    }
  } );
} );

