// Copyright 2002-2012, University of Colorado
/**
 * Node that represents the sun, clouds, and a slider to control the level
 * of cloudiness in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Ellipse2D = require( 'java.awt.geom.Ellipse2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var And = require( 'edu.colorado.phet.common.phetcommon.model.property.And' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var RoundGradientPaint = require( 'edu.colorado.phet.common.phetcommon.view.graphics.RoundGradientPaint' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var ControlPanelNode = require( 'edu.colorado.phet.common.piccolophet.nodes.ControlPanelNode' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var PhetPText = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPText' );
  var ZeroOffsetNode = require( 'edu.colorado.phet.common.piccolophet.nodes.kit.ZeroOffsetNode' );
  var HBox = require( 'edu.colorado.phet.common.piccolophet.nodes.layout.HBox' );
  var VBox = require( 'edu.colorado.phet.common.piccolophet.nodes.layout.VBox' );
  var VSliderNode = require( 'edu.colorado.phet.common.piccolophet.nodes.slider.VSliderNode' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EnergyFormsAndChangesSimSharing = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesSimSharing' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunkLayer = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/EnergyChunkLayer' );
  var Cloud = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/energysystems/model/Cloud' );
  var Sun = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/energysystems/model/Sun' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PImage = require( 'edu.umd.cs.piccolo.nodes.PImage' );


  //private
  var CONTROL_PANEL_TITLE_FONT = new PhetFont( 16, true );
// For debug.

  //private
  var SHOW_EMISSION_SECTORS = false;

  //private
  var EMISSION_SECTOR_LINE_LENGTH = 700;

  // static class: TickMarkLabel
  var TickMarkLabel =
// Convenience class for creating a tick mark label that works for the slider.

    //private
    define( function( require ) {

      //private
      var INDENT = 4;

      //private
      var LENGTH = 7;

      //private
      var STROKE_WIDTH = 1;

      //private
      var STROKE = new BasicStroke( STROKE_WIDTH );

      //private
      var LABEL_FONT = new PhetFont( 14, false );

      //private
      function TickMarkLabel( text ) {
        var path = new DoubleGeneralPath( INDENT, STROKE_WIDTH / 2 ).withAnonymousClassBody( {
          initializer: function() {
            lineTo( INDENT + LENGTH, STROKE_WIDTH / 2 );
          }
        } );
        var tickMark = new PhetPPath( path.getGeneralPath(), STROKE, Color.BLACK );
        var label = new PhetPText( text, LABEL_FONT ).withAnonymousClassBody( {
          initializer: function() {
            setOffset( tickMark.getFullBoundsReference().getMaxX() + 3, -getFullBoundsReference().height / 2 );
          }
        } );
        var rootNode = new Node();
        rootNode.addChild( tickMark );
        rootNode.addChild( label );
        addChild( new ZeroOffsetNode( rootNode ) );
      }

      return inherit( Node, TickMarkLabel, {
      } );
    } );
  ;
  function SunNode( sun, energyChunksVisible, mvt ) {
    PositionableFadableModelElementNode.call( this, sun, mvt );
    var sunRadius = mvt.modelToViewDeltaX( Sun.RADIUS );
    // Add the rays of sunlight.
    var lightRays = new LightRays( mvt.modelToViewDelta( Sun.OFFSET_TO_CENTER_OF_SUN ), sunRadius, 1000, 40, Color.YELLOW );
    addChild( lightRays );
    energyChunksVisible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyChunksVisible ) {
        // Only show the rays then the energy chunks are not shown.
        lightRays.setVisible( !energyChunksVisible );
      }
    } ) );
    // Add the energy chunks, which reside on their own layer.
    addChild( new EnergyChunkLayer( sun.energyChunkList, sun.getObservablePosition(), mvt ) );
    // Add the emission sectors, if enabled.
    if ( SHOW_EMISSION_SECTORS ) {
      for ( var i = 0; i < Sun.NUM_EMISSION_SECTORS; i++ ) {
        var path = new DoubleGeneralPath( mvt.modelToViewDelta( Sun.OFFSET_TO_CENTER_OF_SUN ) );
        var angle = i * Sun.EMISSION_SECTOR_SPAN + Sun.EMISSION_SECTOR_OFFSET;
        path.lineToRelative( EMISSION_SECTOR_LINE_LENGTH * Math.cos( angle ), -EMISSION_SECTOR_LINE_LENGTH * Math.sin( angle ) );
        addChild( new PhetPPath( path.getGeneralPath() ) );
      }
    }
    // Add the sun.
    var sunNode = new PhetPPath( new Ellipse2D.Number( -sunRadius, -sunRadius, sunRadius * 2, sunRadius * 2 ) ).withAnonymousClassBody( {
      initializer: function() {
        setOffset( mvt.modelToViewDelta( Sun.OFFSET_TO_CENTER_OF_SUN ).toPoint2D() );
        setPaint( new RoundGradientPaint( 0, 0, Color.WHITE, new Vector2( sunRadius * 0.7, sunRadius * 0.7 ), new Color( 255, 215, 0 ) ) );
        setStroke( new BasicStroke( 1 ) );
        setStrokePaint( Color.YELLOW );
      }
    } );
    addChild( sunNode );
    // Add the clouds.
    for ( var cloud in sun.clouds ) {
      var cloudNode = new CloudNode( cloud, mvt );
      addChild( cloudNode );
      var lightAbsorbingShape = new LightAbsorbingShape( cloudNode.getFullBoundsReference(), 0.0 );
      lightRays.addLightAbsorbingShape( lightAbsorbingShape );
      cloud.existenceStrength.addObserver( new VoidFunction1().withAnonymousClassBody( {
        apply: function( existenceStrength ) {
          lightAbsorbingShape.lightAbsorptionCoefficient.set( existenceStrength / 10 );
        }
      } ) );
    }
    // Add the control panel for the clouds.
    var cloudIcon = new PImage( EnergyFormsAndChangesResources.Images.CLOUD_1 ).withAnonymousClassBody( {
      initializer: function() {
        setScale( 0.2 );
      }
    } );
    var cloudinessControlPanel = new ControlPanelNode( new VBox( new HBox( new PhetPText( EnergyFormsAndChangesResources.Strings.CLOUDS, CONTROL_PANEL_TITLE_FONT ), cloudIcon ), new VSliderNode( EnergyFormsAndChangesSimSharing.UserComponents.cloudinessSlider, 0, 1, 5, 80, sun.cloudiness, new BooleanProperty( true ) ).withAnonymousClassBody( {
      initializer: function() {
        addLabel( 0, new TickMarkLabel( EnergyFormsAndChangesResources.Strings.NONE ) );
        addLabel( 1, new TickMarkLabel( EnergyFormsAndChangesResources.Strings.LOTS ) );
      }
    } ) ), EFACConstants.CONTROL_PANEL_BACKGROUND_COLOR );
    cloudinessControlPanel.setOffset( -cloudinessControlPanel.getFullBoundsReference().width / 2, -cloudinessControlPanel.getFullBoundsReference().height / 2 );
    addChild( cloudinessControlPanel );
    // Add/remove the absorption area for the solar panel.
    var sunAndSolarPanelActive = new And( sun.getObservableActiveState(), sun.solarPanel.getObservableActiveState() );
    sunAndSolarPanelActive.addObserver( new VoidFunction1().withAnonymousClassBody( {

      //private
      var currentLightAbsorbingShape = null,
      apply: function( bothActive ) {
        if ( bothActive ) {
          // relative to this node.
          var transformedButUncompensatedShape = mvt.modelToView( sun.solarPanel.getAbsorptionShape() );
          var compensationTransform = AffineTransform.getTranslateInstance( -mvt.modelToViewX( sun.getPosition().x ), -mvt.modelToViewY( sun.getPosition().y ) );
          var absorptionShape = compensationTransform.createTransformedShape( transformedButUncompensatedShape );
          currentLightAbsorbingShape = new LightAbsorbingShape( absorptionShape, 1 );
          lightRays.addLightAbsorbingShape( currentLightAbsorbingShape );
        }
        else if ( currentLightAbsorbingShape != null ) {
          lightRays.removeLightAbsorbingShape( currentLightAbsorbingShape );
          currentLightAbsorbingShape = null;
        }
      }
    } ) );
  }

  return inherit( PositionableFadableModelElementNode, SunNode, {
  } );
} );

