// Copyright 2002-2012, University of Colorado
/**
 * Piccolo node that represents a teapot with a burner beneath it in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Area = require( 'java.awt.geom.Area' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var IClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.IClock' );
  var Property = require( 'AXON/Property' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var RoundGradientPaint = require( 'edu.colorado.phet.common.phetcommon.view.graphics.RoundGradientPaint' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var ShapeUtils = require( 'edu.colorado.phet.common.phetcommon.view.util.ShapeUtils' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var BurnerStandNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/BurnerStandNode' );
  var EnergyChunkLayer = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/EnergyChunkLayer' );
  var TeaPot = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/energysystems/model/TeaPot' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PPath = require( 'edu.umd.cs.piccolo.nodes.PPath' );
  var PBounds = require( 'edu.umd.cs.piccolo.util.PBounds' );

// Empirically determined.

  //private
  var BURNER_WIDTH = 125;

  //private
  var BURNER_HEIGHT = BURNER_WIDTH * 0.75;

  //private
  var BURNER_OPENING_WIDTH = BURNER_WIDTH * 0.1;

  // static class: SteamNode
  var SteamNode =

    //private
    define( function( require ) {

      //private
      var MAX_HEIGHT_AND_WIDTH = 200;
// For debug.

      //private
      var SHOW_BOUNDS = false;

      //private
      var RANDOM = new Random();

      //private
      function SteamNode( clock, origin, energyOutput, maxEnergyOutput, isActive ) {
        var cloud = new PhetPPath( Color.LIGHT_GRAY );
        addChild( cloud );
        var overallBounds = new PhetPPath( new BasicStroke( 1 ), Color.RED );
        if ( SHOW_BOUNDS ) {
          addChild( overallBounds );
        }
        // Update the steam cloud shape on each clock tick.
        clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
          clockTicked: function( clockEvent ) {
            if ( isActive.get() ) {
              var proportion = energyOutput.get() / maxEnergyOutput;
              var heightAndWidth = proportion * MAX_HEIGHT_AND_WIDTH;
              var cloudStemShapePoints = [].withAnonymousClassBody( {
                initializer: function() {
                  // Empirically chosen
                  var stemBaseWidth = 8;
                  var startingPoint = new Vector2( 0, heightAndWidth ).plus( new Vector2( -stemBaseWidth / 2, 0 ).getRotatedInstance( Math.PI / 4 ) );
                  add( startingPoint );
                  add( startingPoint.plus( new Vector2( stemBaseWidth, 0 ).getRotatedInstance( Math.PI / 4 ) ) );
                  var stemAngularWidth = Math.PI / 4 * (1 + 0.3 * (RANDOM.nextDouble() - 0.5));
                  add( startingPoint.plus( new Vector2( heightAndWidth / 2, -heightAndWidth / 2 ).getRotatedInstance( stemAngularWidth / 2 ) ) );
                  add( startingPoint.plus( new Vector2( heightAndWidth / 2, -heightAndWidth / 2 ).getRotatedInstance( -stemAngularWidth / 2 ) ) );
                }
              } );
              var cloudBodyShapePoints = [].withAnonymousClassBody( {
                initializer: function() {
                  // Multiplier empirically chosen.
                  var cloudBodyHeightAndWidth = heightAndWidth * 0.9;
                  var centerPoint = new Vector2( heightAndWidth - cloudBodyHeightAndWidth / 2, cloudBodyHeightAndWidth / 2 );
                  var numPoints = 16;
                  for ( var i = 0; i < numPoints; i++ ) {
                    var distanceFromCenter = cloudBodyHeightAndWidth / 2 * (1 + 0.1 * (RANDOM.nextDouble() - 0.5));
                    add( centerPoint.plus( new Vector2( distanceFromCenter, 0 ).getRotatedInstance( i * (Math.PI * 2 / numPoints) ) ) );
                  }
                }
              } );
              var overallShape = new Area( ShapeUtils.createShapeFromPoints( cloudStemShapePoints ) );
              overallShape.add( new Area( ShapeUtils.createRoundedShapeFromVectorPoints( cloudBodyShapePoints ) ) );
              cloud.setPathTo( overallShape );
              overallBounds.setPathTo( new Rectangle.Number( 0, 0, heightAndWidth, heightAndWidth ) );
              var opacity = MathUtil.clamp( 0, Math.round( 255 * (proportion * 0.9) ), 255 );
              cloud.setPaint( new RoundGradientPaint( cloud.getFullBoundsReference().getWidth() / 2, cloud.getFullBoundsReference().getHeight() / 2, new Color( 255, 255, 255, opacity ), new Vector2( cloud.getFullBoundsReference().getWidth() * 0.5 + 0.005, cloud.getFullBoundsReference().getHeight() * 0.5 + 0.005 ), new Color( 200, 200, 200, opacity ) ) );
              // Move so that the lower left corner is at the origin.
              cloud.setOffset( origin.getX(), origin.getY() - heightAndWidth );
              overallBounds.setOffset( origin.getX(), origin.getY() - heightAndWidth );
              cloud.setVisible( true );
            }
            else {
              cloud.setVisible( false );
            }
          }
        } ) );
      }

      return inherit( Node, SteamNode, {
      } );
    } );
  ;
  function TeaPotNode( teaPot, clock, energyChunksVisible, mvt ) {
    PositionableFadableModelElementNode.call( this, teaPot, mvt );
    // Create the burner.
    var heaterCooler = new HeaterCoolerView( teaPot.heatCoolAmount, true, false, EnergyFormsAndChangesResources.Strings.HEAT, "", BURNER_WIDTH, BURNER_HEIGHT, BURNER_OPENING_WIDTH, false, new ObservableArray(), mvt );
    // Create the tea pot.
    var teaPotImageNode = new ModelElementImageNode( TeaPot.TEAPOT_IMAGE, mvt );
    // Create the steam node.  There are some tweak factors in the position.
    var steamNode = new SteamNode( clock, new Vector2( teaPotImageNode.getFullBoundsReference().getMaxX() - 5, teaPotImageNode.getFullBoundsReference().getMinY() + 16 ), teaPot.getEnergyProductionRate(), EFACConstants.MAX_ENERGY_PRODUCTION_RATE, teaPot.getObservableActiveState() );
    // Create the burner stand.
    var burnerStandWidth = teaPotImageNode.getFullBoundsReference().getWidth() * 0.9;
    var burnerStandHeight = burnerStandWidth * 0.7;
    var burnerStandRect = new Rectangle.Number( teaPotImageNode.getFullBoundsReference().getCenterX() - burnerStandWidth / 2, teaPotImageNode.getFullBoundsReference().getMaxY() - 35, burnerStandWidth, burnerStandHeight );
    var burnerStandNode = new BurnerStandNode( burnerStandRect, burnerStandWidth * 0.2 );
    // Make the tea pot & stand transparent when the energy chunks are visible.
    energyChunksVisible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( chunksVisible ) {
        var opaqueness = chunksVisible ? 0.7f
        :
        1;
        teaPotImageNode.setTransparency( opaqueness );
        burnerStandNode.setTransparency( opaqueness );
      }
    } ) );
    // Create the layer that will contain and manage the energy chunks.
    var energyChunkLayer = new EnergyChunkLayer( teaPot.energyChunkList, teaPot.getObservablePosition(), mvt );
    // Do the layout.
    var heaterCoolerBounds = heaterCooler.getFrontNode().getFullBoundsReference();
    heaterCooler.setOffset( burnerStandRect.getCenterX() - heaterCoolerBounds.width / 2, burnerStandRect.getMaxY() - heaterCoolerBounds.getHeight() + 15 );
    // Add the nodes in the appropriate order for the desired layering affect.
    addChild( heaterCooler.getHoleNode() );
    addChild( burnerStandNode );
    addChild( steamNode );
    addChild( energyChunkLayer );
    addChild( heaterCooler.getFrontNode() );
    addChild( teaPotImageNode );
  }

  return inherit( PositionableFadableModelElementNode, TeaPotNode, {
  } );
} );

