// Copyright 2002-2012, University of Colorado
/**
 * Class that represents the air in the model.  Air can hold heat, and can
 * exchange thermal energy with other model objects.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Dimension2D = require( 'java.awt.geom.Dimension2D' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var PDimension = require( 'edu.umd.cs.piccolo.util.PDimension' );
  var MAX_HEAT_EXCHANGE_TIME_STEP = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants/MAX_HEAT_EXCHANGE_TIME_STEP' );//static
  var getHeatTransferFactor = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/HeatTransferConstants/getHeatTransferFactor' );//static

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// 2D size of the air.  It is sized such that it will extend off the left,
// right, and top edges of screen for the most common aspect ratios of the
// view.

  //private
  var SIZE = new PDimension( 0.7, 0.3 );
// The thickness of the slice of air being modeled.  This is basically the
// z dimension, and is used solely for volume calculations.
// In meters.

  //private
  var DEPTH = 0.1;
// Constants that define the heat carrying capacity of the air.
// In J/kg-K, source = design document.

  //private
  var SPECIFIC_HEAT = 1012;
//    private static final double DENSITY = 0.001; // In kg/m^3, source = design document (and common knowledge).
// In kg/m^3, far more dense than real air, done to make things cool faster.

  //private
  var DENSITY = 10;
// Derived constants.

  //private
  var VOLUME = SIZE.getWidth() * SIZE.getHeight() * DEPTH;

  //private
  var MASS = VOLUME * DENSITY;

  //private
  var INITIAL_ENERGY = MASS * SPECIFIC_HEAT * EFACConstants.ROOM_TEMPERATURE;

  //private
  var THERMAL_CONTACT_AREA = new ThermalContactArea( new Rectangle.Number( -SIZE.getWidth() / 2, 0, SIZE.getWidth(), SIZE.getHeight() ), true );

  /**
   * Constructor.
   */
  function Air( clock, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.energy = INITIAL_ENERGY;

    //private
    this.energyChunksVisible;

    //private
    this.energyChunkList = new ObservableArray();

    //private
    this.energyChunkWanderControllers = [];
    this.energyChunksVisible = energyChunksVisible;
    // Hook up to the clock for time dependent behavior.
    clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        stepInTime( clockEvent.getSimulationTimeChange() );
      }
    } ) );
  }

  return inherit( Object, Air, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------

    //private
    stepInTime: function( dt ) {
      // Update the position of any energy chunks.
      for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
        energyChunkWanderController.updatePosition( dt );
        if ( !getThermalContactArea().getBounds().contains( energyChunkWanderController.getEnergyChunk().position.get().toPoint2D() ) ) {
          // Remove this energy chunk.
          energyChunkList.remove( energyChunkWanderController.getEnergyChunk() );
          energyChunkWanderControllers.remove( energyChunkWanderController );
        }
      }
      equalizeWithSurroundingAir( dt );
    },

    //private
    equalizeWithSurroundingAir: function( dt ) {
      if ( Math.abs( getTemperature() - EFACConstants.ROOM_TEMPERATURE ) > EFACConstants.SIGNIFICANT_TEMPERATURE_DIFFERENCE ) {
        var numFullTimeStepExchanges = Math.floor( dt / MAX_HEAT_EXCHANGE_TIME_STEP );
        var leftoverTime = dt - (numFullTimeStepExchanges * MAX_HEAT_EXCHANGE_TIME_STEP);
        for ( var i = 0; i < numFullTimeStepExchanges + 1; i++ ) {
          var timeStep = i < numFullTimeStepExchanges ? MAX_HEAT_EXCHANGE_TIME_STEP : leftoverTime;
          var thermalEnergyLost = (getTemperature() - EFACConstants.ROOM_TEMPERATURE) * HeatTransferConstants.AIR_TO_SURROUNDING_AIR_HEAT_TRANSFER_FACTOR.get() * timeStep;
          changeEnergy( -thermalEnergyLost );
        }
      }
    },
    changeEnergy: function( deltaEnergy ) {
      energy += deltaEnergy;
    },
    getEnergy: function() {
      return energy;
    },
    reset: function() {
      energy = INITIAL_ENERGY;
      energyChunkList.clear();
      energyChunkWanderControllers.clear();
    },
    exchangeEnergyWith: function( energyContainer, dt ) {
      var thermalContactLength = getThermalContactArea().getThermalContactLength( energyContainer.getThermalContactArea() );
      if ( thermalContactLength > 0 ) {
        var excessEnergy = energyContainer.getEnergyBeyondMaxTemperature();
        if ( excessEnergy == 0 ) {
          // Container is below max temperature, exchange energy normally.
          var heatTransferConstant = getHeatTransferFactor( this.getEnergyContainerCategory(), energyContainer.getEnergyContainerCategory() );
          var numFullTimeStepExchanges = Math.floor( dt / MAX_HEAT_EXCHANGE_TIME_STEP );
          var leftoverTime = dt - (numFullTimeStepExchanges * MAX_HEAT_EXCHANGE_TIME_STEP);
          for ( var i = 0; i < numFullTimeStepExchanges + 1; i++ ) {
            var timeStep = i < numFullTimeStepExchanges ? MAX_HEAT_EXCHANGE_TIME_STEP : leftoverTime;
            var thermalEnergyGained = (energyContainer.getTemperature() - getTemperature()) * thermalContactLength * heatTransferConstant * timeStep;
            energyContainer.changeEnergy( -thermalEnergyGained );
            changeEnergy( thermalEnergyGained );
          }
        }
        else {
          // Item is at max temperature.  Shed all excess energy into the air.
          energyContainer.changeEnergy( -excessEnergy );
          changeEnergy( excessEnergy );
        }
      }
    },
    addEnergyChunk: function( ec, initialWanderConstraint ) {
      ec.zPosition.set( 0.0 );
      energyChunkList.add( ec );
      energyChunkWanderControllers.add( new EnergyChunkWanderController( ec, new Property( new Vector2( ec.position.get().getX(), SIZE.getHeight() ) ), initialWanderConstraint ) );
    },
    requestEnergyChunk: function( point ) {
      // Create a new chunk at the top of the air above the specified point.
      return new EnergyChunk( EnergyType.THERMAL, point.getX(), SIZE.getHeight(), energyChunksVisible );
    },
    getCenterPoint: function() {
      return new Vector2( 0, SIZE.getHeight() / 2 );
    },
    getThermalContactArea: function() {
      return THERMAL_CONTACT_AREA;
    },
    getTemperature: function() {
      return energy / (MASS * SPECIFIC_HEAT);
    },
    getEnergyChunkList: function() {
      return energyChunkList;
    },
    getEnergyContainerCategory: function() {
      return EnergyContainerCategory.AIR;
    },
    getEnergyBeyondMaxTemperature: function() {
      // Air temperature is unlimited.
      return 0;
    }
  } );
} );

