// Copyright 2002-2012, University of Colorado
/**
 * Model element that represents a beaker that can contain other thermal
 * model elements.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Shape = require( 'java.awt.Shape' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Vector2 = require( 'DOT/Vector2' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var Beaker = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/Beaker' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );

  /*
   * Constructor.
   */
  function BeakerContainer( clock, initialPosition, width, height, potentiallyContainedElements, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Class Data
    //-------------------------------------------------------------------------
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // List of model elements that may be moved into this beaker.

    //private
    this.potentiallyContainedElements;
    Beaker.call( this, clock, initialPosition, width, height, energyChunksVisible );
    this.potentiallyContainedElements = potentiallyContainedElements;
  }

  return inherit( Beaker, BeakerContainer, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    /*
     * Update the fluid level in the beaker based upon any displacement that
     * could be caused by the given rectangles.  This algorithm is strictly
     * two dimensional, even though displacement is more of the 3D concept.
     *
     * @param potentiallyDisplacingRectangles
     *
     */
    updateFluidLevel: function( potentiallyDisplacingRectangles ) {
      // represents the fluid and the displacing rectangles.
      var fluidRectangle = new Rectangle.Number( getRect().getX(), getRect().getY(), width, height * fluidLevel.get() );
      var overlappingArea = 0;
      for ( var rectangle2D in potentiallyDisplacingRectangles ) {
        if ( rectangle2D.intersects( fluidRectangle ) ) {
          var intersection = rectangle2D.createIntersection( fluidRectangle );
          overlappingArea += intersection.getWidth() * intersection.getHeight();
        }
      }
      // empirically determined to look good.
      var newFluidLevel = Math.min( INITIAL_FLUID_LEVEL + overlappingArea * 120, 1 );
      var proportionateIncrease = newFluidLevel / fluidLevel.get();
      fluidLevel.set( newFluidLevel );
      // Update the shapes of the energy chunk slices.
      for ( var slice in slices ) {
        var originalShape = slice.getShape();
        var expandedOrCompressedShape = AffineTransform.getScaleInstance( 1, proportionateIncrease ).createTransformedShape( originalShape );
        var translationTransform = AffineTransform.getTranslateInstance( originalShape.getBounds2D().getX() - expandedOrCompressedShape.getBounds2D().getX(), originalShape.getBounds2D().getY() - expandedOrCompressedShape.getBounds2D().getY() );
        slice.setShape( translationTransform.createTransformedShape( expandedOrCompressedShape ) );
      }
    },

    //private
    isEnergyChunkObscured: function( ec ) {
      for ( var potentiallyContainedElement in potentiallyContainedElements ) {
        if ( this.getThermalContactArea().getBounds().contains( potentiallyContainedElement.getRect() ) && potentiallyContainedElement.getProjectedShape().contains( ec.position.get().toPoint2D() ) ) {
          return true;
        }
      }
      return false;
    },
    animateNonContainedEnergyChunks: function( dt ) {
      for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
        var ec = energyChunkWanderController.getEnergyChunk();
        if ( isEnergyChunkObscured( ec ) ) {
          // beaker to the fluid, so move it sideways.
          var xVel = 0.05 * dt * (getCenterPoint().getX() > ec.position.get().getX() ? -1 : 1);
          var motionVector = new Vector2( xVel, 0 );
          ec.translate( motionVector );
        }
        else {
          energyChunkWanderController.updatePosition( dt );
        }
        if ( !isEnergyChunkObscured( ec ) && getSliceBounds().contains( ec.position.get().toPoint2D() ) ) {
          // stop moving.
          moveEnergyChunkToSlices( energyChunkWanderController.getEnergyChunk() );
        }
      }
    },
    addEnergyChunk: function( ec ) {
      if ( isEnergyChunkObscured( ec ) ) {
        // because the chunk just came from the model element.
        ec.zPosition.set( 0.0 );
        approachingEnergyChunks.add( ec );
        energyChunkWanderControllers.add( new EnergyChunkWanderController( ec, position ) );
      }
      else {
        super.addEnergyChunk( ec );
      }
    }
  } );
} );

