// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a block in the model.  In the model, a block is two-
 * dimensional, so its shape is represented by a rectangle.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Vector2 = require( 'DOT/Vector2' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var Property = require( 'AXON/Property' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );

// Height and width of all block surfaces, since it is a cube.
// In meters
  var SURFACE_WIDTH = 0.045;
// Number of slices where energy chunks may be placed.

  //private
  var NUM_ENERGY_CHUNK_SLICES = 4;
// Degrees Kelvin, value is pretty much arbitrary. Whatever works.

  //private
  var MAX_TEMPERATURE = 450;
  /*
   * Constructor.
   */
  function Block( clock, initialPosition, density, specificHeat, energyChunksVisible ) {
    // Surfaces used for stacking and thermal interaction.

    //private
    this.topSurface = new Property( null );

    //private
    this.bottomSurface = new Property( null );
    RectangularThermalMovableModelElement.call( this, clock, initialPosition, SURFACE_WIDTH, SURFACE_WIDTH, Math.pow( SURFACE_WIDTH, 3 ) * density, specificHeat, energyChunksVisible );
    // Update the top and bottom surfaces whenever the position changes.
    position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( immutableVector2D ) {
        updateTopSurfaceProperty();
        updateBottomSurfaceProperty();
      }
    } ) );
  }

  return inherit( RectangularThermalMovableModelElement, Block, {
      getColor: function() {},
      /**
       * Get an image to use for the texture of the front of the block when
       * portrayed in the view, if one should be used.
       *
       * @return An image to use for the texture, or null if no texture should
       *         be used.
       */
      getFrontTextureImage: function() {
        return null;
      },
      /**
       * Get an image to use for the texture of the top of the block when
       * portrayed in the view, if one should be used.
       *
       * @return An image to use for the texture, or null if no texture should
       *         be used.
       */
      getTopTextureImage: function() {
        return null;
      },
      /**
       * Get an image to use for the texture of the side of the block when
       * portrayed in the view, if one should be used.
       *
       * @return An image to use for the texture, or null if no texture should
       *         be used.
       */
      getSideTextureImage: function() {
        return null;
      },
      getLabel: function() {},
      getTopSurfaceProperty: function() {
        return topSurface;
      },
      getBottomSurfaceProperty: function() {
        return bottomSurface;
      },
      getThermalContactArea: function() {
        return new ThermalContactArea( getRect(), false );
      },
      addEnergyChunkSlices: function() {
        // The slices for the block are intended to match the projection used in the view.
        var projectionToFront = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( SURFACE_WIDTH / 2 );
        for ( var i = 0; i < NUM_ENERGY_CHUNK_SLICES; i++ ) {
          var projectionOffsetVector = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( i * (-SURFACE_WIDTH / (NUM_ENERGY_CHUNK_SLICES - 1)) );
          var transform = AffineTransform.getTranslateInstance( projectionToFront.getX() + projectionOffsetVector.getX(), projectionToFront.getY() + projectionOffsetVector.getY() );
          slices.add( new EnergyChunkContainerSlice( transform.createTransformedShape( getRect() ), -i * (SURFACE_WIDTH / (NUM_ENERGY_CHUNK_SLICES - 1)), position ) );
        }
      },
      /**
       * Get a rectangle the defines the current shape in model space.  By
       * convention for this simulation, the position is the middle of the
       * bottom of the block's defining rectangle.
       *
       * @return rectangle that defines this item's 2D shape
       */
      getRect: function() {
        return new Rectangle.Number( position.get().getX() - SURFACE_WIDTH / 2, position.get().getY(), SURFACE_WIDTH, SURFACE_WIDTH );
      },

      //private
      updateTopSurfaceProperty: function() {
        topSurface.set( new HorizontalSurface( new DoubleRange( getRect().getMinX(), getRect().getMaxX() ), getRect().getMaxY(), this ) );
      },

      //private
      updateBottomSurfaceProperty: function() {
        bottomSurface.set( new HorizontalSurface( new DoubleRange( getRect().getMinX(), getRect().getMaxX() ), getRect().getMinY(), this ) );
      },
      /**
       * Get the "raw shape" that should be used for depicting this block in the
       * view.  In this context, "raw" means that it is untranslated.  By
       * convention for this simulation, the point (0, 0) is the bottom center
       * of the block.
       *
       * @return Untranslated essential shape of this block.
       */
      getRawShape: function() {
        return new Rectangle.Number( -SURFACE_WIDTH / 2, 0, SURFACE_WIDTH, SURFACE_WIDTH );
      },
      getEnergyBeyondMaxTemperature: function() {
        return Math.max( energy - (MAX_TEMPERATURE * mass * specificHeat), 0 );
      }
    },
//statics
    {
      SURFACE_WIDTH: SURFACE_WIDTH
    } );
} );

