// Copyright 2002-2012, University of Colorado
/**
 * Model element that represents a burner in the simulation.  The burner can
 * heat and also cool other model elements.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var MathUtil = require( 'edu.colorado.phet.common.phetcommon.math.MathUtil' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var ChangeObserver = require( 'edu.colorado.phet.common.phetcommon.model.property.ChangeObserver' );
  var Property = require( 'AXON/Property' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var BurnerStandNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/BurnerStandNode' );
  var EFACIntroCanvas = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/view/EFACIntroCanvas' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// In meters.

  //private
  var WIDTH = 0.075;

  //private
  var HEIGHT = WIDTH * 1;
// joules/sec, empirically chosen.

  //private
  var MAX_ENERGY_GENERATION_RATE = 5000;
// In meters.

  //private
  var CONTACT_DISTANCE = 0.001;
// In meters, empirically chosen.

  //private
  var ENERGY_CHUNK_CAPTURE_DISTANCE = 0.2;
// Because of the way that energy chunks are exchanged between thermal
// modeling elements within this simulation, things can end up looking a
// bit odd if a burner is turned on with nothing on it.  To account for
// this, a separate energy generation rate is used when a burner is
// exchanging energy directly with the air.
// joules/sec, multiplier empirically chosen.

  //private
  var MAX_ENERGY_GENERATION_RATE_INTO_AIR = MAX_ENERGY_GENERATION_RATE * 0.3;

  // static class: BoundedDoubleProperty
  var BoundedDoubleProperty =
// Convenience class - a Property<Double> with a limited range.
    define( function( require ) {
      function BoundedDoubleProperty( value, minValue, maxValue ) {

        //private
        this.bounds;
        Property.call( this, value );
        bounds = new Property( new DoubleRange( minValue, maxValue ) );
      }

      return inherit( Property, BoundedDoubleProperty, {
        set: function( value ) {
          var boundedValue = MathUtil.clamp( bounds.get().getMin(), value, bounds.get().getMax() );
          super.set( boundedValue );
        }
      } );
    } );
  ;
  /**
   * Constructor.
   *
   * @param clock               Clock that steps this burner in time.
   * @param position            The position in model space where this burner
   *                            exists. By convention for this simulation,
   *                            the position is
   * @param energyChunksVisible Property that controls whether the energy
   *                            chunks are visible.
   */
  function Burner( clock, position, energyChunksVisible ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.position;
    this.heatCoolLevel = new BoundedDoubleProperty( 0.0, -1, 1 );

    //private
    this.topSurface;

    //private
    this.energyChunksVisible;
    this.energyChunkList = new ObservableArray();

    //private
    this.energyChunkWanderControllers = [];
    // Track energy transferred to anything sitting on the burner.

    //private
    this.energyExchangedWithObjectSinceLastChunkTransfer = 0;
    // Track build up of energy for transferring chunks to/from the air.

    //private
    this.energyExchangedWithAirSinceLastChunkTransfer = 0;
    this.position = new Vector2( position );
    this.energyChunksVisible = energyChunksVisible;
    // dropping objects on top of burner.
    var perspectiveCompensation = getOutlineRect().getHeight() * EFACIntroCanvas.BURNER_EDGE_TO_HEIGHT_RATIO * Math.cos( BurnerStandNode.PERSPECTIVE_ANGLE );
    topSurface = new Property( new HorizontalSurface( new DoubleRange( getOutlineRect().getMinX() - perspectiveCompensation, getOutlineRect().getMaxX() + perspectiveCompensation ), getOutlineRect().getMaxY(), this ) );
    clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        stepInTime( clockEvent.getSimulationTimeChange() );
      }
    } ) );
    heatCoolLevel.addObserver( new ChangeObserver().withAnonymousClassBody( {
      update: function( newValue, oldValue ) {
        if ( newValue == 0 || (Math.signum( oldValue ) != Math.signum( newValue )) ) {
          // clear accumulated heat/cool amount.
          energyExchangedWithAirSinceLastChunkTransfer = 0;
        }
      }
    } ) );
    // Clear the accumulated energy transfer if thing is removed from burner.
    var somethingOnTop = new BooleanProperty( false );
    somethingOnTop.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( somethingOnTop ) {
        if ( !somethingOnTop ) {
          energyExchangedWithObjectSinceLastChunkTransfer = 0;
        }
      }
    } ) );
  }

  return inherit( ModelElement, Burner, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    /**
     * Get a rectangle that defines the outline of the burner.  In the model,
     * the burner is essentially a 2D rectangle.
     *
     * @return Rectangle that defines the outline in model space.
     */
    getOutlineRect: function() {
      return new Rectangle.Number( position.getX() - WIDTH / 2, position.getY(), WIDTH, HEIGHT );
    },
    getTopSurfaceProperty: function() {
      return topSurface;
    },
    /**
     * Interact with a thermal energy container, adding or removing energy
     * based on the current heat/cool setting.
     *
     * @param thermalEnergyContainer Model object that will get or give energy.
     * @param dt                     Amount of time (delta time).
     */
    addOrRemoveEnergyToFromObject: function( thermalEnergyContainer, dt ) {
      // This shouldn't be used for air - there is a specific method for that.
      assert && assert( !(thermalEnergyContainer instanceof Air) );
      if ( inContactWith( thermalEnergyContainer ) ) {
        var deltaEnergy = 0;
        if ( thermalEnergyContainer.getTemperature() > EFACConstants.FREEZING_POINT_TEMPERATURE ) {
          deltaEnergy = MAX_ENERGY_GENERATION_RATE * heatCoolLevel.get() * dt;
        }
        thermalEnergyContainer.changeEnergy( deltaEnergy );
        energyExchangedWithObjectSinceLastChunkTransfer += deltaEnergy;
      }
    },
    addOrRemoveEnergyToFromAir: function( air, dt ) {
      var deltaEnergy = MAX_ENERGY_GENERATION_RATE_INTO_AIR * heatCoolLevel.get() * dt;
      air.changeEnergy( deltaEnergy );
      energyExchangedWithAirSinceLastChunkTransfer += deltaEnergy;
    },
    inContactWith: function( thermalEnergyContainer ) {
      var containerThermalArea = thermalEnergyContainer.getThermalContactArea().getBounds();
      return (containerThermalArea.getCenterX() > getOutlineRect().getMinX() && containerThermalArea.getCenterX() < getOutlineRect().getMaxX() && Math.abs( containerThermalArea.getMinY() - getOutlineRect().getMaxY() ) < CONTACT_DISTANCE);
    },
    addEnergyChunk: function( ec ) {
      ec.zPosition.set( 0.0 );
      energyChunkList.add( ec );
      energyChunkWanderControllers.add( new EnergyChunkWanderController( ec, new Property( getEnergyChunkStartEndPoint() ) ) );
      energyExchangedWithAirSinceLastChunkTransfer = 0;
      energyExchangedWithObjectSinceLastChunkTransfer = 0;
    },

    //private
    getEnergyChunkStartEndPoint: function() {
      return new Vector2( getCenterPoint().getX(), getCenterPoint().getY() );
    },
    /**
     * Request an energy chunk from the burner.
     *
     * @param point Point from which to search for closest chunk.
     * @return Closest energy chunk, null if none are contained.
     */
    extractClosestEnergyChunk: function( point ) {
      var closestEnergyChunk = null;
      if ( energyChunkList.size() > 0 ) {
        for ( var energyChunk in energyChunkList ) {
          if ( energyChunk.position.get().distance( position ) > ENERGY_CHUNK_CAPTURE_DISTANCE && (closestEnergyChunk == null || energyChunk.position.get().distance( point ) < closestEnergyChunk.position.get().distance( point )) ) {
            // Found a closer chunk.
            closestEnergyChunk = energyChunk;
          }
        }
        energyChunkList.remove( closestEnergyChunk );
        for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
          if ( energyChunkWanderController.getEnergyChunk() == closestEnergyChunk ) {
            energyChunkWanderControllers.remove( energyChunkWanderController );
          }
        }
      }
      if ( closestEnergyChunk == null && heatCoolLevel.get() > 0 ) {
        // Create an energy chunk.
        closestEnergyChunk = new EnergyChunk( EnergyType.THERMAL, getEnergyChunkStartEndPoint(), energyChunksVisible );
      }
      if ( closestEnergyChunk != null ) {
        energyExchangedWithAirSinceLastChunkTransfer = 0;
        energyExchangedWithObjectSinceLastChunkTransfer = 0;
      }
      else {
        console.log( getClass().getName() + " - Warning: Request for energy chunk from burner when not in heat mode and no chunks contained, returning null." );
      }
      return closestEnergyChunk;
    },
    getCenterPoint: function() {
      return new Vector2( position.getX(), position.getY() + HEIGHT / 2 );
    },
    reset: function() {
      super.reset();
      energyChunkList.clear();
      energyChunkWanderControllers.clear();
      energyExchangedWithAirSinceLastChunkTransfer = 0;
      energyExchangedWithObjectSinceLastChunkTransfer = 0;
      heatCoolLevel.reset();
    },
    areAnyOnTop: function( thermalEnergyContainers ) {
      for ( var thermalEnergyContainer in thermalEnergyContainers ) {
        if ( inContactWith( thermalEnergyContainer ) ) {
          return true;
        }
      }
      return false;
    },
    getEnergyChunkCountForAir: function() {
      var count = 0;
      // almost to the burner).
      if ( energyChunkList.size() > 0 && heatCoolLevel.get() >= 0 ) {
        for ( var energyChunk in energyChunkList ) {
          if ( position.distance( energyChunk.position.get() ) > ENERGY_CHUNK_CAPTURE_DISTANCE ) {
            count++;
          }
        }
      }
      if ( count == 0 ) {
        // chunk transfer warrants another chunk.
        count = Math.round( energyExchangedWithAirSinceLastChunkTransfer / EFACConstants.ENERGY_PER_CHUNK );
      }
      return count;
    },

    //private
    stepInTime: function( dt ) {
      // Animate energy chunks.
      for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
        energyChunkWanderController.updatePosition( dt );
        if ( energyChunkWanderController.destinationReached() ) {
          energyChunkList.remove( energyChunkWanderController.getEnergyChunk() );
          energyChunkWanderControllers.remove( energyChunkWanderController );
        }
      }
    },
    getFlameIceRect: function() {
      // be coordinated with the view.
      var outlineRect = getOutlineRect();
      return new Rectangle.Number( outlineRect.getCenterX() - outlineRect.getWidth() / 4, outlineRect.getCenterY(), outlineRect.getWidth() / 2, outlineRect.getHeight() / 2 );
    },
    getTemperature: function() {
      // low value is limited to the freezing point of water.
      return Math.max( EFACConstants.ROOM_TEMPERATURE + heatCoolLevel.get() * 100, EFACConstants.FREEZING_POINT_TEMPERATURE );
    },
    /**
     * Get the number of excess of deficit energy chunks for interaction with
     * thermal objects (as opposed to air).
     *
     * @return Number of energy chunks that could be supplied or consumed.
     *         Negative value indicates that chunks should come in.
     */
    getEnergyChunkBalanceWithObjects: function() {
      return (Math.floor( Math.abs( energyExchangedWithObjectSinceLastChunkTransfer ) / EFACConstants.ENERGY_PER_CHUNK ) * Math.signum( energyExchangedWithObjectSinceLastChunkTransfer ));
    },
    canSupplyEnergyChunk: function() {
      return heatCoolLevel.get() > 0;
    },
    canAcceptEnergyChunk: function() {
      return heatCoolLevel.get() < 0;
    },
  } );
} );

