// Copyright 2002-2012, University of Colorado
/**
 * Primary model class for the "Intro" tab of the Energy Forms and Changes
 * simulation.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Line2D = require( 'java.awt.geom.Line2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var Arrays = require( 'java.util.Arrays' );
  var Collections = require( 'java.util.Collections' );
  var Comparator = require( 'java.util.Comparator' );
  var List = require( 'java.util.List' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var ConstantDtClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.ConstantDtClock' );
  var Property = require( 'AXON/Property' );
  var ChangeObserver = require( 'edu.colorado.phet.common.phetcommon.model.property.ChangeObserver' );
  var Property = require( 'AXON/Property' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var Beaker = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/Beaker' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var ITemperatureModel = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/ITemperatureModel' );
  var BurnerStandNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/BurnerStandNode' );
  var BlockNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/view/BlockNode' );
  var EFACIntroCanvas = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/view/EFACIntroCanvas' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Minimum distance allowed between two objects.  The basically prevents
// floating point issues.
// In meters.

  //private
  var MIN_INTER_ELEMENT_DISTANCE = 1E-9;
// Threshold of temperature difference between the bodies in a multi-body
// system below which energy can be exchanged with air.
// In degrees K, empirically determined.

  //private
  var MIN_TEMPERATURE_DIFF_FOR_MULTI_BODY_AIR_ENERGY_EXCHANGE = 2.0;
// Initial thermometer location, intended to be away from any model objects.

  //private
  var INITIAL_THERMOMETER_LOCATION = new Vector2( 100, 100 );

  //private
  var NUM_THERMOMETERS = 3;
// In meters.

  //private
  var BEAKER_WIDTH = 0.085;

  //private
  var BEAKER_HEIGHT = BEAKER_WIDTH * 1.1;
// Flag that can be turned on in order to print out some profiling info.

  //private
  var ENABLE_INTERNAL_PROFILING = false;
// For operations that require random behavior.

  //private
  var RAND = new Random();

  function EFACIntroModel() {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Main model clock.

    //private
    this.clock = new ConstantDtClock( EFACConstants.FRAMES_PER_SECOND );
    // Burners.

    //private
    this.leftBurner;

    //private
    this.rightBurner;
    // Movable thermal model objects.

    //private
    this.brick;

    //private
    this.ironBlock;

    //private
    this.beaker;
    // Thermometers.
    this.thermometers = [];
    // Air.

    //private
    this.air;
    // Boolean property that controls whether the energy chunks are visible to the user.
    this.energyChunksVisible = new BooleanProperty( false );

    //private
    this.movableThermalEnergyContainers = [];

    //private
    this.previousTime = 0;

    //private
    this.TIME_ARRAY_LENGTH = 100;

    //private
    this.times = new long[TIME_ARRAY_LENGTH];

    //private
    this.countUnderMin = 0;
    clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        stepInTime( clockEvent.getSimulationTimeChange() );
      }
    } ) );
    // Add the air.
    air = new Air( clock, energyChunksVisible );
    // Add the burners.
    rightBurner = new Burner( clock, new Vector2( 0.18, 0 ), energyChunksVisible );
    leftBurner = new Burner( clock, new Vector2( 0.08, 0 ), energyChunksVisible );
    // Add and position the blocks
    brick = new Brick( clock, new Vector2( -0.1, 0 ), energyChunksVisible );
    ironBlock = new IronBlock( clock, new Vector2( -0.175, 0 ), energyChunksVisible );
    // Add and position the beaker.
    {
      var listOfThingsThatCanGoInBeaker = [].withAnonymousClassBody( {
        initializer: function() {
          add( brick );
          add( ironBlock );
        }
      } );
      beaker = new BeakerContainer( clock, new Vector2( -0.015, 0 ), BEAKER_WIDTH, BEAKER_HEIGHT, listOfThingsThatCanGoInBeaker, energyChunksVisible );
    }
    // Put all the thermal containers on a list for easy iteration.
    movableThermalEnergyContainers.add( brick );
    movableThermalEnergyContainers.add( ironBlock );
    movableThermalEnergyContainers.add( beaker );
    // Add the thermometers.
    for ( var i = 0; i < NUM_THERMOMETERS; i++ ) {
      var thermometer = new ElementFollowingThermometer( this, INITIAL_THERMOMETER_LOCATION, false );
      thermometers.add( thermometer );
      // This was requested after interviews.
      thermometer.sensedElementColor.addObserver( new ChangeObserver().withAnonymousClassBody( {
        var blockWidthIncludingPerspective = ironBlock.getProjectedShape().getBounds2D().getWidth(),
        update: function( newColor, oldColor ) {
          var xRange = new DoubleRange( beaker.getRect().getCenterX() - blockWidthIncludingPerspective / 2, beaker.getRect().getCenterX() + blockWidthIncludingPerspective / 2 );
          if ( oldColor == EFACConstants.WATER_COLOR_IN_BEAKER && !thermometer.userControlled.get() && xRange.contains( thermometer.position.get().getX() ) ) {
            // Must toggle userControlled to enable element following.
            thermometer.userControlled.set( true );
            thermometer.position.set( new Vector2( beaker.getRect().getMaxX() - 0.01, beaker.getRect().getMinY() + beaker.getRect().getHeight() * 0.33 ) );
            // Must toggle userControlled to enable element following.
            thermometer.userControlled.set( false );
          }
        }
      } ) );
    }
  }

  return inherit( Object, EFACIntroModel, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    /**
     * Restore the initial conditions of the model.
     */
    reset: function() {
      clock.resetSimulationTime();
      energyChunksVisible.reset();
      air.reset();
      leftBurner.reset();
      rightBurner.reset();
      ironBlock.reset();
      brick.reset();
      beaker.reset();
      for ( var thermometer in thermometers ) {
        thermometer.reset();
      }
    },
    /**
     * Update the state of the model.
     *
     * @param dt Time step.
     */

    //private
    stepInTime: function( dt ) {
      if ( ENABLE_INTERNAL_PROFILING ) {
        var time = System.currentTimeMillis();
        times[countUnderMin++] = time - previousTime;
        if ( previousTime != 0 && time - previousTime > 48 || countUnderMin + 1 >= TIME_ARRAY_LENGTH ) {
          console.log( "----------------------" );
          for ( var i = 0; i < countUnderMin; i++ ) {
            System.out.print( times[i] + " " );
          }
          console.log();
          countUnderMin = 0;
        }
        previousTime = time;
      }
      // supporting surface.
      for ( var movableModelElement in Arrays.asList( ironBlock, brick, beaker ) ) {
        if ( !movableModelElement.userControlled.get() && movableModelElement.getSupportingSurface() == null && movableModelElement.position.get().getY() != 0 ) {
          var minYPos = 0;
          // it can land upon.
          var potentialSupportingSurface = findBestSupportSurface( movableModelElement );
          if ( potentialSupportingSurface != null ) {
            minYPos = potentialSupportingSurface.get().yPos;
            // Center the movableModelElement above its new parent
            var targetX = potentialSupportingSurface.get().getCenterX();
            movableModelElement.setX( targetX );
          }
          // meters/s*s
          var acceleration = -9.8;
          var velocity = movableModelElement.verticalVelocity.get() + acceleration * dt;
          var proposedYPos = movableModelElement.position.get().getY() + velocity * dt;
          if ( proposedYPos < minYPos ) {
            // The element has landed on the ground or some other surface.
            proposedYPos = minYPos;
            movableModelElement.verticalVelocity.set( 0.0 );
            if ( potentialSupportingSurface != null ) {
              movableModelElement.setSupportingSurface( potentialSupportingSurface );
              potentialSupportingSurface.get().addElementToSurface( movableModelElement );
            }
          }
          else {
            movableModelElement.verticalVelocity.set( velocity );
          }
          movableModelElement.position.set( new Vector2( movableModelElement.position.get().getX(), proposedYPos ) );
        }
      }
      // one or more of the blocks.
      beaker.updateFluidLevel( Arrays.asList( brick.getRect(), ironBlock.getRect() ) );
      // exchange energy with one another.
      for ( var ec1 in movableThermalEnergyContainers ) {
        for ( var ec2 in movableThermalEnergyContainers.subList( movableThermalEnergyContainers.indexOf( ec1 ) + 1, movableThermalEnergyContainers.size() ) ) {
          ec1.exchangeEnergyWith( ec2, dt );
        }
      }
      // model elements, including air.
      for ( var burner in Arrays.asList( leftBurner, rightBurner ) ) {
        if ( burner.areAnyOnTop( ironBlock, brick, beaker ) ) {
          for ( var movableThermalEnergyContainer in movableThermalEnergyContainers ) {
            burner.addOrRemoveEnergyToFromObject( movableThermalEnergyContainer, dt );
          }
        }
        else {
          // Nothing on burner, so heat/cool the air.
          burner.addOrRemoveEnergyToFromAir( air, dt );
        }
      }
      // Exchange energy chunks between burners and non-air energy containers.
      for ( var thermalModelElement in Arrays.asList( ironBlock, brick, beaker ) ) {
        for ( var burner in Arrays.asList( leftBurner, rightBurner ) ) {
          if ( burner.inContactWith( thermalModelElement ) ) {
            if ( burner.canSupplyEnergyChunk() && (burner.getEnergyChunkBalanceWithObjects() > 0 || thermalModelElement.getEnergyChunkBalance() < 0) ) {
              // Push an energy chunk into the item on the burner.
              thermalModelElement.addEnergyChunk( burner.extractClosestEnergyChunk( thermalModelElement.getCenterPoint() ) );
            }
            else if ( burner.canAcceptEnergyChunk() && (burner.getEnergyChunkBalanceWithObjects() < 0 || thermalModelElement.getEnergyChunkBalance() > 0) ) {
              // Extract an energy chunk from the model element.
              var ec = thermalModelElement.extractClosestEnergyChunk( burner.getFlameIceRect() );
              if ( ec != null ) {
                burner.addEnergyChunk( ec );
              }
            }
          }
        }
      }
      // Exchange energy chunks between movable thermal energy containers.
      for ( var ec1 in movableThermalEnergyContainers ) {
        for ( var ec2 in movableThermalEnergyContainers.subList( movableThermalEnergyContainers.indexOf( ec1 ) + 1, movableThermalEnergyContainers.size() ) ) {
          if ( ec1.getThermalContactArea().getThermalContactLength( ec2.getThermalContactArea() ) > 0 ) {
            // Exchange chunks if appropriate.
            if ( ec1.getEnergyChunkBalance() > 0 && ec2.getEnergyChunkBalance() < 0 ) {
              ec2.addEnergyChunk( ec1.extractClosestEnergyChunk( ec2.getThermalContactArea().getBounds() ) );
            }
            else if ( ec1.getEnergyChunkBalance() < 0 && ec2.getEnergyChunkBalance() > 0 ) {
              ec1.addEnergyChunk( ec2.extractClosestEnergyChunk( ec1.getThermalContactArea().getBounds() ) );
            }
          }
        }
      }
      // energy containers and the air.
      for ( var movableEnergyContainer in movableThermalEnergyContainers ) {
        // energy should be exchanged with air.
        var contactWithOtherMovableElement = false;
        var immersedInBeaker = false;
        var maxTemperatureDifference = 0;
        // energy containers.
        for ( var otherMovableEnergyContainer in movableThermalEnergyContainers ) {
          if ( movableEnergyContainer == otherMovableEnergyContainer ) {
            continue;
          }
          if ( movableEnergyContainer.getThermalContactArea().getThermalContactLength( otherMovableEnergyContainer.getThermalContactArea() ) > 0 ) {
            contactWithOtherMovableElement = true;
            maxTemperatureDifference = Math.max( Math.abs( movableEnergyContainer.getTemperature() - otherMovableEnergyContainer.getTemperature() ), maxTemperatureDifference );
          }
        }
        if ( beaker.getThermalContactArea().getBounds().contains( movableEnergyContainer.getRect() ) ) {
          // This model element is immersed in the beaker.
          immersedInBeaker = true;
        }
        // conditions met.
        if ( !contactWithOtherMovableElement || (!immersedInBeaker && (maxTemperatureDifference < MIN_TEMPERATURE_DIFF_FOR_MULTI_BODY_AIR_ENERGY_EXCHANGE || movableEnergyContainer.getEnergyBeyondMaxTemperature() > 0)) ) {
          air.exchangeEnergyWith( movableEnergyContainer, dt );
          if ( movableEnergyContainer.getEnergyChunkBalance() > 0 ) {
            var pointAbove = new Vector2( RAND.nextDouble() * movableEnergyContainer.getRect().getWidth() + movableEnergyContainer.getRect().getMinX(), movableEnergyContainer.getRect().getMaxY() );
            var ec = movableEnergyContainer.extractClosestEnergyChunk( pointAbove );
            if ( ec != null ) {
              var ecInitialMotionConstraints = null;
              if ( movableEnergyContainer instanceof Beaker ) {
                // and not just the center, stay in bounds.
                var energyChunkWidth = 0.01;
                ecInitialMotionConstraints = new Rectangle.Number( movableEnergyContainer.getRect().getX() + energyChunkWidth / 2, movableEnergyContainer.getRect().getY(), movableEnergyContainer.getRect().getWidth() - energyChunkWidth, movableEnergyContainer.getRect().getHeight() );
              }
              air.addEnergyChunk( ec, ecInitialMotionConstraints );
            }
          }
          else if ( movableEnergyContainer.getEnergyChunkBalance() < 0 && movableEnergyContainer.getTemperature() < air.getTemperature() ) {
            movableEnergyContainer.addEnergyChunk( air.requestEnergyChunk( movableEnergyContainer.getCenterPoint() ) );
          }
        }
      }
      // Exchange energy chunks between the air and the burners.
      for ( var burner in Arrays.asList( leftBurner, rightBurner ) ) {
        if ( burner.getEnergyChunkCountForAir() > 0 ) {
          air.addEnergyChunk( burner.extractClosestEnergyChunk( burner.getCenterPoint() ), null );
        }
        else if ( burner.getEnergyChunkCountForAir() < 0 ) {
          burner.addEnergyChunk( air.requestEnergyChunk( burner.getCenterPoint() ) );
        }
      }
    },
    /**
     * Validate the position being proposed for the given model element.  This
     * evaluates whether the proposed position would cause the model element
     * to move through another solid element, or the side of the beaker, or
     * something that would look weird to the user and, if so, prevent the odd
     * behavior from happening by returning a location that works better.
     *
     * @param modelElement     Element whose position is being validated.
     * @param proposedPosition Proposed new position for element
     * @return The original proposed position if valid, or alternative position
     *         if not.
     */
    validatePosition: function( modelElement, proposedPosition ) {
      // Compensate for the model element's center X position.
      var translation = new Vector2( proposedPosition ).minus( modelElement.position.get() );
      // the side due to perspective.
      var blockPerspectiveExtension = Block.SURFACE_WIDTH * BlockNode.PERSPECTIVE_EDGE_PROPORTION * Math.cos( BlockNode.PERSPECTIVE_ANGLE ) / 2;
      // difficult z-order issues.
      var standPerspectiveExtension = leftBurner.getOutlineRect().getHeight() * EFACIntroCanvas.BURNER_EDGE_TO_HEIGHT_RATIO * Math.cos( BurnerStandNode.PERSPECTIVE_ANGLE ) / 2;
      var burnerRectX = leftBurner.getOutlineRect().getX() - standPerspectiveExtension - (modelElement != beaker ? blockPerspectiveExtension : 0);
      var burnerBlockingRect = new Rectangle.Number( burnerRectX, leftBurner.getOutlineRect().getY(), rightBurner.getOutlineRect().getMaxX() - burnerRectX, leftBurner.getOutlineRect().getHeight() );
      translation = determineAllowedTranslation( modelElement.getRect(), burnerBlockingRect, translation, false );
      // Validate against the sides of the beaker.
      if ( modelElement != beaker ) {
        // 1 mm thick walls.
        var testRectThickness = 1E-3;
        var beakerRect = beaker.getRect();
        var beakerLeftSide = new Rectangle.Number( beakerRect.getMinX() - blockPerspectiveExtension, beaker.getRect().getMinY(), testRectThickness + blockPerspectiveExtension * 2, beaker.getRect().getHeight() + blockPerspectiveExtension );
        var beakerRightSide = new Rectangle.Number( beaker.getRect().getMaxX() - testRectThickness - blockPerspectiveExtension, beaker.getRect().getMinY(), testRectThickness + blockPerspectiveExtension * 2, beaker.getRect().getHeight() + blockPerspectiveExtension );
        var beakerBottom = new Rectangle.Number( beaker.getRect().getMinX(), beaker.getRect().getMinY(), beaker.getRect().getWidth(), testRectThickness );
        // the beaker will simply be lifted up.
        var restrictPositiveY = !beaker.isStackedUpon( modelElement );
        // Clamp the translation based on the beaker position.
        translation = determineAllowedTranslation( modelElement.getRect(), beakerLeftSide, translation, restrictPositiveY );
        translation = determineAllowedTranslation( modelElement.getRect(), beakerRightSide, translation, restrictPositiveY );
        translation = determineAllowedTranslation( modelElement.getRect(), beakerBottom, translation, restrictPositiveY );
      }
      // Now check the model element's motion against each of the blocks.
      for ( var block in Arrays.asList( ironBlock, brick ) ) {
        if ( modelElement == block ) {
          // Don't test against self.
          continue;
        }
        // element - the block will simply be lifted up.
        var restrictPositiveY = !block.isStackedUpon( modelElement );
        var testRect = modelElement.getRect();
        if ( modelElement == beaker ) {
          // simplify z-order handling.
          testRect = new Rectangle.Number( testRect.getX() - blockPerspectiveExtension, testRect.getY(), testRect.getWidth() + blockPerspectiveExtension * 2, testRect.getHeight() );
        }
        // handle the case where the block is immersed in the beaker.
        if ( modelElement != beaker || !beaker.getRect().contains( block.getRect() ) ) {
          translation = determineAllowedTranslation( testRect, block.getRect(), translation, restrictPositiveY );
        }
      }
      // Determine the new position based on the resultant translation.
      var newPosition = new Vector2( modelElement.position.get().plus( translation ) );
      // Clamp Y position to be positive to prevent dragging below table.
      newPosition.setY( Math.max( newPosition.getY(), 0 ) );
      return newPosition.toPoint2D();
    },
    dumpEnergies: function() {
      for ( var thermalEnergyContainer in Arrays.asList( ironBlock, brick, beaker, air ) ) {
        console.log( thermalEnergyContainer.getClass().getName() + " - energy = " + thermalEnergyContainer.getEnergy() );
      }
    },
    /*
     * Determine the portion of a proposed translation that may occur given
     * a moving rectangle and a stationary rectangle that can block the moving
     * one.
     *
     * @param movingRect
     * @param stationaryRect
     * @param proposedTranslation
     * @param restrictPosY        Boolean that controls whether the positive Y
     *                            direction is restricted.  This is often set
     *                            false if there is another model element on
     *                            top of the one being tested.
     * @return
     */

    //private
    determineAllowedTranslation: function( movingRect, stationaryRect, proposedTranslation, restrictPosY ) {
      // Test for case where rectangles already overlap.
      if ( movingRect.intersects( stationaryRect ) ) {
        // The rectangles already overlap.  Are they right on top of one another?
        if ( movingRect.getCenterX() == stationaryRect.getCenterX() && movingRect.getCenterY() == stationaryRect.getCenterY() ) {
          console.log( getClass().getName() + " - Warning: Rectangle centers in same location, returning zero vector." );
          return new Vector2( 0, 0 );
        }
        // the overlap.
        var xOverlapCure = 0;
        if ( movingRect.getMaxX() > stationaryRect.getMinX() && movingRect.getMinX() < stationaryRect.getMinX() ) {
          xOverlapCure = stationaryRect.getMinX() - movingRect.getMaxX();
        }
        else if ( stationaryRect.getMaxX() > movingRect.getMinX() && stationaryRect.getMinX() < movingRect.getMinX() ) {
          xOverlapCure = stationaryRect.getMaxX() - movingRect.getMinX();
        }
        var yOverlapCure = 0;
        if ( movingRect.getMaxY() > stationaryRect.getMinY() && movingRect.getMinY() < stationaryRect.getMinY() ) {
          yOverlapCure = stationaryRect.getMinY() - movingRect.getMaxY();
        }
        else if ( stationaryRect.getMaxY() > movingRect.getMinY() && stationaryRect.getMinY() < movingRect.getMinY() ) {
          yOverlapCure = stationaryRect.getMaxY() - movingRect.getMinY();
        }
        // since overlap was detected by the "intersects" method.
        assert && assert( !(xOverlapCure == 0 && yOverlapCure == 0) );
        // the other translation value unchanged.
        if ( xOverlapCure != 0 && Math.abs( xOverlapCure ) < Math.abs( yOverlapCure ) ) {
          return new Vector2( xOverlapCure, proposedTranslation.getY() );
        }
        else {
          return new Vector2( proposedTranslation.getX(), yOverlapCure );
        }
      }
      var xTranslation = proposedTranslation.getX();
      var yTranslation = proposedTranslation.getY();
      // X direction.
      if ( proposedTranslation.getX() > 0 ) {
        // Check for collisions moving right.
        var rightEdge = new Line2D.Number( movingRect.getMaxX(), movingRect.getMinY(), movingRect.getMaxX(), movingRect.getMaxY() );
        var rightEdgeSmear = projectShapeFromLine( rightEdge, proposedTranslation );
        if ( rightEdge.getX1() <= stationaryRect.getMinX() && rightEdgeSmear.intersects( stationaryRect ) ) {
          // Collision detected, limit motion.
          xTranslation = stationaryRect.getMinX() - rightEdge.getX1() - MIN_INTER_ELEMENT_DISTANCE;
        }
      }
      else if ( proposedTranslation.getX() < 0 ) {
        // Check for collisions moving left.
        var leftEdge = new Line2D.Number( movingRect.getMinX(), movingRect.getMinY(), movingRect.getMinX(), movingRect.getMaxY() );
        var leftEdgeSmear = projectShapeFromLine( leftEdge, proposedTranslation );
        if ( leftEdge.getX1() >= stationaryRect.getMaxX() && leftEdgeSmear.intersects( stationaryRect ) ) {
          // Collision detected, limit motion.
          xTranslation = stationaryRect.getMaxX() - leftEdge.getX1() + MIN_INTER_ELEMENT_DISTANCE;
        }
      }
      // Y direction.
      if ( proposedTranslation.getY() > 0 && restrictPosY ) {
        // Check for collisions moving up.
        var movingTopEdge = new Line2D.Number( movingRect.getMinX(), movingRect.getMaxY(), movingRect.getMaxX(), movingRect.getMaxY() );
        var topEdgeSmear = projectShapeFromLine( movingTopEdge, proposedTranslation );
        if ( movingTopEdge.getY1() <= stationaryRect.getMinY() && topEdgeSmear.intersects( stationaryRect ) ) {
          // Collision detected, limit motion.
          yTranslation = stationaryRect.getMinY() - movingTopEdge.getY1() - MIN_INTER_ELEMENT_DISTANCE;
        }
      }
      if ( proposedTranslation.getY() < 0 ) {
        // Check for collisions moving down.
        var movingBottomEdge = new Line2D.Number( movingRect.getMinX(), movingRect.getMinY(), movingRect.getMaxX(), movingRect.getMinY() );
        var bottomEdgeSmear = projectShapeFromLine( movingBottomEdge, proposedTranslation );
        if ( movingBottomEdge.getY1() >= stationaryRect.getMaxY() && bottomEdgeSmear.intersects( stationaryRect ) ) {
          // Collision detected, limit motion.
          yTranslation = stationaryRect.getMaxY() - movingBottomEdge.getY1() + MIN_INTER_ELEMENT_DISTANCE;
        }
      }
      return new Vector2( xTranslation, yTranslation );
    },
    /* Project a line into a 2D shape based on the provided projection vector.
     * This is a convenience function used by the code that detects potential
     * collisions between the 2D objects in model space.
     */

    //private
    projectShapeFromLine: function( edge, projection ) {
      var path = new DoubleGeneralPath();
      path.moveTo( edge.getP1() );
      path.lineTo( edge.getX1() + projection.getX(), edge.getY1() + projection.getY() );
      path.lineTo( edge.getX2() + projection.getX(), edge.getY2() + projection.getY() );
      path.lineTo( edge.getP2() );
      path.closePath();
      return path.getGeneralPath();
    },
    getClock: function() {
      return clock;
    },
    getBrick: function() {
      return brick;
    },
    getIronBlock: function() {
      return ironBlock;
    },
    getLeftBurner: function() {
      return leftBurner;
    },
    getRightBurner: function() {
      return rightBurner;
    },
    getBeaker: function() {
      return beaker;
    },
    getAir: function() {
      return air;
    },

    //private
    findBestSupportSurface: function( element ) {
      var bestOverlappingSurface = null;
      // if this element can go on top of it.
      for ( var potentialSupportingElement in Arrays.asList( leftBurner, rightBurner, brick, ironBlock, beaker ) ) {
        if ( potentialSupportingElement == element || potentialSupportingElement.isStackedUpon( element ) ) {
          // so skip it.
          continue;
        }
        if ( element.getBottomSurfaceProperty().get().overlapsWith( potentialSupportingElement.getTopSurfaceProperty().get() ) ) {
          // is the best one so far.
          var surfaceOverlap = getHorizontalOverlap( potentialSupportingElement.getTopSurfaceProperty().get(), element.getBottomSurfaceProperty().get() );
          // current one.
          if ( bestOverlappingSurface == null || (surfaceOverlap > getHorizontalOverlap( bestOverlappingSurface.get(), element.getBottomSurfaceProperty().get() ) && !isDirectlyAbove( bestOverlappingSurface.get(), potentialSupportingElement.getTopSurfaceProperty().get() )) || (isDirectlyAbove( potentialSupportingElement.getTopSurfaceProperty().get(), bestOverlappingSurface.get() )) ) {
            bestOverlappingSurface = potentialSupportingElement.getTopSurfaceProperty();
          }
        }
      }
      // tested isn't directly above the best surface's center.
      if ( bestOverlappingSurface != null ) {
        while ( bestOverlappingSurface.get().getElementOnSurface() != null ) {
          bestOverlappingSurface = bestOverlappingSurface.get().getElementOnSurface().getTopSurfaceProperty();
        }
      }
      return bestOverlappingSurface;
    },
    getBlockList: function() {
      return Arrays.asList( brick, ironBlock );
    },
// Get the amount of overlap in the x direction between two horizontal surfaces.

    //private
    getHorizontalOverlap: function( s1, s2 ) {
      var lowestMax = Math.min( s1.xRange.getMax(), s2.xRange.getMax() );
      var highestMin = Math.max( s1.xRange.getMin(), s2.xRange.getMin() );
      return Math.max( lowestMax - highestMin, 0 );
    },
// Returns true if surface s1's center is above surface s2.

    //private
    isDirectlyAbove: function( s1, s2 ) {
      return s2.xRange.contains( s1.getCenterX() ) && s1.yPos > s2.yPos;
    },
    /**
     * Get the temperature and color that would be sensed by a thermometer at
     * the provided location.
     *
     * @param location Location to be sensed.
     * @return Composite object with temperature and color at the provided
     *         location.
     */
    getTemperatureAndColorAtLocation: function( location ) {
      var locationAsPoint = location.toPoint2D();
      // must be taken into account.
      var copyOfBlockList = new ArrayList( getBlockList() );
      Collections.sort( copyOfBlockList, new Comparator().withAnonymousClassBody( {
        compare: function( b1, b2 ) {
          if ( b1.position.get().equals( b2.position.get() ) ) {
            return 0;
          }
          if ( b2.position.get().getX() > b1.position.get().getX() || b2.position.get().getY() > b1.position.get().getY() ) {
            return 1;
          }
          return -1;
        }
      } ) );
      for ( var block in copyOfBlockList ) {
        if ( block.getProjectedShape().contains( locationAsPoint ) ) {
          return new TemperatureAndColor( block.getTemperature(), block.getColor() );
        }
      }
      // Test if this point is in the water or steam associated with the beaker.
      if ( beaker.getThermalContactArea().getBounds().contains( locationAsPoint ) ) {
        return new TemperatureAndColor( beaker.getTemperature(), EFACConstants.WATER_COLOR_IN_BEAKER );
      }
      else if ( beaker.getSteamArea().contains( locationAsPoint ) && beaker.steamingProportion > 0 ) {
        return new TemperatureAndColor( beaker.getSteamTemperature( locationAsPoint.getY() - beaker.getSteamArea().getMinY() ), Color.WHITE );
      }
      // Test if the point is a burner.
      for ( var burner in Arrays.asList( leftBurner, rightBurner ) ) {
        if ( burner.getFlameIceRect().contains( locationAsPoint ) ) {
          return new TemperatureAndColor( burner.getTemperature(), EFACConstants.FIRST_TAB_BACKGROUND_COLOR );
        }
      }
      // Point is in nothing else, so return the air temperature.
      return new TemperatureAndColor( air.getTemperature(), EFACConstants.FIRST_TAB_BACKGROUND_COLOR );
    }
  } );
} );

