// Copyright 2002-2012, University of Colorado
/**
 * A class that contains static methods for redistributing a set of energy
 * chunks within a shape.  The basic approach is to simulate a set of
 * particles embedded in a fluid, and each particle repels all others as well
 * as the edges of the containers.
 * <p/>
 * Reuse Notes: This could probably be generalized fairly easily to distribute
 * any number items within a container of arbitrary size in a way that looks
 * pretty random.  Either that, or the code itself could be copied and the
 * various parameters changed as needed.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Rectangle = require( 'KITE/Rectangle' );
  var HashMap = require( 'java.util.HashMap' );
  var List = require( 'java.util.List' );
  var Map = require( 'java.util.Map' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Vector2 = require( 'DOT/Vector2' );
  var DoubleRange = require( 'edu.colorado.phet.common.phetcommon.util.DoubleRange' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );

// In Newtons, empirically determined.

  //private
  var OUTSIDE_CONTAINER_FORCE = 0.01;
// Seeded for greater consistency.

  //private
  var RAND = new Random( 2 );

  //private
  var ZERO_VECTOR = new Vector2( 0, 0 );
// Parameters that can be adjusted to change they nature of the redistribution.
// In seconds, for algorithm that moves the points.

  //private
  var MAX_TIME_STEP = 5E-3;
// In kilograms, chosen arbitrarily.

  //private
  var ENERGY_CHUNK_MASS = 1E-3;
// In kg / m ^ 3, same as water, used for drag.

  //private
  var FLUID_DENSITY = 1000;
// In meters, chosen empirically.

  //private
  var ENERGY_CHUNK_DIAMETER = 1E-3;
// Treat energy chunk as if it is shaped like a sphere.

  //private
  var ENERGY_CHUNK_CROSS_SECTIONAL_AREA = Math.PI * Math.pow( ENERGY_CHUNK_DIAMETER, 2 );
// Unitless, empirically chosen.

  //private
  var DRAG_COEFFICIENT = 500;
// Thresholds for deciding whether or not to perform redistribution.
// These value should be chosen such that particles spread out, then stop
// all movement.
// In joules, I think.

  //private
  var REDISTRIBUTION_THRESHOLD_ENERGY = 1E-4;

  return inherit( Object, EnergyChunkDistributor, {
    /**
     * Redistribute a set of energy chunks that are contained in energy chunk
     * "slices".  This is done in this way because all of the energy chunks in
     * a set of slices interact with each other, but the container for each is
     * defined by the boundary of its containing slice.
     *
     * @param energyChunkContainerSlices Set of slices, each containing a set of energy chunks.
     * @param dt                         Delta time
     */
    updatePositions: function( energyChunkContainerSlices, dt ) {
      // Determine a rectangle that bounds all of the slices.
      var boundingRect;
      {
        var minX = Number.POSITIVE_INFINITY;
        var minY = Number.POSITIVE_INFINITY;
        var maxX = Number.NEGATIVE_INFINITY;
        var maxY = Number.NEGATIVE_INFINITY;
        for ( var slice in energyChunkContainerSlices ) {
          minX = Math.min( slice.getShape().getBounds2D().getMinX(), minX );
          maxX = Math.max( slice.getShape().getBounds2D().getMaxX(), maxX );
          minY = Math.min( slice.getShape().getBounds2D().getMinY(), minY );
          maxY = Math.max( slice.getShape().getBounds2D().getMaxY(), maxY );
        }
        boundingRect = new Rectangle.Number( minX, minY, maxX - minX, maxY - minY );
      }
      // Create a map that tracks the force applied to each energy chunk.
      var mapEnergyChunkToForceVector = new HashMap();
      for ( var energyChunkContainerSlice in energyChunkContainerSlices ) {
        for ( var ec in energyChunkContainerSlice.energyChunkList ) {
          mapEnergyChunkToForceVector.put( ec, ZERO_VECTOR );
        }
      }
      // Make sure that there is actually something to distribute.
      if ( mapEnergyChunkToForceVector.isEmpty() ) {
        // Nothing to do - abort.
        return false;
      }
      // Divisor empirically determined.
      var minDistance = Math.min( boundingRect.getWidth(), boundingRect.getHeight() ) / 20;
      // made up, and can be adjusted if needed.
      var forceConstant = ENERGY_CHUNK_MASS * boundingRect.getWidth() * boundingRect.getHeight() * 0.1 / mapEnergyChunkToForceVector.size();
      // Loop once for each max time step plus any remainder.
      var particlesRedistributed = false;
      var numForceCalcSteps = (dt / MAX_TIME_STEP);
      var extraTime = dt - numForceCalcSteps * MAX_TIME_STEP;
      for ( var forceCalcStep = 0; forceCalcStep <= numForceCalcSteps; forceCalcStep++ ) {
        var timeStep = forceCalcStep < numForceCalcSteps ? MAX_TIME_STEP : extraTime;
        // container, other particles, and drag.
        for ( var energyChunkContainerSlice in energyChunkContainerSlices ) {
          var containerShape = energyChunkContainerSlice.getShape();
          // Determine the max possible distance to an edge.
          var maxDistanceToEdge = Math.sqrt( Math.pow( containerShape.getBounds2D().getWidth(), 2 ) + Math.pow( containerShape.getBounds2D().getHeight(), 2 ) );
          // Determine forces on each energy chunk.
          for ( var ec in energyChunkContainerSlice.energyChunkList ) {
            // Reset accumulated forces.
            mapEnergyChunkToForceVector.put( ec, ZERO_VECTOR );
            if ( containerShape.contains( ec.position.get().toPoint2D() ) ) {
              // edges at the given angle.
              for ( var angle = 0; angle < 2 * Math.PI; angle += Math.PI / 2 ) {
                var edgeDetectSteps = 8;
                var lengthBounds = new DoubleRange( 0, maxDistanceToEdge );
                for ( var edgeDetectStep = 0; edgeDetectStep < edgeDetectSteps; edgeDetectStep++ ) {
                  var vectorToEdge = new Vector2( lengthBounds.getCenter(), 0 ).getRotatedInstance( angle );
                  if ( containerShape.contains( ec.position.get().plus( vectorToEdge ).toPoint2D() ) ) {
                    lengthBounds = new DoubleRange( lengthBounds.getCenter(), lengthBounds.getMax() );
                  }
                  else {
                    lengthBounds = new DoubleRange( lengthBounds.getMin(), lengthBounds.getCenter() );
                  }
                }
                // Handle case where point is too close to the container's edge.
                if ( lengthBounds.getCenter() < minDistance ) {
                  lengthBounds = new DoubleRange( minDistance, minDistance );
                }
                // Apply the force due to this edge.
                var edgeForce = new Vector2( forceConstant / Math.pow( lengthBounds.getCenter(), 2 ), 0 ).getRotatedInstance( angle + Math.PI );
                mapEnergyChunkToForceVector.put( ec, mapEnergyChunkToForceVector.get( ec ).plus( edgeForce ) );
              }
              // that can be applied.
              for ( var otherEnergyChunk in mapEnergyChunkToForceVector.keySet() ) {
                if ( ec == otherEnergyChunk ) {
                  continue;
                }
                // Calculate force vector, but handle cases where too close.
                var vectorToOther = ec.position.get().minus( otherEnergyChunk.position.get() );
                if ( vectorToOther.magnitude() < minDistance ) {
                  if ( vectorToOther.magnitude() == 0 ) {
                    // Create a random vector of min distance.
                    var randomAngle = RAND.nextDouble() * Math.PI * 2;
                    vectorToOther = new Vector2( minDistance * Math.cos( randomAngle ), minDistance * Math.sin( randomAngle ) );
                  }
                  else {
                    vectorToOther = vectorToOther.getInstanceOfMagnitude( minDistance );
                  }
                }
                // Add the force to the accumulated forces on this energy chunk.
                mapEnergyChunkToForceVector.put( ec, mapEnergyChunkToForceVector.get( ec ).plus( vectorToOther.getInstanceOfMagnitude( forceConstant / (vectorToOther.magnitudeSquared()) ) ) );
              }
            }
            else {
              // Point is outside container, move it towards center of shape.
              var vectorToCenter = new Vector2( boundingRect.getCenterX(), boundingRect.getCenterY() ).minus( ec.position.get() );
              mapEnergyChunkToForceVector.put( ec, vectorToCenter.getInstanceOfMagnitude( OUTSIDE_CONTAINER_FORCE ) );
            }
          }
        }
        // Update energy chunk velocities, drag force, and position.
        var maxEnergy = 0;
        for ( var energyChunk in mapEnergyChunkToForceVector.keySet() ) {
          // Calculate the energy chunk's velocity as a result of forces acting on it.
          var forceOnThisChunk = mapEnergyChunkToForceVector.get( energyChunk );
          var newVelocity = energyChunk.getVelocity().plus( forceOnThisChunk.times( timeStep / ENERGY_CHUNK_MASS ) );
          // Calculate drag force.  Uses standard drag equation.
          var dragMagnitude = 0.5 * FLUID_DENSITY * DRAG_COEFFICIENT * ENERGY_CHUNK_CROSS_SECTIONAL_AREA * newVelocity.magnitudeSquared();
          var dragForceVector = dragMagnitude > 0 ? newVelocity.getRotatedInstance( Math.PI ).getInstanceOfMagnitude( dragMagnitude ) : ZERO_VECTOR;
          // Update velocity based on drag force.
          newVelocity = newVelocity.plus( dragForceVector.times( timeStep / ENERGY_CHUNK_MASS ) );
          energyChunk.setVelocity( newVelocity );
          // Update max energy.
          var totalParticleEnergy = 0.5 * ENERGY_CHUNK_MASS * newVelocity.magnitudeSquared() + forceOnThisChunk.magnitude() * Math.PI / 2;
          if ( totalParticleEnergy > maxEnergy ) {
            maxEnergy = totalParticleEnergy;
          }
        }
        particlesRedistributed = maxEnergy > REDISTRIBUTION_THRESHOLD_ENERGY;
        if ( particlesRedistributed ) {
          for ( var energyChunk in mapEnergyChunkToForceVector.keySet() ) {
            // Update position.
            energyChunk.position.set( energyChunk.position.get().plus( energyChunk.getVelocity().times( timeStep ) ) );
          }
        }
      }
      return particlesRedistributed;
    },
    /*
     * Super simple alternative energy chunk distribution algorithm - just puts
     * all energy chunks in center of slide.  This is useful for debugging.
     * Rename it to substitute if for the 'real' algorithm.
     */
    updatePositionsDbg: function( energyChunkContainerSlices, dt ) {
      // Update the positions of the energy chunks.
      for ( var slice in energyChunkContainerSlices ) {
        var sliceCenter = new Vector2( slice.getShape().getBounds2D().getCenterX(), slice.getShape().getBounds2D().getCenterY() );
        for ( var energyChunk in slice.energyChunkList ) {
          energyChunk.position.set( sliceCenter );
        }
      }
    },
    generateRandomLocation: function( rect ) {
      return new Vector2( rect.getMinX() + (RAND.nextDouble() * rect.getWidth()), rect.getMinY() + (RAND.nextDouble() * rect.getHeight()) );
    }
  } );
} );

