// Copyright 2002-2012, University of Colorado
/**
 * This class is used to make an energy chunk wander, i.e. to perform somewhat
 * of a random walk while moving towards a destination.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Random = require( 'java.util.Random' );
  var Vector2 = require( 'DOT/Vector2' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// In m/s.

  //private
  var MIN_VELOCITY = 0.06;
// In m/s.

  //private
  var MAX_VELOCITY = 0.10;

  //private
  var RAND = new Random();

  //private
  var MIN_TIME_IN_ONE_DIRECTION = 0.4;

  //private
  var MAX_TIME_IN_ONE_DIRECTION = 0.8;
// In meters, empirically chosen.

  //private
  var DISTANCE_AT_WHICH_TO_STOP_WANDERING = 0.05;
// Max deviation from angle to destination, in radians, empirically chosen.

  //private
  var MAX_ANGLE_VARIATION = Math.PI * 0.2;

  function EnergyChunkWanderController( energyChunk, destination ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.energyChunk;

    //private
    this.destination;

    //private
    this.velocity = new Vector2( 0, MAX_VELOCITY );

    //private
    this.countdownTimer = 0;

    //private
    this.initialWanderConstraint;
    this( energyChunk, destination, null );
  }

  function EnergyChunkWanderController( energyChunk, destination, initialWanderConstraint ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.energyChunk;

    //private
    this.destination;

    //private
    this.velocity = new Vector2( 0, MAX_VELOCITY );

    //private
    this.countdownTimer = 0;

    //private
    this.initialWanderConstraint;
    this.energyChunk = energyChunk;
    this.destination = destination;
    this.initialWanderConstraint = initialWanderConstraint;
    resetCountdownTimer();
    changeVelocityVector();
  }

  return inherit( Object, EnergyChunkWanderController, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    updatePosition: function( dt ) {
      var distanceToDestination = energyChunk.position.get().distance( destination.get() );
      if ( distanceToDestination < velocity.magnitude() * dt && !energyChunk.position.get().equals( destination.get() ) ) {
        // Destination reached.
        energyChunk.position.set( destination.get() );
        velocity.setMagnitude( 0 );
      }
      else if ( energyChunk.position.get().distance( destination.get() ) < dt * velocity.magnitude() ) {
        // Prevent overshoot.
        velocity.times( energyChunk.position.get().distance( destination.get() ) * dt );
      }
      // Stay within the horizontal confines of the initial bounds.
      if ( initialWanderConstraint != null && energyChunk.position.get().getY() < initialWanderConstraint.getMaxY() ) {
        var proposedPosition = energyChunk.position.get().plus( velocity.times( dt ) );
        if ( proposedPosition.getX() < initialWanderConstraint.getMinX() || proposedPosition.getX() > initialWanderConstraint.getMaxX() ) {
          // Bounce in the x direction to prevent going outside initial bounds.
          velocity.setComponents( -velocity.getX(), velocity.getY() );
        }
      }
      energyChunk.position.set( energyChunk.position.get().plus( velocity.times( dt ) ) );
      countdownTimer -= dt;
      if ( countdownTimer <= 0 ) {
        changeVelocityVector();
        resetCountdownTimer();
      }
    },

    //private
    changeVelocityVector: function() {
      var vectorToDestination = destination.get().minus( energyChunk.position.get() );
      var angle = vectorToDestination.getAngle();
      if ( vectorToDestination.magnitude() > DISTANCE_AT_WHICH_TO_STOP_WANDERING ) {
        // Add some randomness to the direction of travel.
        angle = angle + (RAND.nextDouble() - 0.5) * 2 * MAX_ANGLE_VARIATION;
      }
      var scalarVelocity = MIN_VELOCITY + (MAX_VELOCITY - MIN_VELOCITY) * RAND.nextDouble();
      velocity.setComponents( scalarVelocity * Math.cos( angle ), scalarVelocity * Math.sin( angle ) );
    },

    //private
    resetCountdownTimer: function() {
      countdownTimer = MIN_TIME_IN_ONE_DIRECTION + (MAX_TIME_IN_ONE_DIRECTION - MIN_TIME_IN_ONE_DIRECTION) * RAND.nextDouble();
    },
    getEnergyChunk: function() {
      return energyChunk;
    },
    destinationReached: function() {
      return destination.get().distance( energyChunk.position.get() ) < 1E-7;
    }
  } );
} );

