// Copyright 2002-2012, University of Colorado
/**
 * Class containing the constants that control the rate of heat transfer
 * between the various model elements that can contain heat, as well as methods
 * for obtaining the heat transfer value for any two model elements that are
 * capable of exchanging heat.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var HashMap = require( 'java.util.HashMap' );
  var Map = require( 'java.util.Map' );
  var Property = require( 'AXON/Property' );

// Heat transfer values.  NOTE: Originally, these were constants, but the
// design team requested that they be changeable via a developer control,
// which is why they are now properties.
  var BRICK_IRON_HEAT_TRANSFER_FACTOR = new Property( 1000.0 );
  var BRICK_WATER_HEAT_TRANSFER_FACTOR = new Property( 1000.0 );
  var BRICK_AIR_HEAT_TRANSFER_FACTOR = new Property( 50.0 );
  var IRON_WATER_HEAT_TRANSFER_FACTOR = new Property( 1000.0 );
  var IRON_AIR_HEAT_TRANSFER_FACTOR = new Property( 50.0 );
  var WATER_AIR_HEAT_TRANSFER_FACTOR = new Property( 50.0 );
  var AIR_TO_SURROUNDING_AIR_HEAT_TRANSFER_FACTOR = new Property( 10000.0 );
// Maps for obtaining transfer constants for a given thermal element.

  //private
  var HEAT_TRANSFER_FACTORS_FOR_BRICK = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyContainerCategory.IRON, BRICK_IRON_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.WATER, BRICK_WATER_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.AIR, BRICK_AIR_HEAT_TRANSFER_FACTOR );
    }
  } );

  //private
  var HEAT_TRANSFER_FACTORS_FOR_IRON = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyContainerCategory.BRICK, BRICK_IRON_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.WATER, BRICK_WATER_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.AIR, BRICK_AIR_HEAT_TRANSFER_FACTOR );
    }
  } );

  //private
  var HEAT_TRANSFER_FACTORS_FOR_WATER = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyContainerCategory.BRICK, BRICK_WATER_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.IRON, IRON_WATER_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.AIR, WATER_AIR_HEAT_TRANSFER_FACTOR );
    }
  } );

  //private
  var HEAT_TRANSFER_FACTORS_FOR_AIR = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyContainerCategory.BRICK, BRICK_AIR_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.IRON, IRON_AIR_HEAT_TRANSFER_FACTOR );
      put( EnergyContainerCategory.WATER, WATER_AIR_HEAT_TRANSFER_FACTOR );
    }
  } );

  //private
  var CONTAINER_CATEGORY_MAP = new HashMap().withAnonymousClassBody( {
    initializer: function() {
      put( EnergyContainerCategory.BRICK, HEAT_TRANSFER_FACTORS_FOR_BRICK );
      put( EnergyContainerCategory.IRON, HEAT_TRANSFER_FACTORS_FOR_IRON );
      put( EnergyContainerCategory.WATER, HEAT_TRANSFER_FACTORS_FOR_WATER );
      put( EnergyContainerCategory.AIR, HEAT_TRANSFER_FACTORS_FOR_AIR );
    }
  } );

  return inherit( Object, HeatTransferConstants, {
      getHeatTransferFactor: function( container1, container2 ) {
        return CONTAINER_CATEGORY_MAP.get( container1 ).get( container2 ).get();
      }
    },
//statics
    {
      BRICK_IRON_HEAT_TRANSFER_FACTOR: BRICK_IRON_HEAT_TRANSFER_FACTOR,
      BRICK_WATER_HEAT_TRANSFER_FACTOR: BRICK_WATER_HEAT_TRANSFER_FACTOR,
      BRICK_AIR_HEAT_TRANSFER_FACTOR: BRICK_AIR_HEAT_TRANSFER_FACTOR,
      IRON_WATER_HEAT_TRANSFER_FACTOR: IRON_WATER_HEAT_TRANSFER_FACTOR,
      IRON_AIR_HEAT_TRANSFER_FACTOR: IRON_AIR_HEAT_TRANSFER_FACTOR,
      WATER_AIR_HEAT_TRANSFER_FACTOR: WATER_AIR_HEAT_TRANSFER_FACTOR,
      AIR_TO_SURROUNDING_AIR_HEAT_TRANSFER_FACTOR: AIR_TO_SURROUNDING_AIR_HEAT_TRANSFER_FACTOR
    } );
} );

