// Copyright 2002-2012, University of Colorado
/**
 * A movable model element that contains thermal energy and that, at least in
 * the model, has an overall shape that can be represented as a rectangle.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Shape = require( 'java.awt.Shape' );
  var Rectangle = require( 'KITE/Rectangle' );
  var ArrayList = require( 'java.util.ArrayList' );
  var List = require( 'java.util.List' );
  var Vector2 = require( 'DOT/Vector2' );
  var ClockAdapter = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockAdapter' );
  var ClockEvent = require( 'edu.colorado.phet.common.phetcommon.model.clock.ClockEvent' );
  var IClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.IClock' );
  var Property = require( 'AXON/Property' );
  var ObservableArray = require( 'AXON/ObservableArray' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyType = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyType' );
  var ENERGY_TO_NUM_CHUNKS_MAPPER = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants/ENERGY_TO_NUM_CHUNKS_MAPPER' );//static
  var MAX_HEAT_EXCHANGE_TIME_STEP = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants/MAX_HEAT_EXCHANGE_TIME_STEP' );//static
  var getHeatTransferFactor = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/HeatTransferConstants/getHeatTransferFactor' );//static

  /*
   * Constructor.
   *
   * @param clock
   */
  function RectangularThermalMovableModelElement( clock, initialPosition, width, height, mass, specificHeat, energyChunksVisible ) {
    this.energyChunksVisible;
    // In Joules.
    this.energy = 0;
    // In J/kg-K
    this.specificHeat;
    // In kg
    this.mass;
    this.width;
    this.height;

    //private
    this.nextSliceIndex;
    // 2D "slices" of the container, used for 3D layering of energy chunks.
    this.slices = [];
    // Energy chunks that are approaching this model element.
    this.approachingEnergyChunks = new ObservableArray();
    this.energyChunkWanderControllers = [];
    UserMovableModelElement.call( this, initialPosition );
    this.mass = mass;
    this.width = width;
    this.height = height;
    this.specificHeat = specificHeat;
    this.energyChunksVisible = energyChunksVisible;
    energy = mass * specificHeat * EFACConstants.ROOM_TEMPERATURE;
    // Hook up to the clock for time dependent behavior.
    clock.addClockListener( new ClockAdapter().withAnonymousClassBody( {
      clockTicked: function( clockEvent ) {
        stepInTime( clockEvent.getSimulationTimeChange() );
      }
    } ) );
    // Add the slices, a.k.a. layers, where the energy chunks will live.
    addEnergyChunkSlices();
    nextSliceIndex = slices.size() / 2;
    // Add the initial energy chunks.
    addInitialEnergyChunks();
  }

  return inherit( UserMovableModelElement, RectangularThermalMovableModelElement, {
    /**
     * Get the rectangle that defines this elements position and shape in
     * model space.
     */
    getRect: function() {},
    changeEnergy: function( deltaEnergy ) {
      energy += deltaEnergy;
    },
    getEnergy: function() {
      return energy;
    },
    getTemperature: function() {
      return energy / (mass * specificHeat);
    },
    reset: function() {
      super.reset();
      energy = mass * specificHeat * EFACConstants.ROOM_TEMPERATURE;
      addInitialEnergyChunks();
    },
    stepInTime: function( dt ) {
      // Distribute the energy chunks contained within this model element.
      EnergyChunkDistributor.updatePositions( slices, dt );
      // Animate the energy chunks that are outside this model element.
      animateNonContainedEnergyChunks( dt );
    },
    animateNonContainedEnergyChunks: function( dt ) {
      for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
        energyChunkWanderController.updatePosition( dt );
        if ( getSliceBounds().contains( energyChunkWanderController.getEnergyChunk().position.get().toPoint2D() ) ) {
          moveEnergyChunkToSlices( energyChunkWanderController.getEnergyChunk() );
        }
      }
    },
    /**
     * Add an energy chunk to this model element.  The energy chunk can be
     * outside of the element's rectangular bounds, in which case it is added
     * to the list of chunks that are moving towards the element, or it can be
     * positioned already inside, in which case it is immediately added to one
     * of the energy chunk "slices".
     *
     * @param ec Energy chunk to add.
     */
    addEnergyChunk: function( ec ) {
      if ( getSliceBounds().contains( ec.position.get().toPoint2D() ) ) {
        // directly to a slice.
        addEnergyChunkToNextSlice( ec );
      }
      else {
        // towards it.
        ec.zPosition.set( 0.0 );
        approachingEnergyChunks.add( ec );
        energyChunkWanderControllers.add( new EnergyChunkWanderController( ec, position ) );
      }
    },
// Add an energy chunk to the next available slice.  Override for more elaborate behavior.
    addEnergyChunkToNextSlice: function( ec ) {
      slices.get( nextSliceIndex ).addEnergyChunk( ec );
      nextSliceIndex = (nextSliceIndex + 1) % slices.size();
    },
    getSliceBounds: function() {
      var minX = Number.POSITIVE_INFINITY;
      var minY = Number.POSITIVE_INFINITY;
      var maxX = Number.NEGATIVE_INFINITY;
      var maxY = Number.NEGATIVE_INFINITY;
      for ( var slice in slices ) {
        var sliceBounds = slice.getShape().getBounds2D();
        if ( sliceBounds.getMinX() < minX ) {
          minX = sliceBounds.getMinX();
        }
        if ( sliceBounds.getMaxX() > maxX ) {
          maxX = sliceBounds.getMaxX();
        }
        if ( sliceBounds.getMinY() < minY ) {
          minY = sliceBounds.getMinY();
        }
        if ( sliceBounds.getMaxY() > maxY ) {
          maxY = sliceBounds.getMaxY();
        }
      }
      return new Rectangle.Number( minX, minY, maxX - minX, maxY - minY );
    },
    moveEnergyChunkToSlices: function( ec ) {
      approachingEnergyChunks.remove( ec );
      for ( var energyChunkWanderController in new ArrayList( energyChunkWanderControllers ) ) {
        if ( energyChunkWanderController.getEnergyChunk() == ec ) {
          energyChunkWanderControllers.remove( energyChunkWanderController );
        }
      }
      addEnergyChunkToNextSlice( ec );
    },
    removeEnergyChunk: function( ec ) {
      for ( var slice in slices ) {
        if ( slice.energyChunkList.remove( ec ) ) {
          return true;
        }
      }
      return false;
    },
    /*
     * Extract the closest energy chunk to the provided point.  Compensate
     * distances for the z-offset so that z-positioning doesn't skew the
     * results, since the provided point is only 2D.
     *
     * @param point Comparison point.
     * @return Energy chunk, null if there are none available.
     */
    extractClosestEnergyChunk: function( point ) {
      var closestEnergyChunk = null;
      var closestCompensatedDistance = Number.POSITIVE_INFINITY;
      // Identify the closest energy chunk.
      for ( var slice in slices ) {
        for ( var ec in slice.energyChunkList ) {
          // almost always be chosen.
          var compensatedEcPosition = ec.position.get().minus( 0, EFACConstants.Z_TO_Y_OFFSET_MULTIPLIER * ec.zPosition.get() );
          var compensatedDistance = compensatedEcPosition.distance( point );
          if ( compensatedDistance < closestCompensatedDistance ) {
            closestEnergyChunk = ec;
          }
        }
      }
      removeEnergyChunk( closestEnergyChunk );
      return closestEnergyChunk;
    },
    /*
     * Extract an energy chunk that is a good choice for being transferred to
     * the provided shape.  Generally, this means that it is close to the
     * shape.  This routine is not hugely general - it makes some assumptions
     * that make it work for blocks in beakers.  If support for other shapes is
     * needed, it will need some work.
     *
     * @param destinationShape
     * @return Energy chunk, null if none are available.
     */
    extractClosestEnergyChunk: function( destinationShape ) {
      var chunkToExtract = null;
      var myBounds = getSliceBounds();
      if ( destinationShape.contains( getThermalContactArea().getBounds() ) ) {
        // our right or left edge.
        var closestDistanceToVerticalEdge = Number.POSITIVE_INFINITY;
        for ( var slice in slices ) {
          for ( var ec in slice.energyChunkList ) {
            var distanceToVerticalEdge = Math.min( Math.abs( myBounds.getMinX() - ec.position.get().getX() ), Math.abs( myBounds.getMaxX() - ec.position.get().getX() ) );
            if ( distanceToVerticalEdge < closestDistanceToVerticalEdge ) {
              chunkToExtract = ec;
              closestDistanceToVerticalEdge = distanceToVerticalEdge;
            }
          }
        }
      }
      else if ( getThermalContactArea().getBounds().contains( destinationShape.getBounds2D() ) ) {
        // is close but doesn't overlap with the destination shape.
        var closestDistanceToDestinationEdge = Number.POSITIVE_INFINITY;
        var destinationBounds = destinationShape.getBounds2D();
        for ( var slice in slices ) {
          for ( var ec in slice.energyChunkList ) {
            var distanceToDestinationEdge = Math.min( Math.abs( destinationBounds.getMinX() - ec.position.get().getX() ), Math.abs( destinationBounds.getMaxX() - ec.position.get().getX() ) );
            if ( !destinationShape.contains( ec.position.get().toPoint2D() ) && distanceToDestinationEdge < closestDistanceToDestinationEdge ) {
              chunkToExtract = ec;
              closestDistanceToDestinationEdge = distanceToDestinationEdge;
            }
          }
        }
      }
      else {
        // There is no or limited overlap, so use center points.
        chunkToExtract = extractClosestEnergyChunk( new Vector2( destinationShape.getBounds2D().getCenterX(), destinationShape.getBounds2D().getCenterY() ) );
      }
      // Fail safe - If nothing found, get the first chunk.
      if ( chunkToExtract == null ) {
        console.log( getClass().getName() + " - Warning: No energy chunk found by extraction algorithm, trying first available.." );
        for ( var slice in slices ) {
          if ( slice.energyChunkList.size() > 0 ) {
            chunkToExtract = slice.energyChunkList.get( 0 );
            break;
          }
        }
        if ( chunkToExtract == null ) {
          console.log( getClass().getName() + " - Warning: No chunks available for extraction." );
        }
      }
      removeEnergyChunk( chunkToExtract );
      return chunkToExtract;
    },
    /**
     * Initialization method that add the "slices" where the energy chunks
     * reside.  Should be called only once at initialization.
     */
    addEnergyChunkSlices: function() {
      // Make sure this method isn't being misused.
      assert && assert( (slices.size() == 0) );
      // for more sophisticated behavior.
      slices.add( new EnergyChunkContainerSlice( getRect(), 0, position ) );
    },
    addInitialEnergyChunks: function() {
      for ( var slice in slices ) {
        slice.energyChunkList.clear();
      }
      var targetNumChunks = ENERGY_TO_NUM_CHUNKS_MAPPER.apply( energy );
      var energyChunkBounds = getThermalContactArea().getBounds();
      while ( getNumEnergyChunks() < targetNumChunks ) {
        // Add a chunk at a random location in the block.
        addEnergyChunk( new EnergyChunk( EnergyType.THERMAL, EnergyChunkDistributor.generateRandomLocation( energyChunkBounds ), energyChunksVisible ) );
      }
      // Distribute the energy chunks within the container.
      for ( var i = 0; i < 1000; i++ ) {
        if ( !EnergyChunkDistributor.updatePositions( slices, EFACConstants.SIM_TIME_PER_TICK_NORMAL ) ) {
          break;
        }
      }
    },
    getNumEnergyChunks: function() {
      var numChunks = 0;
      for ( var slice in slices ) {
        numChunks += slice.getNumEnergyChunks();
      }
      return numChunks + approachingEnergyChunks.size();
    },
    getSlices: function() {
      return slices;
    },
    exchangeEnergyWith: function( otherEnergyContainer, dt ) {
      var thermalContactLength = getThermalContactArea().getThermalContactLength( otherEnergyContainer.getThermalContactArea() );
      if ( thermalContactLength > 0 ) {
        if ( Math.abs( otherEnergyContainer.getTemperature() - getTemperature() ) > EFACConstants.TEMPERATURES_EQUAL_THRESHOLD ) {
          // Exchange energy between this and the other energy container.
          var heatTransferConstant = getHeatTransferFactor( this.getEnergyContainerCategory(), otherEnergyContainer.getEnergyContainerCategory() );
          var numFullTimeStepExchanges = Math.floor( dt / MAX_HEAT_EXCHANGE_TIME_STEP );
          var leftoverTime = dt - (numFullTimeStepExchanges * MAX_HEAT_EXCHANGE_TIME_STEP);
          for ( var i = 0; i < numFullTimeStepExchanges + 1; i++ ) {
            var timeStep = i < numFullTimeStepExchanges ? MAX_HEAT_EXCHANGE_TIME_STEP : leftoverTime;
            var thermalEnergyGained = (otherEnergyContainer.getTemperature() - getTemperature()) * thermalContactLength * heatTransferConstant * timeStep;
            otherEnergyContainer.changeEnergy( -thermalEnergyGained );
            changeEnergy( thermalEnergyGained );
          }
        }
      }
    },
    /*
     * Get the shape as is is projected into 3D in the view.  Ideally, this
     * wouldn't even be in the model, because it would be purely handled in the
     * view, but it proved necessary.
     */
    getProjectedShape: function() {
      // This projects a rectangle, override for other behavior.
      var forwardPerspectiveOffset = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( Block.SURFACE_WIDTH / 2 );
      var backwardPerspectiveOffset = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( -Block.SURFACE_WIDTH / 2 );
      var path = new DoubleGeneralPath();
      var rect = getRect();
      path.moveTo( new Vector2( rect.getX(), rect.getY() ).plus( forwardPerspectiveOffset ) );
      path.lineTo( new Vector2( rect.getMaxX(), rect.getY() ).plus( forwardPerspectiveOffset ) );
      path.lineTo( new Vector2( rect.getMaxX(), rect.getY() ).plus( backwardPerspectiveOffset ) );
      path.lineTo( new Vector2( rect.getMaxX(), rect.getMaxY() ).plus( backwardPerspectiveOffset ) );
      path.lineTo( new Vector2( rect.getMinX(), rect.getMaxY() ).plus( backwardPerspectiveOffset ) );
      path.lineTo( new Vector2( rect.getMinX(), rect.getMaxY() ).plus( forwardPerspectiveOffset ) );
      path.closePath();
      return path.getGeneralPath();
    },
    getCenterPoint: function() {
      return new Vector2( position.get().getX(), position.get().getY() + height / 2 );
    },
    /**
     * Get a number indicating the balance between the energy level and the
     * number of energy chunks owned by this model element.
     *
     * @return 0 if the number of energy chunks matches the energy level, a
     *         negative value if there is a deficit, and a positive value if there is
     *         a surplus.
     */
    getEnergyChunkBalance: function() {
      return getNumEnergyChunks() - EFACConstants.ENERGY_TO_NUM_CHUNKS_MAPPER.apply( energy );
    }
  } );
} );

