// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a 2D space that can come into contact with other
 * thermal areas.  This is basically just a shape and a flag that indicates
 * whether immersion can occur (such as when
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Rectangle = require( 'KITE/Rectangle' );

// Threshold of distance for determining whether two areas are in contact.
// In meters.

  //private
  var TOUCH_DISTANCE_THRESHOLD = 0.001;

  function ThermalContactArea( bounds, supportsImmersion ) {

    //private
    this.bounds = new Rectangle.Number();

    //private
    this.supportsImmersion;
    this.bounds.setFrame( bounds );
    this.supportsImmersion = supportsImmersion;
  }

  return inherit( Object, ThermalContactArea, {
    getBounds: function() {
      return bounds;
    },
    /**
     * Get the amount of thermal contact that exists between this and another
     * thermal area.  Since thermal contact areas are 2D, the amount of
     * contact is a 1D quantity.  For example, when a rectangle is sitting
     * on top of another that is the same width, the contact length is the
     * width of the shared edge.
     *
     * @param that Other thermal contact area.
     * @return Length of contact
     */
    getThermalContactLength: function( that ) {
      var xOverlap = getHorizontalOverlap( this.bounds, that.bounds );
      var yOverlap = getVerticalOverlap( this.bounds, that.bounds );
      var contactLength = 0;
      if ( xOverlap > 0 && yOverlap > 0 ) {
        // the other.
        if ( this.supportsImmersion || that.supportsImmersion ) {
          var immersionRect = this.bounds.createIntersection( that.bounds );
          contactLength = immersionRect.getWidth() * 2 + immersionRect.getHeight() * 2;
          if ( immersionRect.getWidth() != this.bounds.getWidth() && immersionRect.getWidth() != that.bounds.getWidth() ) {
            // Not fully overlapping in X direction, so adjust contact length accordingly.
            contactLength -= immersionRect.getHeight();
          }
          if ( immersionRect.getHeight() != this.bounds.getHeight() && immersionRect.getHeight() != that.bounds.getHeight() ) {
            // Not fully overlapping in Y direction, so adjust contact length accordingly.
            contactLength -= immersionRect.getWidth();
          }
        }
        else {
          // by testing.
          if ( yOverlap > 1E-6 && xOverlap > 1E-6 ) {
            console.log( getClass().getName() + " - Error: Double overlap detected in case where neither energy container supports immersion.  Ignoring." );
            console.log( "yOverlap = " + yOverlap );
            console.log( "xOverlap = " + xOverlap );
          }
        }
      }
      else if ( xOverlap > 0 || yOverlap > 0 ) {
        // see if the two containers are touching.
        if ( xOverlap > 0 && Math.abs( this.bounds.getMaxY() - that.bounds.getMinY() ) < TOUCH_DISTANCE_THRESHOLD || Math.abs( this.bounds.getMinY() - that.bounds.getMaxY() ) < TOUCH_DISTANCE_THRESHOLD ) {
          contactLength = xOverlap;
        }
        else if ( yOverlap > 0 && Math.abs( this.bounds.getMaxX() - that.bounds.getMinX() ) < TOUCH_DISTANCE_THRESHOLD || Math.abs( this.bounds.getMinX() - that.bounds.getMaxX() ) < TOUCH_DISTANCE_THRESHOLD ) {
          contactLength = xOverlap;
        }
      }
      return contactLength;
    },
// Convenience method for determining overlap of rectangles in X dimension.

    //private
    getHorizontalOverlap: function( r1, r2 ) {
      var lowestMax = Math.min( r1.getMaxX(), r2.getMaxX() );
      var highestMin = Math.max( r1.getMinX(), r2.getMinX() );
      return Math.max( lowestMax - highestMin, 0 );
    },
// Convenience method for determining overlap of rectangles in X dimension.

    //private
    getVerticalOverlap: function( r1, r2 ) {
      var lowestMax = Math.min( r1.getMaxY(), r2.getMaxY() );
      var highestMin = Math.max( r1.getMinY(), r2.getMinY() );
      return Math.max( lowestMax - highestMin, 0 );
    }
  } );
} );

