// Copyright 2002-2012, University of Colorado
/**
 * Class that represents a "beaker container" in the view.  A beaker container
 * is a beaker that contains fluid, and in which other objects can be placed,
 * generally displacing the fluid.
 * <p/>
 * See the header comments in the parent class for some important information
 * about how this class is used on the canvas.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Area = require( 'java.awt.geom.Area' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Vector2 = require( 'DOT/Vector2' );
  var IClock = require( 'edu.colorado.phet.common.phetcommon.model.clock.IClock' );
  var SimpleObserver = require( 'edu.colorado.phet.common.phetcommon.util.SimpleObserver' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var CursorHandler = require( 'edu.colorado.phet.common.piccolophet.event.CursorHandler' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var Beaker = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/Beaker' );
  var BeakerView = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/BeakerView' );
  var Block = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/Block' );
  var EFACIntroModel = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EFACIntroModel' );
  var PClip = require( 'edu.umd.cs.piccolox.nodes.PClip' );

  function BeakerContainerView( clock, model, mvt ) {
    BeakerView.call( this, clock, model.getBeaker(), model.energyChunksVisible, mvt );
    // mask hides energy chunks that overlap with blocks.
    for ( var block in model.getBlockList() ) {
      block.position.addObserver( new SimpleObserver().withAnonymousClassBody( {
        update: function() {
          updateEnergyChunkClipMask( model, energyChunkClipNode );
        }
      } ) );
    }
    var beaker = model.getBeaker();
    model.getBeaker().position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( position ) {
        BeakerContainerView.this.updateEnergyChunkClipMask( model, energyChunkClipNode );
      }
    } ) );
    // Add the cursor handler.
    grabNode.addInputEventListener( new CursorHandler( CursorHandler.HAND ) );
    // Add the drag handler.
    var offsetPosToCenter = new Vector2( grabNode.getFullBoundsReference().getCenterX() - mvt.modelToViewX( beaker.position.get().getX() ), grabNode.getFullBoundsReference().getCenterY() - mvt.modelToViewY( beaker.position.get().getY() ) );
    grabNode.addInputEventListener( new ThermalElementDragHandler( beaker, grabNode, mvt, new ThermalItemMotionConstraint( model, beaker, grabNode, mvt, offsetPosToCenter ) ) );
  }

  return inherit( BeakerView, BeakerContainerView, {
// Update the clipping mask that hides energy chunks behind blocks that are in the beaker.
    updateEnergyChunkClipMask: function( model, clip ) {
      var forwardPerspectiveOffset = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( Block.SURFACE_WIDTH / 2 );
      var backwardPerspectiveOffset = EFACConstants.MAP_Z_TO_XY_OFFSET.apply( -Block.SURFACE_WIDTH / 2 );
      var clippingMask = new Area( frontNode.getFullBoundsReference() );
      for ( var block in model.getBlockList() ) {
        if ( model.getBeaker().getRect().contains( block.getRect() ) ) {
          var path = new DoubleGeneralPath();
          var rect = block.getRect();
          path.moveTo( new Vector2( rect.getX(), rect.getY() ).plus( forwardPerspectiveOffset ) );
          path.lineTo( new Vector2( rect.getMaxX(), rect.getY() ).plus( forwardPerspectiveOffset ) );
          path.lineTo( new Vector2( rect.getMaxX(), rect.getY() ).plus( backwardPerspectiveOffset ) );
          path.lineTo( new Vector2( rect.getMaxX(), rect.getMaxY() ).plus( backwardPerspectiveOffset ) );
          path.lineTo( new Vector2( rect.getMinX(), rect.getMaxY() ).plus( backwardPerspectiveOffset ) );
          path.lineTo( new Vector2( rect.getMinX(), rect.getMaxY() ).plus( forwardPerspectiveOffset ) );
          path.closePath();
          clippingMask.subtract( new Area( mvt.modelToView( path.getGeneralPath() ) ) );
        }
      }
      clip.setPathTo( clippingMask );
    }
  } );
} );

