// Copyright 2002-2012, University of Colorado
/**
 * This node was copied from BlockNode in order to prototype a block that
 * changes color as the heat capacity changes.
 * ----------------------------
 * Piccolo node that represents a block in the view.  The blocks in the model
 * are 2D, and this class gives them some perspective in order to make them
 * appear to be 3D.
 * ----------------------------
 * NOTE: This was created for heat capacity lab (HCL) prototype, and has been
 * removed from active service as of 8/6/2012.  A decision was made in late
 * March 2013 to leave this in the code in case we ever decide to revive the
 * HCL simulation.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var BasicStroke = require( 'java.awt.BasicStroke' );
  var Color = require( 'SCENERY/util/Color' );
  var Font = require( 'SCENERY/util/Font' );
  var Image = require( 'java.awt.Image' );
  var Shape = require( 'java.awt.Shape' );
  var Stroke = require( 'java.awt.Stroke' );
  var AffineTransform = require( 'java.awt.geom.AffineTransform' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Vector2 = require( 'DOT/Vector2' );
  var Property = require( 'AXON/Property' );
  var VoidFunction1 = require( 'edu.colorado.phet.common.phetcommon.util.function.VoidFunction1' );
  var ModelViewTransform = require( 'edu.colorado.phet.common.phetcommon.view.graphics.transforms.ModelViewTransform' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var PhetFont = require( 'edu.colorado.phet.common.phetcommon.view.util.PhetFont' );
  var CursorHandler = require( 'edu.colorado.phet.common.piccolophet.event.CursorHandler' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var EnergyChunk = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/model/EnergyChunk' );
  var EnergyChunkNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/view/EnergyChunkNode' );
  var Block = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/Block' );
  var ConfigurableHeatCapacityBlock = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/ConfigurableHeatCapacityBlock' );
  var EFACIntroModel = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EFACIntroModel' );
  var EnergyChunkContainerSliceNode = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/intro/model/EnergyChunkContainerSliceNode' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PImage = require( 'edu.umd.cs.piccolo.nodes.PImage' );
  var PText = require( 'edu.umd.cs.piccolo.nodes.PText' );
  var PClip = require( 'edu.umd.cs.piccolox.nodes.PClip' );
  var PComposite = require( 'edu.umd.cs.piccolox.nodes.PComposite' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Constants that define the 3D projection.  Public so that model can reference.
  var PERSPECTIVE_ANGLE = Math.atan2( -EFACConstants.Z_TO_Y_OFFSET_MULTIPLIER, -EFACConstants.Z_TO_X_OFFSET_MULTIPLIER );
  var PERSPECTIVE_EDGE_PROPORTION = Math.sqrt( Math.pow( EFACConstants.Z_TO_X_OFFSET_MULTIPLIER, 2 ) + Math.pow( EFACConstants.Z_TO_Y_OFFSET_MULTIPLIER, 2 ) );

  //private
  var LABEL_FONT = new PhetFont( 32, false );

  //private
  var OUTLINE_STROKE = new BasicStroke( 3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL );

  //private
  var OUTLINE_STROKE_COLOR = Color.DARK_GRAY;
// Debug controls.

  //private
  var SHOW_2D_REPRESENTATION = false;

  function ConfigurableHeatCapacityBlockNode( model, block, mvt ) {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------
    // Node where approaching energy chunks are placed if set.  This can be
    // used to make sure that approaching energy chunks stay "out front".

    //private
    this.approachingEnergyChunkParentNode = null;
    // shape from the position of the block.
    var scaleTransform = AffineTransform.getScaleInstance( mvt.getTransform().getScaleX(), mvt.getTransform().getScaleY() );
    // Create the shape for the front of the block.
    var blockRectInViewCoords = scaleTransform.createTransformedShape( Block.getRawShape() ).getBounds2D();
    var perspectiveEdgeSize = mvt.modelToViewDeltaX( block.getRect().getWidth() * PERSPECTIVE_EDGE_PROPORTION );
    var blockFaceOffset = new Vector2( -perspectiveEdgeSize / 2, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE );
    var backCornersOffset = new Vector2( perspectiveEdgeSize, 0 ).getRotatedInstance( -PERSPECTIVE_ANGLE );
    var lowerLeftFrontCorner = new Vector2( blockRectInViewCoords.getMinX(), blockRectInViewCoords.getMaxY() ).plus( blockFaceOffset );
    var lowerRightFrontCorner = new Vector2( blockRectInViewCoords.getMaxX(), blockRectInViewCoords.getMaxY() ).plus( blockFaceOffset );
    var upperRightFrontCorner = new Vector2( blockRectInViewCoords.getMaxX(), blockRectInViewCoords.getMinY() ).plus( blockFaceOffset );
    var upperLeftFrontCorner = new Vector2( blockRectInViewCoords.getMinX(), blockRectInViewCoords.getMinY() ).plus( blockFaceOffset );
    var blockFaceShape = new Rectangle.Number( lowerLeftFrontCorner.getX(), upperLeftFrontCorner.getY(), blockRectInViewCoords.getWidth(), blockRectInViewCoords.getHeight() );
    // Create the shape of the top of the block.
    var upperLeftBackCorner = upperLeftFrontCorner.plus( backCornersOffset );
    var upperRightBackCorner = upperRightFrontCorner.plus( backCornersOffset );
    var blockTopPath = new DoubleGeneralPath();
    blockTopPath.moveTo( upperLeftFrontCorner );
    blockTopPath.lineTo( upperRightFrontCorner );
    blockTopPath.lineTo( upperRightBackCorner );
    blockTopPath.lineTo( upperLeftBackCorner );
    blockTopPath.lineTo( upperLeftFrontCorner );
    var blockTopShape = blockTopPath.getGeneralPath();
    // Create the shape of the side of the block.
    var lowerRightBackCorner = lowerRightFrontCorner.plus( backCornersOffset );
    var blockSidePath = new DoubleGeneralPath();
    blockSidePath.moveTo( upperRightFrontCorner );
    blockSidePath.lineTo( lowerRightFrontCorner );
    blockSidePath.lineTo( lowerRightBackCorner );
    blockSidePath.lineTo( upperRightBackCorner );
    blockSidePath.lineTo( upperRightFrontCorner );
    var blockSideShape = blockSidePath.getGeneralPath();
    // Create the shape for the back of the block.
    var lowerLeftBackCorner = lowerLeftFrontCorner.plus( backCornersOffset );
    var blockBackPath = new DoubleGeneralPath();
    blockBackPath.moveTo( lowerLeftBackCorner );
    blockBackPath.lineTo( lowerRightBackCorner );
    blockBackPath.moveTo( lowerLeftBackCorner );
    blockBackPath.lineTo( lowerLeftFrontCorner );
    blockBackPath.moveTo( lowerLeftBackCorner );
    blockBackPath.lineTo( upperLeftBackCorner );
    var blockBackShape = blockBackPath.getGeneralPath();
    // Add the back of the block.
    var blockBack = new PhetPPath( blockBackShape, OUTLINE_STROKE, OUTLINE_STROKE_COLOR );
    addChild( blockBack );
    // Create the layers where the energy chunks will be placed.
    var energyChunkRootNode = new Node();
    addChild( energyChunkRootNode );
    for ( var i = block.getSlices().size() - 1; i >= 0; i-- ) {
      energyChunkRootNode.addChild( new EnergyChunkContainerSliceNode( block.getSlices().get( i ), mvt ) );
    }
    // Add the face, top, and sides of the block.
    var blockFace = createSurface( blockFaceShape, block.color, block.getFrontTextureImage() );
    var blockTop = createSurface( blockTopShape, block.color, block.getTopTextureImage() );
    var blockSide = createSurface( blockSideShape, block.color, block.getSideTextureImage() );
    addChild( blockFace );
    addChild( blockTop );
    addChild( blockSide );
    if ( SHOW_2D_REPRESENTATION ) {
      addChild( new PhetPPath( scaleTransform.createTransformedShape( Block.getRawShape() ), new BasicStroke( 1 ), Color.RED ) );
    }
    // Position and add the label.
    var label = new PText( block.getLabel() );
    label.setFont( LABEL_FONT );
    if ( label.getFullBoundsReference().width >= mvt.modelToViewDeltaX( Block.SURFACE_WIDTH * 0.9 ) ) {
      // Scale the label to fit on the face of the block.
      var scale = (mvt.modelToViewDeltaX( Block.SURFACE_WIDTH * 0.9 ) / label.getFullBoundsReference().width);
      label.setScale( scale );
    }
    var labelCenterX = (upperLeftFrontCorner.getX() + upperRightFrontCorner.getX()) / 2;
    var labelCenterY = (upperLeftFrontCorner.getY() - mvt.modelToViewDeltaY( Block.SURFACE_WIDTH ) / 2);
    label.centerFullBoundsOnPoint( labelCenterX, labelCenterY );
    addChild( label );
    // this model element and add/remove them as needed.
    block.approachingEnergyChunks.addElementAddedObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( addedEnergyChunk ) {
        var energyChunkNode = new EnergyChunkNode( addedEnergyChunk, mvt );
        var parentNode = approachingEnergyChunkParentNode == null ? energyChunkRootNode : approachingEnergyChunkParentNode;
        parentNode.addChild( energyChunkNode );
        block.approachingEnergyChunks.addElementRemovedObserver( new VoidFunction1().withAnonymousClassBody( {
          apply: function( removedEnergyChunk ) {
            if ( removedEnergyChunk == addedEnergyChunk ) {
              parentNode.removeChild( energyChunkNode );
              block.approachingEnergyChunks.removeElementRemovedObserver( this );
            }
          }
        } ) );
      }
    } ) );
    // that it looks like they are in the block.
    block.energyChunksVisible.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( energyChunksVisible ) {
        var opaqueness = energyChunksVisible ? 0.5f
        :
        1.0
        f;
        blockFace.setTransparency( opaqueness );
        blockTop.setTransparency( opaqueness );
        blockSide.setTransparency( opaqueness );
        label.setTransparency( opaqueness );
      }
    } ) );
    // Update the offset if and when the model position changes.
    block.position.addObserver( new VoidFunction1().withAnonymousClassBody( {
      apply: function( newPosition ) {
        setOffset( mvt.modelToView( newPosition ).toPoint2D() );
        // nodes can handle their own positioning.
        energyChunkRootNode.setOffset( mvt.modelToView( newPosition ).getRotatedInstance( Math.PI ).toPoint2D() );
      }
    } ) );
    // Add the cursor handler.
    addInputEventListener( new CursorHandler( CursorHandler.HAND ) );
    // Add the drag handler.
    var offsetPosToCenter = new Vector2( getFullBoundsReference().getCenterX() - mvt.modelToViewX( block.position.get().getX() ), getFullBoundsReference().getCenterY() - mvt.modelToViewY( block.position.get().getY() ) );
    addInputEventListener( new ThermalElementDragHandler( block, this, mvt, new ThermalItemMotionConstraint( model, block, this, mvt, offsetPosToCenter ) ) );
  }

  return inherit( PComposite, ConfigurableHeatCapacityBlockNode, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
      setApproachingEnergyChunkParentNode: function( node ) {
        // This should not be set more than once.
        assert && assert( approachingEnergyChunkParentNode == null );
        approachingEnergyChunkParentNode = node;
      },
      /*
       * Convenience method to avoid code duplication.  Adds a node of the given
       * shape, color, and texture (if a texture is specified).
       */

      //private
      createSurface: function( shape, fillColor, textureImage ) {
        var root = new Node();
        // provided, this may end up getting partially or entirely covered up.
        var shapeNode = new PhetPPath( shape );
        fillColor.addObserver( new VoidFunction1().withAnonymousClassBody( {
          apply: function( color ) {
            shapeNode.setPaint( color );
          }
        } ) );
        root.addChild( shapeNode );
        if ( textureImage != null ) {
          // Add the clipped texture.
          var clippedTexture = new PClip();
          clippedTexture.setPathTo( shape );
          var texture = new PImage( textureImage );
          // Scale up the texture image if needed.
          var textureScale = 1;
          if ( texture.getFullBoundsReference().width < clippedTexture.getFullBoundsReference().width ) {
            textureScale = clippedTexture.getFullBoundsReference().width / texture.getFullBoundsReference().width;
          }
          if ( texture.getFullBoundsReference().height < clippedTexture.getFullBoundsReference().height ) {
            textureScale = Math.max( clippedTexture.getFullBoundsReference().height / texture.getFullBoundsReference().height, textureScale );
          }
          texture.setScale( textureScale );
          // Add the texture to the clip node in order to clip it.
          texture.setOffset( clippedTexture.getFullBoundsReference().getMinX(), clippedTexture.getFullBoundsReference().getMinY() );
          clippedTexture.addChild( texture );
          root.addChild( clippedTexture );
        }
        // Add the outlined shape so that edges are visible.
        root.addChild( new PhetPPath( shape, OUTLINE_STROKE, OUTLINE_STROKE_COLOR ) );
        return root;
      }
    },
//statics
    {
      PERSPECTIVE_ANGLE: PERSPECTIVE_ANGLE,
      PERSPECTIVE_EDGE_PROPORTION: PERSPECTIVE_EDGE_PROPORTION
    } );
} );

