// Copyright 2002-2013, University of Colorado
/**
 * Base class for a Piccolo node that represents a thermometer in the view.
 *
 * @author John Blanco
 */
define( function( require ) {
  'use strict';

  // modules
  var inherit = require( 'PHET_CORE/inherit' );
  var Dimension2D = require( 'java.awt.geom.Dimension2D' );
  var Line2D = require( 'java.awt.geom.Line2D' );
  var Vector2 = require( 'java.awt.geom.Vector2' );
  var Rectangle = require( 'KITE/Rectangle' );
  var Function = require( 'edu.colorado.phet.common.phetcommon.math.Function' );
  var Vector2 = require( 'DOT/Vector2' );
  var DoubleGeneralPath = require( 'edu.colorado.phet.common.phetcommon.view.util.DoubleGeneralPath' );
  var CursorHandler = require( 'edu.colorado.phet.common.piccolophet.event.CursorHandler' );
  var PhetPPath = require( 'edu.colorado.phet.common.piccolophet.nodes.PhetPPath' );
  var ZeroOffsetNode = require( 'edu.colorado.phet.common.piccolophet.nodes.kit.ZeroOffsetNode' );
  var EnergyFormsAndChangesResources = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/EnergyFormsAndChangesResources' );
  var EFACConstants = require( 'ENERGY_FORMS_AND_CHANGES/energy-forms-and-changes/common/EFACConstants' );
  var Node = require( 'SCENERY/nodes/Node' );
  var PImage = require( 'edu.umd.cs.piccolo.nodes.PImage' );
  var PPath = require( 'edu.umd.cs.piccolo.nodes.PPath' );
  var PDimension = require( 'edu.umd.cs.piccolo.util.PDimension' );
  var PClip = require( 'edu.umd.cs.piccolox.nodes.PClip' );
  var PComposite = require( 'edu.umd.cs.piccolox.nodes.PComposite' );

//-------------------------------------------------------------------------
// Class Data
//-------------------------------------------------------------------------
// Constants that define the size and relative position of the triangle.
// These values will need tweaking if the images used for the thermometer
// are changed.
// In screen coordinates, which is close to pixels.

  //private
  var TRIANGLE_SIDE_SIZE = 15;

  //private
  var NUM_TICK_MARKS = 13;

  //private
  var TICK_MARK_STROKE = new BasicStroke( 2 );
  /*
   * Constructor.
   *
   * @param thermometer
   * @param mvt
   */
  function ThermometerNode() {
    //-------------------------------------------------------------------------
    // Instance Data
    //-------------------------------------------------------------------------

    //private
    this.liquidHeightMapFunction;

    //private
    this.triangle;

    //private
    this.liquidShaftWidth;

    //private
    this.liquidShaft;

    //private
    this.centerOfBulb;
    this.triangleTipOffset;
    // Root node, all children should be added to this.
    var rootNode = new Node();
    // Create and add nodes that will act as layers.
    var backLayer = new Node();
    rootNode.addChild( backLayer );
    var middleLayer = new Node();
    rootNode.addChild( middleLayer );
    var frontLayer = new Node();
    rootNode.addChild( frontLayer );
    // Add the back of the thermometer.
    var thermometerBack = new PImage( EnergyFormsAndChangesResources.Images.THERMOMETER_MEDIUM_BACK );
    backLayer.addChild( thermometerBack );
    // clip will prevent the liquid from ever appearing to pop out the top.
    var liquidShaftClipNode = new PClip();
    liquidShaftClipNode.setStroke( null );
    backLayer.addChild( liquidShaftClipNode );
    // Set up reference values for layout.
    centerOfBulb = new Vector2( thermometerBack.getFullBoundsReference().getCenterX(), thermometerBack.getFullBoundsReference().getMaxY() - thermometerBack.getFullBoundsReference().height * 0.1 );
    // Add the liquid shaft, the shape of which will indicate the temperature.
    {
      liquidShaft = new PhetPPath( new Color( 237, 28, 36 ) );
      liquidShaftClipNode.addChild( liquidShaft );
      // There are some tweak factors in here for setting the shape of the thermometer liquid.
      liquidShaftWidth = thermometerBack.getFullBoundsReference().getWidth() * 0.45;
      // Tweak multiplier to align with desired tick mark at 100 degrees C.
      var boilingPointLiquidHeight = (centerOfBulb.getY() - thermometerBack.getFullBoundsReference().getMinY()) * 0.84;
      // Tweak multiplier to align with desired tick mark at 0 degrees C.
      var freezingPointLiquidHeight = thermometerBack.getFullBoundsReference().height * 0.21;
      liquidHeightMapFunction = new Function.LinearFunction( EFACConstants.FREEZING_POINT_TEMPERATURE, EFACConstants.BOILING_POINT_TEMPERATURE, freezingPointLiquidHeight, boilingPointLiquidHeight );
      // manually coordinated with the image used for the thermometer.
      var clipPath = new DoubleGeneralPath().withAnonymousClassBody( {
        initializer: function() {
          var thermometerTopY = thermometerBack.getFullBoundsReference().getMinY();
          var curveStartOffset = thermometerBack.getFullBoundsReference().getHeight() * 0.05;
          var clipWidth = liquidShaftWidth * 1.1;
          var centerX = thermometerBack.getFullBoundsReference().getCenterX();
          moveTo( centerX - clipWidth / 2, centerOfBulb.getY() );
          lineTo( centerX - clipWidth / 2, thermometerTopY + curveStartOffset );
          curveTo( centerX - clipWidth / 4, thermometerTopY - curveStartOffset / 4, centerX + clipWidth / 4, thermometerTopY - curveStartOffset / 4, centerX + clipWidth / 2, thermometerTopY + curveStartOffset );
          lineTo( centerX + clipWidth / 2, centerOfBulb.getY() );
          closePath();
        }
      } );
      liquidShaftClipNode.setPathTo( clipPath.getGeneralPath() );
    }
    // Add the image for the front of the thermometer.
    frontLayer.addChild( new PImage( EnergyFormsAndChangesResources.Images.THERMOMETER_MEDIUM_FRONT ) );
    // Add the tick marks.  There are some tweak factors here.
    var tickMarkXOffset = thermometerBack.getFullBoundsReference().width * 0.3;
    var shortTickMarkWidth = thermometerBack.getFullBoundsReference().width * 0.1;
    var longTickMarkWidth = shortTickMarkWidth * 2;
    var tickMarkMinY = centerOfBulb.getY() - thermometerBack.getFullBoundsReference().getHeight() * 0.15;
    var tickMarkSpacing = ((tickMarkMinY - thermometerBack.getFullBoundsReference().getMinY()) / NUM_TICK_MARKS) * 0.945;
    for ( var i = 0; i < NUM_TICK_MARKS; i++ ) {
      // Tick marks are set to have a longer one at freezing, boiling, and half way between.
      var tickMarkShape = new Line2D.Number( 0, 0, (i - 1) % 5 == 0 ? longTickMarkWidth : shortTickMarkWidth, 0 );
      var tickMark = new PhetPPath( tickMarkShape, TICK_MARK_STROKE, Color.BLACK );
      tickMark.setOffset( tickMarkXOffset, tickMarkMinY - i * tickMarkSpacing );
      frontLayer.addChild( tickMark );
    }
    // on the left side of this triangle.
    triangleTipOffset = new PDimension( -thermometerBack.getWidth() / 2 - TRIANGLE_SIDE_SIZE * Math.cos( Math.PI / 6 ) - 2, thermometerBack.getHeight() / 2 - thermometerBack.getWidth() / 2 );
    var triangleLeftmostPoint = new Vector2( backLayer.getFullBoundsReference().getCenterX() + triangleTipOffset.getWidth(), backLayer.getFullBoundsReference().getCenterY() + triangleTipOffset.getHeight() );
    var triangleUpperRightPoint = triangleLeftmostPoint.plus( new Vector2( TRIANGLE_SIDE_SIZE, 0 ).getRotatedInstance( Math.PI / 5 ) );
    var triangleLowerRightPoint = triangleLeftmostPoint.plus( new Vector2( TRIANGLE_SIDE_SIZE, 0 ).getRotatedInstance( -Math.PI / 5 ) );
    var trianglePath = new DoubleGeneralPath( triangleLeftmostPoint ).withAnonymousClassBody( {
      initializer: function() {
        lineTo( triangleUpperRightPoint );
        lineTo( triangleLowerRightPoint );
        closePath();
      }
    } );
    triangle = new PhetPPath( trianglePath.getGeneralPath(), new Color( 0, 0, 0, 0 ), new BasicStroke( 2 ), Color.BLACK );
    middleLayer.addChild( triangle );
    // be at point (0, 0).
    addChild( new ZeroOffsetNode( rootNode ) );
    // Add the cursor handler.
    addInputEventListener( new CursorHandler( CursorHandler.HAND ) );
  }

  return inherit( PComposite, ThermometerNode, {
//-------------------------------------------------------------------------
// Methods
//-------------------------------------------------------------------------
    getOffsetCenterShaftToTriangleTip: function() {
      return new Vector2( -getFullBoundsReference().getWidth() + liquidShaft.getFullBoundsReference().getCenterX(), getFullBoundsReference().getHeight() / 2 );
    },
    setSensedColor: function( sensedColor ) {
      triangle.setPaint( sensedColor );
    },
    setSensedTemperature: function( temperature ) {
      var liquidShaftHeight = liquidHeightMapFunction.evaluate( temperature );
      liquidShaft.setPathTo( new Rectangle.Number( centerOfBulb.getX() - liquidShaftWidth / 2 + 0.75, centerOfBulb.getY() - liquidShaftHeight, liquidShaftWidth, liquidShaftHeight ) );
    }
  } );
} );

